import scipy.optimize

import numpy

def smooth(x,window_len=10,window='hanning'):
    """smooth the data using a window with requested size.
    
    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal 
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.
    
    input:
        x: the input signal 
        window_len: the dimension of the smoothing window
        window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
            flat window will produce a moving average smoothing.

    output:
        the smoothed signal
        
    example:

    t=linspace(-2,2,0.1)
    x=sin(t)+randn(len(t))*0.1
    y=smooth(x)
    
    see also: 
    
    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.lfilter
 
    TODO: the window parameter could be the window itself if an array instead of a string   
    """

    if x.ndim != 1:
        raise ValueError, "smooth only accepts 1 dimension arrays."

    if x.size < window_len:
        raise ValueError, "Input vector needs to be bigger than window size."


    if window_len<3:
        return x


    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError, "Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'"


    s=numpy.r_[2*x[0]-x[window_len:1:-1],x,2*x[-1]-x[-1:-window_len:-1]]
    #print(len(s))
    if window == 'flat': #moving average
        w=ones(window_len,'d')
    else:
        w=eval('numpy.'+window+'(window_len)')

    y=numpy.convolve(w/w.sum(),s,mode='same')
    return y[window_len-1:-window_len+1]

def analytic_single_ss(z,phi_0,c0,valence=2):
    from numpy import linspace, trapz, exp
    kappa = 1.0
    return c0*exp(-valence*phi_0*exp(-kappa*z))

def analytic_ss(mesh,phi_0,c0,valence=2,bulk_conc=False):
    from numpy import linspace, trapz, exp
    # Analytic func
    z_vec = mesh.coordinates()[:,-1]
    kappa   = 1.0
    if not bulk_conc:
        h   = mesh.coordinates()[:,-1].max()
        dz  = 0.001
        z   = linspace(0,h,h/dz+1)
        
        phi = phi_0*exp(-kappa*z)
        c_vec = exp(-valence*phi)
        
        weight = c0*h/trapz(c_vec,dx=dz)
        return weight*exp(-valence*phi_0*exp(-kappa*z_vec))
    return c0*exp(-valence*phi_0*exp(-kappa*z_vec))
        

def ext_logspace(start, stop, num = 50, first_step = 1.0):
    " An extended logspace"
    from numpy import cumsum, arange, insert
    fmin = scipy.optimize.optimize.fmin
    
    def basefind_cost(base,start,stop,num,weight):
        vec = cumsum((base**arange(num-1))/weight)
        return abs(vec[-1]-stop)

    opt = {}
    opt["xtol"]    = 1e-13
    opt["ftol"]    = 1e-13
    opt["maxfun"]  = 600
    opt["maxiter"] = 300
    
    weight = 1/first_step
    base0 = 1.1
    base = fmin(basefind_cost,base0,(start,stop,num,weight),**opt)
    dx = insert((base**arange(num-1))/weight,0,0)
    return cumsum(dx)

