#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Johan Hake (johan.hake@gmail.com)"
__copyright__ = "Copyright (C) 2008 Johan Hake"
__license__  = "GNU LGPL Version 3.0 or any later"
__date__ = "2008-02-19 -- 2008-09-12"

# NOTE: We need to import scipy.optimize before dolfin!!
#       Otherwise -> segfault

__all__ = ['default_params','setup','run_system','post_process']

from help_func import ext_logspace, analytic_ss

from diffsim import *
from diffsim.dyadiccleft import *

from diffsim.logger import *
from os import path

set_log_level(DEBUG,ALL)
set_log_level(CRITICAL,"instant")
#et_log_level(INFO,'diffsim.diffusiveligand')


_global_names = ParameterDict(dd = None,
                              pl = None,
                              dl = None,
                              sf = None,
                              case_name = None)

scale_dict = {0:3.0, 1:2.5, 2:2.0, 3:1.8, 4:1.6, 5:1.4, 6:1.2,
              7:1.0, 8:0.8, 9:0.6, 10:0.4,11:0.2}
nice_num_iter = filter(lambda item:item[1]==1.0,scale_dict.iteritems())[0][0]
def default_params():
    # Ca specific parameters
    p = ParameterDict(
        below_zero_tol      = 0.1,
        crank_nich          = False,
        c0                  = 0.1,
        D_Ca                = 1e5,
        tstop               = 1.0,
        dt_weight_start     = 1e-6,
        dt_weight_max       = 5.0,
        dt_max              = 1.0,
        dt_start            = 1e-10,
        dt_min              = 1e-10,
        use_scale_dt        = True,
        dt_scale            = 1.5,
        use_cut_plane_plot  = 0,
        use_plot_TT         = 0,
        use_plot_SR         = 0,
        plot_below_zero     = 0,
        plot_range          = [0,5],
        channels            = [2],#[0],
        lcc_amp             = 0.2,
        constant_flux       = 1,
        ryr_channels        = 1,
        markov_model_scale  = 2.0,
        cytosole            = 1,
        cytosole_dn         = 30.0,
        save_data           = 0,
        value_logger        = 1,
        plot_saved_data     = 0,
        run_times           = 1,
        python_func         = False,
        vector_degree       = 1,
        hight               = 12.5,
        radius              = 20,
        resolution          = 0,
        phi_0               = -2.2,
        SUPG_type           = 3,
        tau_dict            = {0:0.6,1:0.6,2:0.6},#{0:0.22,1:0.28,2:0.38}, # From experiments
        )
    return p

def setup(p):
    # Clear any previously created DiffsimObjects
    clear_all()

    # Define mesh name
    mesh_name = "mesh/cleft_mesh_3D_h%s_r%s_res_%d.xml.gz"%(str(p.hight),str(p.radius),p.resolution)

    # Instantiate the DiffsimObjects
    dd = DiffusionDomain("Dyadic_cleft",mesh_name,vector_degree = p.vector_degree)
    if abs(p.phi_0) > 0.0:
        c0 = analytic_ss(dd.mesh(),p.phi_0,p.c0)
    else:
        c0 = p.c0
    
    dl = DiffusiveLigand(dd.name(),"Ca",c0)
    if abs(p.phi_0) > 0.0 and p.resolution != 0:
        sf = StaticField("Bi_lipid_field",dd.name(),python_func = p.python_func)
        _global_names.sf = sf.name()
    
    # Store some global names
    _global_names.dd = dd.name()
    _global_names.dl = dl.name()

    # Init any boundaries
    times  = [0,p.tstop/2]
    values = [[1,0],[0,1]]
    for i, m in enumerate(p.channels):
        vc = ValueClamp("LCC",times,values[i], True)
        if p.constant_flux:
            ConstantChannelFlux("LCC",dd.name(), "Ca",m, amp = p.lcc_amp,
                                activator = vc)
        else:
            LCCVoltageDepFlux("LCC",dd.name(), "Ca",m,
                              activator = vc)
    
    if not p.constant_flux:
        ValueClamp("V",[0.0],[0.0])
    
    
    if p.ryr_channels:
        ryr_markers = []
        if p.radius == 20:
            ryr_markers = [20]
        
        if p.radius == 50:
            ryr_markers = range(20,24)
        
        if p.radius == 100:
            ryr_markers = range(20,36)
        
        for i, m in enumerate(ryr_markers):
            RyRMarkovModel_Stern("RyR_%d"%m,m,True)
    
    if p.cytosole:
        ConcDepLinearFluxBoundary("Cytosole",dd.name(),"Ca",100, u0 = p.c0,\
                                  dn = p.cytosole_dn)
    
    if p.use_cut_plane_plot:
        pl0 = Plotter(dl.name(),plot_range = p.plot_range,
                      cutplane_origin = (0,0,0),#(0,0,3.02),
                      cutplane_normal = (0,1,0),#(0,0,1),
                      below_zero = p.plot_below_zero,
        #                save_png = True,
                      )
    
    if p.use_plot_SR:
        pl1 = Plotter(dl.name(),plot_range = plot_range,
                      elevate = -60, below_zero = p.plot_below_zero)
    
    if p.use_plot_TT:
        pl2 = Plotter(dl.name(),plot_range = p.plot_range,
                      elevate = -60,azimuth = 180, below_zero = p.plot_below_zero)
        
    if p.save_data:
        from numpy import zeros, linspace
        N = 101
        max_z = dd.mesh().coordinates()[:,2].max()
        min_z = dd.mesh().coordinates()[:,2].min()
        points = zeros((N,3))
        points[:,2] = ext_logspace(min_z,max_z,N,0.01)
        
        pl = PointLogger(dl.name(),points)
        _global_names.pl = pl.name()
    
    if p.value_logger:
        sl = ValueLogger(dl.name(),[1,2,3,20,100],[MIN,MAX,MEAN])
    
    # Define case name
    _global_names.case_name = path.basename(__file__).split('.')[0] + \
                              "_h%s_r%d_res_%d_vec_deg_%d_SUPG_TYPE_%d_tau_%.2f"%\
                              (str(p.hight),p.radius,p.resolution,\
                               p.vector_degree,p.SUPG_type,p.tau_dict[p.resolution])
    
    return dict(DiffsimObject.iter_objects()), _global_names.case_name
    
    
def run_system(p):
    # Setting the parameters
    params = get_params()
    
    params.time.dt_max          = p.dt_max
    params.time.tstop           = p.tstop
    
    params.Dyadic_cleft_Dt.weight_start = p.dt_weight_start
    params.Dyadic_cleft_Dt.weight_max   = p.dt_weight_max  
    
    params.Dyadic_cleft_Dt.scale      = p.dt_scale
    params.Dyadic_cleft_Dt.minimal    = p.dt_min
    #params.Dyadic_cleft_Ca_Dt.weight_scale  = p.dt_min
    
    
    params.solver.tol_rel       = 0.001
    params.solver.tol_abs       = 0.0001
    params.solver.tol_ss        = 0.00001
    params.solver.crank_nich    = p.crank_nich
    
    if abs(p.phi_0) > 0.0 and p.resolution != 0:
        params[_global_names.sf].phi_0     = p.phi_0
        params[_global_names.sf].SUPG_type = p.SUPG_type
        params[_global_names.sf].tau       = p.tau_dict[p.resolution]

    print params
    
    # Run system
    if p.run_times > 0:
        run_sim(p.run_times,_global_names.case_name)
    else:
        init_system()

def post_process(p):
    phi_0 = p.phi_0
    dd = DiffsimObject.get_object(_global_names.dd)
    dl = DiffsimObject.get_object(_global_names.dl)
    L2_w = None
    Max_w = None
    L2_n  = None
    if abs(phi_0) != 0.0 and p.run_times > 0:
        from numpy import linspace, exp, trapz
        
        # Analytic func
        h   = dd.mesh().coordinates()[:,-1].max()
        dz  = 0.001
        z   = linspace(0,h,h/dz+1)
        valence = 2.0
        kappa   = 1.0
        
        phi = phi_0*exp(-kappa*z)
        c_vec = exp(-valence*phi)
        
        weight = p.c0*h/trapz(c_vec,dx=dz)
        
        z_vec = dd.mesh().coordinates()[:,-1]
        analytic = weight*exp(-valence*phi_0*exp(-kappa*z_vec))
        
        if not p.channels and not p.cytosole:
            L2_n = dd.normalized_L2_norm(dl.value(),analytic)
            L2_w = dd.L2_norm_w(analytic-dl.value())
            Max_w = dd.Max_norm_w(analytic-dl.value())
            print "Simulated weighted L2 norm  :%13.5e"%L2_w
            print "Simulated weighted Max norm :%13.5e"%Max_w
            print "Simulated normalized L2 norm:%13.5e"%L2_n
        
        if p.plot_saved_data:
            pass
            from scitools import easyviz
            pl = DiffsimObject.get_object(_global_names.pl)
            data = load(pl.file_name())
            
            values = data["values"]
            time   = data["time"]
            z_vec2 = data['points'][:,2]
            analytic2 = weight*exp(-valence*phi_0*exp(-kappa*z_vec2))
            
            plot_basename = path.abspath(path.join("figures",\
                                path.basename(pl.file_name().split(".cpickle")[0])))
            easyviz.plot(z_vec2,values[-1,:],z_vec2,analytic2,
                         legend=("Numeric", "Analytic"),
                         title = _global_names.case_name.replace("_","-"))
            
            easyviz.hardcopy(plot_basename + ".eps", color = True, \
                             enhanced = True, fontname = "cmr10")
            converting = Popen("eps2pdf  %s.eps --outfile  %s.pdf"%\
                               (plot_basename,plot_basename),shell=True)
            converting.wait()
        
    return _global_names.case_name, L2_w, Max_w, L2_n

if __name__ == "__main__":
    params = default_params()
    params.tstop = 10.0
    params.radius = 20
    params.resolution = 2
    params.constant_flux = 0
    params.cytosole = 0
    params.channels = []
    params.ryr_channels = 0
    params.run_times = 1
    
    setup(params)
    run_system(params)
    a,b,c,d = post_process(params)
    print a,b,c,d
