__author__ = "Johan Hake (johan.hake@gmail.com)"
__copyright__ = "Copyright (C) 2008 Johan Hake"
__license__  = "GNU LGPL Version 3.0 or any later"
__date__ = "2008-02-19 -- 2008-11-27"

__all__ = ['default_params','setup','run_system','post_process']

from help_func import analytic_ss

from diffsim import *
from diffsim.dyadiccleft import *

from diffsim.logger import *
from os import path

set_log_level(DEBUG,ALL)
set_log_level(CRITICAL,"instant")
set_log_level(CRITICAL,"UFL")
#set_log_level(INFO,'diffsim.diffusiveligand')

_global_names = ParameterDict(dd = None,
                              pl = None,
                              sf = None,
                              case_name = None)
def default_params():
    # Ca specific parameters
    p = ParameterDict(
        time   = ParameterDict(
                 tstop           = 1.0,
                 dt_weight_start = 10.,
                 dt_weight_max   = 1.e7,
                 dt_weight_scale = 2.,
                 dt_max          = 1.0,
                 dt_start        = 1.0,
                 dt_min          = 1e-10,
                 use_scale_dt    = True,
                 dt_scale        = 1.5,
                 dt_value_stat   = "max",
                 num_runs        = 1,
                 prec            = "amg",
                 cn              = 0,
                 tr_bdf2         = OptionCheck(0,[0,1]),
                 num_previous_values = 4,
                 dt_num_previous_values = 3,
                 newton          = 0,
                 nonlinear       = 0,
                 nl_tol_abs      = 1.0e-5,
                 relax_list      = [1.0,0.5,0.25,0.1],
                 max_fix_point_iter = 50,
                 ),
        
        io     = ParameterDict(
                 use_cut_plane_plot  = 0,
                 use_plot_TT         = 0,
                 use_plot_SR         = 0,
                 plot_below_zero     = 0,
#                 plot_range          = [0,5],
                 save_traces         = 0,
                 save_data           = 0,
                 plot_saved_data     = 0,
                 use_log_file        = 0,
                 dolfin_silent       = 0,
                 case_name           = "",
                 ),
        
        domain = ParameterDict(
                 T                    = 37.,
                 phi_0                = -2.2,
                 hight                = 12.0,
                 radius               = 110,
                 resolution           = 2,
                 cytosole             = 1,
                 cytosole_dn          = 10.0,
                 ryr_channels         = 1,
                 markov_model_scale   = 1.0,
                 ncx                  = [],
                 ncx_Kd_Act           = 0.256,
                 max_dist_ncx         = -1.0,
                 same_boundary_tensor = [],
                 #mesh_str            = "mesh/cleft_mesh_3D_h%s_r%s_res_%d.xml.gz"
                 ),

        voltage = ParameterDict(
                  sigmoidal           = 1,
                  max_step            = 1.,
                  sigma               = .2, 
                  voltage_step        = 10.0,
                  holding_pot         = -80.0,
                  clamp_time          = 1.,
                  minimize_dt         = 0,
                  ),

        ions    = ParameterDict(
                  Ca      = ParameterDict(
                            D               = 1.e5,
                            use_vc       = 0,
                            vc_times        = [0.,2.3,2.7,3.2,3.6],
                            vc_values       = [0,1,0,1,0],
                            plot_range      = [0.,5.],
                            valuelogger     = [1,2,3,20,100] + range(70,82),
                            valuelogger_to_screen = 1,
                            use             = 1,
                            lcc_channels    = [2],#,4],#[0],
                            lcc_amp         = 0.2,
                            lcc_CaO_part    = 1.0,
                            constant_flux   = 0,
                            nonlinear_flux  = 0,
                            u0              = 0.1,
                            ),
                  
                  Na      = ParameterDict(
                            D             = .5e5,
                            use_vc        = 0,
                            vc_times      = [0.,1.3,1.6,1.9,2.1],
                            vc_values     = [0,1,0,1,0],
                            plot_range    = [10.e3,15.e3],
                            valuelogger   = [1,2,3,20,60,100] + range(70,82),
                            valuelogger_to_screen = 1,
                            use           = 1,
                            na_channels   = [60],#,4],#[0],
                            na_g          = 20.,
                            u0            = 10.0e3,
                            ),
                  ),
        
        fem     = ParameterDict(
                  python_func         = False,
                  vector_degree       = 1,
                  SUPG_type           = 3,
                  tau_dict            = {0:0.6,1:0.6,2:0.6},#{0:0.22,1:0.28,2:0.38}, # From experiments
                  ),
        )
    return p

use_efield = False

def setup(p):
    global use_efield

    # Set the temperature
    set_T(p.domain.T)
    
    # Define case name
    if p.io.case_name:
        case_name_append = "_" + p.io.case_name
    else:
        case_name_append = ""
    _global_names.case_name = path.basename(__file__).split('.')[0] + \
                              case_name_append
                              
    if p.io.use_log_file:
        filename =  _global_names.case_name + "_" + present_time_str() + ".log"
        if path.isdir("log"):
            filename = path.join("log",filename)
        set_log_destination(FILE,filename)

    if p.io.dolfin_silent:
        from dolfin import dolfin_set
        dolfin_set("output destination", "silent")
    
    # Clear any previously created DiffsimObjects
    clear_all()

    # Define mesh name
    mesh_name = "mesh/cleft_mesh_3D_h%s_r%s_res_%d_dist_ncx_na%s.xml.gz"%\
                (str(p.domain.hight), str(p.domain.radius),\
                 p.domain.resolution, str(p.domain.max_dist_ncx))
    
    # Instantiate the DiffsimObjects
    dd = DiffusionDomain("Dyadic_cleft",
                         mesh_name,
                         vector_degree        = p.fem.vector_degree,
                         same_boundary_tensor = p.domain.same_boundary_tensor)
    
    # Store some global names
    _global_names.dd = dd.name()

    # If using electric field
    if abs(p.domain.phi_0) > 0.0 and p.domain.resolution != 0:
        use_efield = True

    # Check for included ions
    for ion in p.ions.keys():
        if p.ions[ion].use:
            #if use_efield:
            #    u0 = analytic_ss(dd.mesh(),p.domain.phi_0,p.ions[ion].u0)
            #else:
            #    u0 = p.ions[ion].u0
            u0 = p.ions[ion].u0
            
            dl = DiffusiveLigand(dd.name(),ion,u0,p.ions[ion].D)
            
            if p.io.use_cut_plane_plot:
                pl0 = Plotter(dl.name(),plot_range = p.ions[ion].plot_range,
                              cutplane_origin = (0,0,0),
                              cutplane_normal = (0,1,0),
                            )
            
            if p.io.use_plot_SR:
                pl1 = Plotter(dl.name(),plot_range = p.ions[ion].plot_range,
                            elevate = -60)
            
            if p.io.use_plot_TT:
                pl2 = Plotter(dl.name(),plot_range = p.ions[ion].plot_range,
                            elevate = -60,azimuth = 180)
            
            if p.io.save_traces:
                from numpy import zeros, linspace
                N = 101
                max_z = dd.mesh().coordinates()[:,2].max()
                min_z = dd.mesh().coordinates()[:,2].min()
                points = zeros((N,3))
                points[:,2] = ext_logspace(min_z,max_z,N,0.01)
                
                pl = PointLogger(dl.name(),points)
                _global_names.pl = pl.name()
            
            if p.ions[ion].valuelogger:
                sl = ValueLogger(dl.name(),p.ions[ion].valuelogger,[MIN,MAX,MEAN],\
                                 p.ions[ion].valuelogger_to_screen)#,STD])
                
            if p.domain.cytosole:
                ConcDepLinearFluxBoundary("Cytosole"+"_"+ion,dd.name(),ion,100,
                                          u0 = p.ions[ion].u0,
                                          dn = p.domain.cytosole_dn)
    
    # Add static electric field, must be done after a DiffusiveLigand is added
    if use_efield:
        sf = StaticField("Bi_lipid_field",dd.name(),python_func = p.fem.python_func)
        _global_names.sf = sf.name()

    
    # Ca specific initialization
    if p.ions.Ca.use:
        # Init the stochastic LCC
        if p.ions.Ca.use_vc:
            act = ValueClamp("LCC",p.ions.Ca.vc_times,p.ions.Ca.vc_values, True)
        else:
            act = LCCMarkovModel_Greenstein
        for i, m in enumerate(p.ions.Ca.lcc_channels):
            if p.ions.Ca.constant_flux:
                ConstantChannelFlux("LCC",dd.name(), "Ca",m,
                                    amp = p.ions.Ca.lcc_amp,
                                    activator = act)
            elif p.ions.Ca.nonlinear_flux:
                LCCVoltageDepNonlinearFlux(dd.name(),m,
                                  activator = act)
            else:
                LCCVoltageDepFlux(dd.name(),m,
                                  activator = act)
                
        # Only add RyR if we have included Ca ions
        if p.domain.ryr_channels:
            ryr_markers = []
            if p.domain.radius == 20:
                ryr_markers = [20]
            
            if p.domain.radius == 50:
                ryr_markers = range(20,24)
            
            if p.domain.radius == 100:
                ryr_markers = range(20,36)
            
            for i, m in enumerate(ryr_markers):
                RyRMarkovModel_Stern("RyR_%d"%m,m,True)
    
    
    # Na specific initialization
    if p.ions.Na.use:
        # Init the stochastic Na channel
        if p.ions.Na.na_channels:
            if not p.ions.Ca.lcc_channels:
                if p.voltage.clamp_time != 0.0:
                    times = [0.0,p.voltage.clamp_time]
                    values = [p.voltage.holding_pot,p.voltage.voltage_step]
                else:
                    times = [p.voltage.clamp_time]
                    values = [p.voltage.voltage_step]
                if p.voltage.sigmoidal:
                    SigmoidalVoltageStep(times,values,p.voltage.sigma,p.voltage.max_step)
                else:
                    VoltageStep(times,values,bool(p.voltage.minimize_dt))
            if p.ions.Na.use_vc:
                act = ValueClamp("Na",p.ions.Na.vc_times,p.ions.Na.vc_values, True)
            else:
                act = NaMarkovModel_Clancy
            for i, m in enumerate(p.ions.Na.na_channels):
                NaVoltageDepFlux(dd.name(),m,g_Na = p.ions.Na.na_g,
                                 activator = act)
    # If using both Ca and Na
    if p.ions.Na.use and p.ions.Ca.use:
        # Iterate over any NCXs
        for m in p.domain.ncx:
            NCX(dd.name(),m)
            

    # If using any of Ca and Na
    if p.ions.Na.use or p.ions.Ca.use:
        
        # If using voltage dep channels add Voltage step
        if p.ions.Ca.lcc_channels or p.ions.Na.na_channels:
            if p.voltage.clamp_time != 0.0:
                times = [0.0,p.voltage.clamp_time]
                values = [p.voltage.holding_pot,p.voltage.voltage_step]
            else:
                times = [p.voltage.clamp_time]
                values = [p.voltage.voltage_step]
            
            if p.voltage.sigmoidal:
                SigmoidalVoltageStep(times,values,p.voltage.sigma,p.voltage.max_step)
            else:
                VoltageStep(times,values,bool(p.voltage.minimize_dt))
    
    return dict(DiffsimObject.iter_objects()), _global_names.case_name
    
    
def run_system(p):
    # Setting the parameters
    params = get_params()

    params.time.dt_max          = p.time.dt_max
    params.time.tstop           = p.time.tstop
    
    params.io.save_data         = bool(p.io.save_data)

    
    params.Dyadic_cleft_Dt.value_stat   = p.time.dt_value_stat
    params.Dyadic_cleft_Dt.weight_start = p.time.dt_weight_start
    params.Dyadic_cleft_Dt.num_previous_values = p.time.dt_num_previous_values
    params.Dyadic_cleft_Dt.weight_max   = p.time.dt_weight_max  
    
    params.Dyadic_cleft_Dt.dt_scale      = p.time.dt_scale
    params.Dyadic_cleft_Dt.weight_scale  = p.time.dt_weight_scale
    params.Dyadic_cleft_Dt.minimal    = p.time.dt_min
    params.Dyadic_cleft_Dt.start      = p.time.dt_min if use_efield else  p.time.dt_max 
    #params.Dyadic_cleft_Ca_Dt.weight_scale  = p.dt_min
    
    if p.domain.ncx and "NCX" in params:
        params.NCX.Kd_Act   = p.domain.ncx_Kd_Act
        
    params.solver.prec          = p.time.prec
    params.solver.tol_rel       = 1.e-10
    params.solver.tol_abs       = 1.e-7
    params.solver.tr_bdf2       = bool(p.time.tr_bdf2)
    params.solver.crank_nich    = bool(p.time.cn)
    params.solver.newton        = bool(p.time.newton)
    params.solver.nonlinear     = bool(p.time.nonlinear)
    params.solver.nl_tol_abs    = p.time.nl_tol_abs
    params.solver.relax_list    = p.time.relax_list
    params.solver.max_fix_point_iter = p.time.max_fix_point_iter
    params.solver.num_previous_values = p.time.num_previous_values
    
    if hasattr(params,'RyRMarkovChain_Stern'):
        params.RyRMarkovChain_Stern.scale = p.domain.markov_model_scale

    #if hasattr(params,'LCCMainGate'):
    #    params.LCCMainGate.scale = p.domain.markov_model_scale

    if hasattr(params,'LCCVoltageDepFlux'):
        params.LCCVoltageDepFlux.CaO_part = p.ions.Ca.lcc_CaO_part
    
    if use_efield:
        params[_global_names.sf].phi_0     = p.domain.phi_0
        params[_global_names.sf].SUPG_type = p.fem.SUPG_type
        params[_global_names.sf].tau       = p.fem.tau_dict[p.domain.resolution]
    
    debug(str(params))
    
    # Run system
    if p.time.num_runs > 0:
        run_sim(p.time.num_runs,_global_names.case_name)
    else:
        init_system()

def post_process(p):
    return _global_names.case_name#, L2_w, Max_w, L2_n
        
if __name__ == "__main__":
    params = default_params()
    params.domain.radius         = 100
    params.domain.T              = 25.
    params.domain.resolution     = 0
    params.domain.hight          = 12.0
    params.domain.ryr_channels   = 0
    params.domain.ncx            = range(70,82)
    
    params.ions.Ca.use           = 1
    params.time.tstop            = 10.0 
    params.time.dt_scale         = 1.2
    params.time.num_runs         = 1
    params.io.use_cut_plane_plot = 0
    
    #print params.optstr()
    
    optparser = params.optparser()
    optparser.parse_args()
    
    objs = setup(params)
    import sys
    profile = False
    if len(sys.argv) > 1:
        if 'profile' in sys.argv:
            import hotshot, hotshot.stats
            profile = True
            profile_file = "%s.prof"%sys.argv[0].split('.')[0]
            profiler = hotshot.Profile(profile_file)
            profiler.start()
    run_system(params)
    a = post_process(params)
    print "Case name:",a

    if profile:
        profiler.stop()
        profiler.close()
        stats = hotshot.stats.load(profile_file)
        stats.strip_dirs()
        stats.sort_stats('time')
        stats.print_stats()
