__author__ = "Johan Hake (johan.hake@gmail.com)"
__copyright__ = "Copyright (C) 2008 Johan Hake"
__license__  = "GNU LGPL Version 3.0 or any later"
__date__ = "2008-02-19 -- 2010-02-02"

# Defining the import * statement
__all__ = ['get_params','init_system','run_sim','single_run',\
           'clear_all','set_T','get_T']

# Import diffsim stuff
from diffsim.base            import(DiffsimObject,
                                    DiscreteVariable,
                                    ContinuousVariable,
                                    Singleton,
                                    )

from diffsim.diffusiondomain       import DiffusionDomain
from diffsim.diffusiveligand       import DiffusiveLigand
from diffsim.diffusiveliganddt     import DiffusiveLigandDt
from diffsim.io                    import Plotter, PointLogger
from diffsim.signal                import Signal
from diffsim.parameterdict         import (ParameterDict, TypeCheck,
                                           RangeCheck, OptionCheck)
from diffsim.stochastictimestepper import (StochasticTimeStepper,
                                           DiscreteMarkovModel,
                                           DiscreteMarkovChain)
from diffsim.disk                  import save
from diffsim.processtime           import *

# Initialize the logging system
from logger           import *
init_local_logger(__name__,globals())

# Import numpy
import numpy

_time_params = ParameterDict(tstop     = RangeCheck(10.0,gt=0,lt=1e5),
                             dt_max    = RangeCheck(1.0,gt=0,lt=100),
                             num_runs  = RangeCheck(1 ,ge=0))

_io_params   = ParameterDict(save_data = TypeCheck(True),
                             collect_state_info = TypeCheck(False))

def set_T(T):
    " Set the temperature"
    DiffsimObject.set_T(T)

def get_T():
    " Get the temperature"
    return DiffsimObject.get_T()

def clear_all():
    " Clear all registered objects and re-establish the time handling objects"
    global _time_handler, _system_initialized, _object_dependencies, \
           _indep_cont, _dep_cont
    
    DiffsimObject.clear_all()
    
    _system_initialized = False
    _object_dependencies = []
    _indep_cont          = []
    _dep_cont            = []
    
    _time_handler = TimeHandler()

def get_params():
    """ Return a ParameterDict with default values

    Parameters can be changed the returned disctionary and passed to the
    set_params function to actually set the parameters to new values.

    """
    kwargs = dict(time   = _time_params,
                  solver = ContinuousVariable.get_solver_params(),
                  io     = _io_params,)
    id_list = []
    for name, obj in DiffsimObject.iter_objects():
        if not name in ["tstop","time"]:
            p = obj.get_params()
            # If there are no parameter registered for this obj, do not use it
            # or if the same parameters allready has been registered.
            if len(p) > 0 and id(p) not in id_list:
                if hasattr(type(obj),"_params"):
                    kwargs[type(obj).__name__] = p
                else:
                    kwargs[name] = p
                id_list.append(id(p))
    return ParameterDict(**kwargs)

def init_system():
    " Initialize the connections, and register precedence of countinuous variables"
    global _object_dependencies, _indep_cont, _dep_cont, _system_initialized,\
           _time_handler
    
    # Called once for each call to single_run()
    info("Initializing the system")
    
    # FIXME: Get a way to find out when we need to set _system_initialized to False
    if not _system_initialized:
        connect_DiffsimObjects()
        _object_dependencies = object_dependencies()
        _indep_cont, _dep_cont = continuous_dependencies()
        _system_initialized = True
    reset_system(_object_dependencies)
    _time_handler.initial_events()

def _format_time(sec):
    "A simple formater for presenting times"
    sec = int(sec)
    hours = sec/3600
    out_str = ""
    if hours > 0:
        out_str += "%dh."%hours
        sec = sec%(hours*3600)
    minutes = sec/60
    if minutes > 0:
        out_str += "%dm."%minutes
        sec = sec%(minutes*60)
    elif out_str !="":
        out_str += "0m."
    out_str += "%ds."%sec
    return out_str

def run_sim(num_runs = None, case_name = "simulation"):
    " Run the simulation 'num_runs' time"
    global _time_params
    tic("All runs")
    run_str = "run no: %d"
    if num_runs is None:
        num_runs = _time_params.num_runs
    else:
        _time_params.num_runs = num_runs

    # If an object can set the case_name then do it
    for name, obj in DiffsimObject.iter_objects():
        if hasattr(obj,"set_case_name"):
            obj.set_case_name(case_name)
    
    info_str = ""
    # Walk through the runs
    for n in xrange(num_runs):
        info("\nStarting run number: %d\n"%(n+1))
        tic(run_str%n)
        single_run()
        info_str = "* Run no: %d took: %s *"%(n+1,_format_time(toc(run_str%n)))
        info("*"*len(info_str))
        info(info_str)
    if num_runs > 1:
        info_str = "* All %d runs took: %s *"%(num_runs,_format_time(toc("All runs")))
        info("*"*len(info_str))
        info(info_str)
    info("*"*len(info_str))
    
    # Finalize the objects
    finalize_last = []
    data = {}
    for name, obj in DiffsimObject.iter_objects():
        # FIXME: Make more general
        if isinstance(obj, Plotter):
            finalize_last.append(obj)
        else:
            ret = obj.finalize()
            if ret is not None:
                data[obj.name()] = ret
    for obj in finalize_last:
        ret = obj.finalize()
        if not ret is None:
            data[obj.name()] = ret
    
    # If there were any collected data, save them to file
    if len(data) > 0 and _io_params.save_data:
        # Save a copy of the parameters in a ordinary dict
        data["params"] = get_params().copy(True)
        save(case_name, with_time = True,**data)

    debug("Done with running simulation")
    
    if len(data) > 0:
        return data


def single_run():
    "Do one run."
    init_system()
    while _time_handler.continue_sim():
        step_system()

# FIXME: Consider changing all the class names and function names to _foo

############### Protected variables, classes and methods ###########################
_system_initialized  = False
_object_dependencies = []
_indep_cont          = []
_dep_cont            = []

def discrete_variable_cmp(x,y):
    tx = x.next_event()
    ty = y.next_event()
    if  tx < ty : return -1
    if  tx > ty : return 1
    return 0

class TStop(DiscreteVariable,Singleton):
    def __init__(self):
        super(TStop,self).__init__("tstop")
        self._tstop = _time_params.tstop
        self._ext_vars['time'] = 0
        self._register_DiffsimObject(TStop)
        
    def reset(self):
        self._t = 0
        self._tstop = _time_params.tstop
        
    def update_time(self,dt):
        pass
    
    def step(self):
        " Call and end simulation"
        self.stop_sim()
        return True
        
    def next_event(self):
        return self._tstop - self._t

    def receive(self, sender):
        assert(sender.name() == 'time')
        self._t = sender.value()
            
    
class DtMax(DiscreteVariable,Singleton):
    def __init__(self):
        super(DtMax,self).__init__("dt max")
        self._dt_max   = _time_params.dt_max

        self._register_DiffsimObject(DtMax)
    
    def reset(self):
        self._dt_max   = _time_params.dt_max
        
    def update_time(self,dt):
        "A dummy function, to complete the interface."
        pass
    
    def step(self):
        "A dummy function, to complete the interface."
        return True
    
    def next_event(self):
        return self._dt_max

class TimeHandler(DiscreteVariable,Singleton):
    """ A class to take care of the time manegement in Diffsim
    
    I inherits from DiscreteVariable to get some nice features, but it will not
    act as one.
    
    """
    def __init__(self):
        super(TimeHandler, self).__init__("time")
        self._event_list = []
        self._all_discrete_objects = None
        self._t = 0.0
        self._steps = 0
        self._continue_sim = True
        self._run_number = 0
        self._data = []
        self._chunk = 100
        self._dt_taken = numpy.zeros(self._chunk)
        
        # Create the tstop and dt_expander
        TStop()
        DtMax()
        
        self._register_DiffsimObject(TimeHandler)
    
    def reset(self):
        " Initialize the list of all discrete objects"
        if self._steps != 0:
            self._data.append(dict(tstop = self._t,
                                   steps = self._steps,
                                   dt_taken = self._dt_taken[:self._steps].copy(),
                                   ))
        self._run_number += 1
        self._t = 0.0
        self._steps = 0
        self._continue_sim = True
        self._dt_taken = numpy.zeros(self._chunk)
        self._all_discrete_objects = DiscreteVariable.objects()
        
    def update_time(self,dt):
        " Updates the value of time"
        self._t += dt
    
    def value(self):
        " Returns the value of the time"
        return self._t

    def next_event(self):
        " Dummy function"
        return numpy.inf

    def step(self):
        " Dummy function that will never be called"
        return True
    
    def next_dt(self):
        "Returns the size of the next time step"
        # Get events that happens on the same time step
        keep_stepping = True
        while keep_stepping:
            event = min(self._all_discrete_objects)
            dt = event.next_event()
            keep_stepping = not event.step()
        
        # Append the event to the event_list.
        # These will be called by the deliver_events function.
        self._event_list.append(event)
        
        # Collect any events that happens at the same time step
        event_at_same_time = min(self._all_discrete_objects)
        while event_at_same_time.next_event() == dt and event_at_same_time != event:
            if event_at_same_time.step():
                self._event_list.append(event_at_same_time)
            event_at_same_time = min(self._all_discrete_objects)
        
        debug("%s caused the size of dt."%", ".join("'%s'"%event.name() for event in self._event_list))
        assert dt>=0.0, "The dt must be zero or positive"
        # Update all discrete objects to the new time step
        for do in self._all_discrete_objects:
            do.update_time(dt)
        
        if len(self._dt_taken) <= self._steps:
            self._dt_taken = numpy.concatenate((self._dt_taken,numpy.zeros(self._chunk)))
        self._dt_taken[self._steps] = dt
        self._steps += 1
        return dt

    def steps(self):
        " Return the number of steps taken" 
        return self._steps

    def run_number(self):
        " Return the run number"
        return self._run_number
    
    def initial_events(self):
        " Handles any events that is schedueled to t = 0"
        debug("All discrete obj: %s"%\
              (", ".join(str(s) for s in self._all_discrete_objects)))
        initial_event = min(self._all_discrete_objects)
        while initial_event.next_event() == 0:
            debug("Initial event: %s"%str(initial_event))
            if initial_event.step():
                # Deliver the event
                initial_event.send()
                
                # If the event triggers an minimization of dt
                if initial_event.minimize_dt():
                    for cont_object in ContinuousVariable.objects():
                        cont_object.minimize_dt()
            
    def deliver_events(self):
        " Call the discrete events that was collected during the time step."
        while self._event_list:
            event = self._event_list.pop()
            debug("Discrete event: %s"%event.name())
            # Send the event
            event.send()
            
            # If the event triggers an minimization of dt
            if event.minimize_dt():
                for cont_object in ContinuousVariable.objects():
                    cont_object.minimize_dt()
            
        # Deliver the time to all recievers at the end
        self.send()

    def continue_sim(self):
        " Return True if the simulation should continue."
        return self._continue_sim
    
    def receive_stop_signal(self,sender):
        " Called whenever a simulation should stop."
        debug("At t = %s, after %d steps, %s called stop simulation"%
              (self._t, self._steps, sender.name()))
        self._continue_sim = False
    
    def finalize(self):
        self._run_number = 0
        self._data.append(dict(tstop = self._t,
                               steps = self._steps,
                               dt_taken = self._dt_taken[:self._steps].copy(),
                               ))
        timedata = self._data
        self._data = []
        return timedata
        
# Instantiate a TimeHandler
_time_handler        = TimeHandler()

def step_system():
    " Step the systems one time step."
    global _time_handler, _indep_cont, _dep_cont
    
    tic("Resolving discrete events")
    dt = _time_handler.next_dt()
    t  = _time_handler.value()
    debug("Resolving next discrete events took: %2.2fs"%toc("Resolving discrete events"))
    info_str0 = "* Run number:  %2d of %2d     *"%(_time_handler.run_number(),
                                              _time_params.num_runs)
    info_str1 = "* Num steps: %4d           *"%_time_handler.steps()
    info_str2 = "* t: %5.3f, dt: %4.9f *"%(t,dt)
    info("*"*max(len(info_str0),len(info_str1),len(info_str2)))
    info(info_str0)
    info(info_str1)
    info(info_str2)
    info("*"*max(len(info_str0),len(info_str1),len(info_str2)))
    
    cont_objects = ContinuousVariable.objects()
    if dt > 0. and len(cont_objects) > 0:
        solver_params = cont_objects[0].get_solver_params()

        # If we have more than one continuous variables and using a nonlinear solver
        if len(cont_objects) > 1 and solver_params.nonlinear:
            # we first solve any independent variables
            def add_tmp_nonlinear(obj,tmp_nonlinear,_indep_cont):
                " Recursevily add dependent objects"
                if obj not in tmp_nonlinear:
                    tmp_nonlinear.append(obj)
                    
                for cont_obj in indep_cont:
                    if obj.name() in cont_obj.demands():
                        indep_cont.remove(cont_obj)
                        add_tmp_nonlinear(cont_obj,tmp_nonlinear,indep_cont)
                
            # Collect temporary nonlinears
            tmp_nonlinear = []
            indep_cont_copy = _indep_cont[:]
            for cont_obj in _indep_cont:
                if cont_obj.nonlinear():
                    add_tmp_nonlinear(cont_obj,tmp_nonlinear,indep_cont_copy)
                    
            for cont_obj in _indep_cont:
                # TODO:implement parameter passing, maybee to the constructor
                # of continuous systems? WHY this comment here??!?!?
                #cont_obj = DiffsimObject.get_object(name)
                if cont_obj in tmp_nonlinear:
                    continue
                ContinuousVariable.solve(cont_obj,dt)
            #print "Tmp: nonlinear:",tmp_nonlinear
            #print "dependent:", _dep_cont
            #print "independent:", _indep_cont
            # Then we solve the dependent one
            iteration = 0
            for relax in solver_params.relax_list:
                info("\nTrying to solve nonlinear problem with relax %s"%str(relax))
                while not converged(_dep_cont + tmp_nonlinear)\
                    and iteration < solver_params.max_fix_point_iter:
                    debug("\nNonlinear solve: iteration %d"%iteration)
                    # Do the fixpoint iterations
                    for cont_obj in _dep_cont + tmp_nonlinear:
                        # TODO:implement parameter passing, maybee to the constructor
                        # of continuous systems?
                        ContinuousVariable.solve(cont_obj,dt,iteration,relax)
                    iteration += 1
                
                # If we hit the max iteration limit
                if iteration == solver_params.max_iter:
                    iteration = 0
                    # Initialize the continuous objects again
                    if relax == solver_params.relax_list[-1]:
                        info("We did not manage to solve nonlinear problem. We try to continue.")
                    else:
                        for cont_obj in _dep_cont + tmp_nonlinear:
                            cont_obj.reset_after_non_convergence()
                        
                else:
                    info('Nonlinear problem solved in %d iterations'%iteration)
                    for cont_obj in _dep_cont + tmp_nonlinear:
                        cont_obj.reset_converged()
                    
                    break
            info("Trying to solve nonlinear problem with relax %s"%str(relax))

        # Solve all continues variables as if they were independent
        elif len(cont_objects) > 1:
            for cont_obj in _indep_cont + _dep_cont:
                ContinuousVariable.solve(cont_obj,dt)
                
        else:
            cont_obj = cont_objects[0]
            
            # Solve nonlinear problem
            if cont_obj.nonlinear() and solver_params.nonlinear:
                iteration = 0
                for relax in solver_params.relax_list:
                    info("Trying to solve nonlinear problem with relax %s"%str(relax))
                    while not converged([cont_obj]) and iteration < solver_params.max_iter:
                        debug("Nonlinear solve: iteration %d"%iteration)
                        ContinuousVariable.solve(cont_obj,dt,iteration,relax)
                        iteration += 1
                    # If we hit the max iteration limit
                    if iteration == solver_params.max_iter:
                        iteration = 0
                        if relax == solver_params.relax_list[-1]:
                            info("We did not manage to solve nonlinear problem. We try to continue.")
                        else:
                            cont_obj.reset_after_non_convergence()
                    else:
                        info('Nonlinear problem solved in %d iterations'%iteration)
                        cont_obj.reset_converged()
                        break
                        
            else:
                ContinuousVariable.solve(cont_obj,dt)
            
    
    # Deliver any collected events
    _time_handler.deliver_events()

def converged(cont_objs):
    " Return True if the dependent continuous variables has all converged"
    cont_obj_iterator = cont_objs.__iter__()
    try:
        # if the var_iterator reaches the last variable it throws a
        # StopIteration exception
        while cont_obj_iterator.next().converged():pass
        return False
    except:
        return True

def connect_DiffsimObjects():
    """ Connects all the DiffsimObjects.

    This function iterate over all DiffsimObjects and connects them
    depending on their demands and the name of the object
    
    Raise a RuntimeWarning if there are instances that still expect values
    from others.
    """
    global _time_handler
    info("Establishing connections")
    
    error = []
    Signal.disconnect_all_signals(_time_handler)
    
    for rec_name, receiver in DiffsimObject.iter_objects():
        # If the receiver does not need any external values, move along
        # Disconnect any signal attached to the receivers first.
        Signal.disconnect_all_signals(receiver)
        
        # Iterate over the demanded names
        for sender_name in receiver.demands():
            sender = DiffsimObject.get_object(sender_name)
            if sender is None:
                # If the demanded name is not registered
                error.append(sender_name)
                continue
            info("Connecting %s with %s"%(sender.name(),rec_name))
            sender.connect(receiver)
    
    # Connect each objects stop_sim_signal to the _time_handler
    for name, sender in DiffsimObject.iter_objects():
        if not isinstance(sender,TimeHandler):
            sender.connect(_time_handler, stop_signal = True)
        
    if error:
        frase = "These DiffsimObjects are" if len(error) == 1 \
                else "This DiffsimObject is"
        raise RuntimeError, """ Error while connecting.
        %s not registered:
        %s"""%(frase,", ".join(error))


def object_dependencies():
    info("Collecting object dependencies")
    dependencies = []
    for name, obj in DiffsimObject.iter_objects():
        deps = obj.demands()[:]
        dependencies.append((name,deps))
    return dependencies

def continuous_dependencies():
    info("Collecting continuous variable dependencies")
    independent = []
    dependent   = []
    # Collect all continuous dependencies
    cont_dep = []
    for obj in ContinuousVariable.objects():
        deps = []
        for dep in obj.demands():
            dep_obj = DiffsimObject.get_object(dep)
            if not dep_obj is None and isinstance(dep_obj,ContinuousVariable):
                # Only collect dependencies of other ContinuousVariables
                deps.append(dep)
        cont_dep.append((obj.name(),deps[:]))
        
    checked = []
    # Build the runtime dependency dicts
    while cont_dep:
        name, depends = cont_dep.pop(0)
        # If no continuous depends
        if len(depends) == 0:
            obj = DiffsimObject.get_object(name)
            # If the continuous object is nonlinear
            if obj.nonlinear():
                cont_dep.append((name,depends))
                checked.append(name)
                continue
            independent.append(obj)
            info("Adding independent continuous variable %s"%name)
            # Remove the dependency of this var from the remaining continuous vars
            for n, deps in cont_dep:
                if name in deps:
                    deps.remove(name)
        else:
            cont_dep.append((name,depends))
            if name in checked:
                # We have allready checked the variable once and circular dependency
                # is detected
                while cont_dep:
                    # Assume that the remaining variables are dependent of each others
                    name, depends = cont_dep.pop(0)
                    dependent.append(DiffsimObject.get_object(name))
                    info("Adding dependent continuous variable %s"%name)
            else:
                checked.append(name)
    
    return independent, dependent

def reset_system(object_dependencies):
    """ Resets the system.
    
    First non dependent instances are reseted.
    """
    import copy
    info("Reseting the system:")
    #self._time_handler.reset()
    obj_dep = copy.deepcopy(object_dependencies)
    reseted = []
    postponed = []
    stochastic_time_stepper = None
    diffusive_ligand_dt = []
    while obj_dep:
        name, depends = obj_dep.pop(0)

        # Get the object
        obj = DiffsimObject.get_object(name)
        assert not obj is None, "DiffsimObject %s is not initialized correctly."%name
        
        if name == _time_handler.name():
            for n, deps in obj_dep:
                if name in deps:
                    deps.remove(name)
            continue
        
        # If the obj is a MarkovChain or a MarkovModel, ignore reseting as these
        # are resetted by the StochasticTimeStepper instance
        if isinstance(obj,(DiscreteMarkovModel, DiscreteMarkovChain)):
            continue
        
        # If the obj is StochasticTimeStepper postpone its reset
        if type(obj) == StochasticTimeStepper:
            stochastic_time_stepper = obj
            continue

        # Postpone the reseting of DiffusiveLigandDt as these are dependent on all
        # other variables beeing reseted
        if isinstance(obj, DiffusiveLigandDt):
            diffusive_ligand_dt.append(obj)
            continue
            
        if len(depends) == 0:
            # If no other dependencies reset instance and send the value
            
            debug("Reseting %s"%name)
            if not obj is None:
                # Reset and send value
                obj.reset()
                obj.send()
            # Remove any occurence of this
            # dependency from the remaining instances.
            for n, deps in obj_dep:
                if name in deps:
                    deps.remove(name)
        else:
            if name in postponed:
                # We have already checked the variable once
                # FIXME: Isn't that expected?
                # We have circular dependency, but have to initialize one or another!
                #raise RuntimeError, "Detected circular dependency of instances."
                debug("############################################")
                debug("Circular!: %s"%name)
                debug("Reseting %s"%name)
                if not obj is None:
                    # Reset and send value
                    obj.reset()
                    obj.send()
                # Remove any occurence of this
                # dependency from the remaining instances.
                for n, deps in obj_dep:
                    if name in deps:
                        deps.remove(name)
            else:    
                # If the instance depends on values from other instances then
                # put it back
                obj_dep.append((name,depends))
                # Register that we have allready checked the variable
                postponed.append(name)

    # Reset and broadcast the time
    debug("Reseting %s"%_time_handler.name())
    _time_handler.reset()
    _time_handler.send()

    # If a StochasticTimeStepper is registered reset, it and do not broadcast
    # its values. # FIXME: Why?!??!
    if not stochastic_time_stepper is None:
        debug("Reseting %s"%stochastic_time_stepper.name())
        stochastic_time_stepper.reset()

    # Reset the diffusive_ligand_dt. Need to be last as they need all other
    # variables to be updated
    for dt in diffusive_ligand_dt:
        debug("Reseting %s"%dt.name())
        dt.reset()

    

    
debug("%s loaded"%__name__)
