__author__ = "Johan Hake (johan.hake@gmail.com)"
__copyright__ = "Copyright (C) 2008 Johan Hake"
__license__  = "GNU LGPL Version 3.0 or any later"
__date__ = "2008-08-19 -- 2010-01-28"

# Defining the import * statement
__all__ = ['DiffsimObject',
           'ContinuousVariable',
           'DiscreteVariable',
           'Q10_scale'
           ]

# Importing DiffSim stuff
from diffsim.signal        import Signal
from diffsim.parameterdict import ParameterDict, TypeCheck, RangeCheck, OptionCheck
from diffsim.common        import T_ABS, ALLOWED_IONS

# Initialize the logging system
from diffsim.logger import *
init_local_logger(__name__,globals(),INFO)

# Importing numpy
import numpy

# Import dolfin
import dolfin

def Q10_scale(Q10,T_ref):
    " Return the Q10 scaling factor given the T_ref and the Q10 value"
    T_ref = check_temp(T_ref)
    return Q10**((DiffsimObject.get_T()-T_ref)/10)

def check_temp(T):
    T_min = 10
    T_max = 50
    if T < 100:
        T += T_ABS
    assert(T_ABS + T_min< T and T < T_ABS +T_max), "%s \xe2\x88\x89[%s, %s] \xc2\xb0K"%(str(T),str(T_ABS+T_min),str(T_ABS+T_max))
    return T

# Some protected help functions
def check_ion_name(ion_name):
    " Check the validity of an ion name"
    if not isinstance(ion_name,str):
        raise TypeError, "Please provide a string for the 'ion_name'"
    if not ion_name in ALLOWED_IONS:
        raise TypeError,\
              "Please provide an ion from one of: %s"%(", ".join(ALLOWED_IONS))

def check_domain_name(domain_name):
    """ Check the validity of a domain name

    Return the domain object if ok, otherwise raise Error
    
    """
    if not isinstance(domain_name,str):
        raise TypeError, "Please provide a string for the 'domain_name'"
        
    domain = DiffsimObject.get_object(domain_name)
    if domain is None or not isinstance(domain,DiffusionDomain):
        raise AttributeError, "No Domain with name %s is registered"%domain_name
    
    return domain

def _check_ligand_argument(obj,ligand_name):
    """ A helper function to check a ligand argument

    @param   obj           : The instantiated object
    @param  ligand_name   : The unique name of the ligand
    @return               : The ligand instance corresponding to the ligand_name
    @raise  TypeError     : If the provided ligand_name is not a string
    @raise  AttributError : If no ligand with ligand_name is registered
    """
    
    # Check the type of ligand_name
    if not isinstance(ligand_name,str):
        raise TypeError, "Please provide a string for the ligand_name"

    # Get the DiffsimObject
    ligand = obj.get_object(ligand_name)
    if ligand is None or not isinstance(ligand,DiffusiveLigand):
        raise AttributeError, "No Ligand with name %s is registered"%ligand_name
    
    return ligand

# FIXME: The use of super in the __init__ functions is not safe, as I am changeing the
#        arguement list. When this becomes an issue I will try fix it.
#        Further reading:
#           <http://wiki.osafoundation.org/bin/view/Chandler/UsingSuper>
#        and
#           <http://fuhm.net/super-harmful>

class Singleton(object):
     """ A Pythonic Singleton """
     def __new__(cls, *args, **kwargs):
         if '_inst' not in vars(cls):
             cls._inst = object.__new__(cls)
         return cls._inst
     
     def clear_singleton_instant(self):
         del type(self)._inst

class DiffsimObject(object):
    " Base class for all variables used in diffsim"

    # A dict with all the instantiated DiffsimObjects with their names as keys
    __all_DiffsimObjects = {}

    __T = 37.+T_ABS
    # All derived classes from DiffsimObject needs to implement __init__. The constructor
    # of the base class needs to be called first in the derived constructor and lastthe
    # function _register_DiffsimObject needs to be called with the derived class as the 
    # only arguement.
    def __init__(self, name = None):
        """ Instantiate the DiffsimObject
        
        This function needs to be called by all derived classes.
        
        @param name : A string defining the unique name of the DiffsimObject.
        """
        # FIXME: We must call the objects.__init__ in case of multiple inheritances
        #        We are doing this in an unsafe way, see the note above.
        super(DiffsimObject,self).__init__()
        
        if not isinstance(name,str):
            raise TypeError, "Please provide a string as a name"
        
        # The signal instant. Used to send the DiffsimObject
        self._signal = Signal()
        
        # Stop simulation signal.
        self._stop_sim_signal = Signal()

        # The unique identifier
        self._name     = name

        # Initialize the parameters if the class has the class attribute _params
        # and it is equal to 0, then call the classmethod to fill it
        if hasattr(type(self),'_params'):
            if type(self)._params == 0:
                type(self)._params = type(self).default_params()
        else: 
            self._params = self.default_params()
        
        # A dict of external variables the keys hold the unique name of the variable
        # and the value holds the value(s) ;)
        self._ext_vars = {}
        
        # A list which should be populated with external variables, that is changed
        # during the previous time step        
        self._ext_vars_changed = []

        # Store any previous values that the DiffsimObject want to keep
        # Used by the solver for continuous variables
        self._previous_values = {}

        # Init hash cache
        self._hash = None
        
        # Register the DiffsimObject if the instantiated object is a DiffsimObject!
        self._register_DiffsimObject(DiffsimObject)
        
    def __hash__(self):
        if self._hash is None:
            self._hash = hash(self.name())
        return self._hash
    
    def __eq__(self,other):
        return self.name() == other.name()

    @staticmethod
    def get_T():
        return DiffsimObject.__T
    
    @staticmethod
    def set_T(T):
        "Set the temperature" 
        DiffsimObject.__T = check_temp(T)

    def reset(self):
        " Reset the DiffsimObject before a simulation"
        pass
    
    def demands(self):
        " Return a list of variables that the DiffsimObject demand"
        return self._ext_vars.keys()
    
    def finalize(self):
        " Finalize the DiffsimObject after a set of simulations is done"
        pass

    def send(self):
        " Send it self to any receivers" 
        self._signal(self)

    def previous_values(self):
        " Return a dict with stored previous values"
        return self._previous_values

    def stop_sim(self):
        " Send the stop sim signal"
        self._stop_sim_signal(self)
        
    def connect(self, receiver, stop_signal = False):
        " Connect this DiffsimObject to a receiving DiffsimObject"
        if not issubclass(type(receiver), DiffsimObject):
            raise TypeError, "Can only connect instances of classes inheritateing DiffsimObject"
        if stop_signal:
            assert(hasattr(receiver,"receive_stop_signal"))
            self._stop_sim_signal.connect(receiver.receive_stop_signal)
        else:
            self._signal.connect(receiver.receive)
    
    def name(self):
        " Return the name of the DiffsimObject"
        return self._name
    
    def receive(self, sender):
        " Called when a signal is delivered to the DiffsimObject"
        pass

    @classmethod
    def default_params(cls):
        " Returns the default parameters" 
        return ParameterDict()
    
    def update_params(self,params):
        """ Update the parameters of the DiffsimObject
        
        @param params : A dict that is used to update the parameters
        """
        self._params.update(params)
    
    def get_params(self):
        " Return the parameter dict"
        return self._params
    
    def _register_DiffsimObject(self, class_type):
        """ Register a created DiffsimObject to a dict with its name as key
        
        This function needs to be called by the __init__ function in each
        subclass of DiffsimObject. It have to be called by the last subclass
        
        @param class : If the object is of type class then it will be registered
        """
        
        assert(isinstance(self,DiffsimObject)),"Can only register an instant of DiffsimObject or a subclass to DiffsimObject"
        # If the handed self is of the same type as class
        if type(self) != class_type:
            #debug("Not registering %s"%self.name())
            return 
        # Check if a DiffsimObject with the same name is already instantiated
        if not self.get_object(self.name()) is None:
            # If trying to register my self twice, and me being a singleton, return
            if self.get_object(self.name()) is self and isinstance(self,Singleton):
                return
            raise TypeError, "Only DiffsimObjects with unique names are allowed. A DiffsimObject with name '%s' already excists."%self.name()
        DiffsimObject.__all_DiffsimObjects[self.name()] = self
        debug("DiffsimObject '%s' is now created"%self.name())
        
        # Register the object dependent on which class it is derived from
        if isinstance(self,DiscreteVariable):
            self._register_DiscreteVariable()
            
        elif isinstance(self,ContinuousVariable):
            self._register_ContinuousVariable()
            
        elif isinstance(self,DiscreteMarkovModel):
            self._register_DiscreteMarkovModel()
    
    @classmethod
    def get_object(cls,name,default=None):
        """ Return the instance corresponding to name
        
        @param name    : The name of the variable
        @param default : If not a variable with name is registered then return default
        """
        assert(isinstance(name,str))
        return cls.__all_DiffsimObjects.get(name,default)
    
    @classmethod
    def clear_all(cls):
        " Clear all registered DiffsimObjects."

        # Disconnect all signals()
        for receiver in cls.iter_objects():
            Signal.disconnect_all_signals(receiver)
        
        # Clear all subtypes of DiffsimObjects
        DiscreteVariable.clear_all()
        ContinuousVariable.clear_all()
        DiscreteMarkovModel.clear_all()
        
        # Clear the actuall diffsim objects
        cls.__all_DiffsimObjects = {}
        
    @classmethod
    def remove_object(cls,name):
        " Remove a diffsim object from storage"
        cls.__all_DiffsimObjects.pop(name,None)
        # FIXME: Implement to remove from the three different variables too.
        
    @classmethod
    def iter_objects(cls):
        " Return an iterator over all registered DiffsimObjects" 
        return cls.__all_DiffsimObjects.iteritems()
    
    @classmethod
    def objects(cls):
        " Return a shallow copy of all registered DiffsimObjects" 
        return cls.__all_DiffsimObjects.values()

    def rename(self,new_name):
        " Rename the DiffsimObject"
        old_name = self._name
        if not self.get_object(new_name) is None:
            raise TypeError, "Only DiffsimObjects with unique names are allowed. A DiffsimObject with name '%s' already excists."%new_name
        self.remove_object(old_name)
        self._name = new_name

        # Clear hash
        self._hash = None
        self._register_DiffsimObject(type(self))
    
    def __str__(self):
        return "'%s' object with name '%s'"%(type(self).__name__,self._name)

    def __repr__(self):
        return str(self)

def default_solver_params():
    return ParameterDict(sym_solver     = OptionCheck('cg',['cg']),
                         asym_solver    = OptionCheck('bicgstab',['bicgstab','gmres']),
                         prec           = OptionCheck('amg_hypre',['jacobi', 'sor',
                                                                   'ilu','icc',
                                                                   'amg_hypre',
                                                                   'amg_ml']),
                         tol_rel        = RangeCheck(1.0e-7,gt=0,le=10),
                         tol_abs        = RangeCheck(1.0e-10,gt=0,le=10),
                         nl_tol_abs     = RangeCheck(1.0e-5,gt=0,le=100),
                         max_iter       = RangeCheck(50,ge=1,le=1e4),
                         crank_nich     = TypeCheck(False),
                         tr_bdf2        = TypeCheck(True),
                         num_previous_values = RangeCheck(5,gt=0,le=10),
                         monitor_convergence = TypeCheck(False),
                         relax_list     = [1.0,0.5,0.25,0.1],
                         newton         = TypeCheck(False),  # Not implemented
                         nonlinear      = TypeCheck(False),
                         max_fix_point_tries  = RangeCheck(3,gt=0,lt=10),
                         max_fix_point_iter = RangeCheck(50,gt=0,lt=1000),
                         )

class ContinuousVariable(DiffsimObject):
    """ Base class for ContinuousVariables
    
    This class implement a solver and a converged function that
    can be called in a fix point iteration loop, together
    with some protected attributes used by these functions.
    """
    
    __all_ContinuousVariables = []
    
    def __init__(self, name):
        """ Instantiate the ContinuousVariable

        This function needs to be called by all derived classes.
        
        @param name : A string defining the name of the ContinuousVariable. 
        """
        super(ContinuousVariable,self).__init__(name)
        
        # Protected attributes
        self._converged    = False
        self._fix_point_it = 0
        self._nonlinear    = False
        self._values_0     = None
        self._values_1     = None
        self._values_2     = None
        
        self._register_DiffsimObject(ContinuousVariable)
        
    # This is a class attribute so the solver parameters are the same for
    # all ContinuousVariables
    _solver_params = default_solver_params()
    
    default_solver_params = staticmethod(default_solver_params)

    @classmethod
    def get_solver_params(cls):
        " Return the solver parameter dict"
        return cls._solver_params

    
    @classmethod
    def update_solver_params(cls,params):
        """ Update the solver parameters
        
        @param params : A dict that is used to update the parameters
        """
        cls._solver_params.update(params)
    
    def minimize_dt(self):
        """ Minimize the next time step.

        If this functionality is relevant for a ContinuousVariable then
        it needs to be implemented in a derived class.
        """
        pass
    
    def nonlinear(self):
        return self._nonlinear

    def solve(self,dt,iteration=0, relax=1.0):
        """ A mother solve method


        This method is called by the app to reset, solve (call the sub class solve)
        and broadcast the values, of a ContinuousVariable """
        
        from numpy import ndarray, absolute
        from dolfin.cpp import GenericVector
        
        assert(self.solve != ContinuousVariable.solve),"You need to implement solve() in a sub class"

        if not isinstance(self._values_0,(int,float,ndarray,GenericVector)):
            raise TypeError, "Values can only be int, float or numpy arrays"
        
        # If we are inside a fix point or newton iteration loop
        if iteration >= 1:
            #debug("Making %s ready for another iteration"%self._name)
            # FIXME: check for ndarray,GenericVector before this!
            self._values_2 *= 1 - relax
            self._values_1 *= relax
            self._values_2[:] += self._values_1
            self._values_1 /= relax
        else:
            # Update any previous registered values.
            variables = self.active_variables()
            # Iterate over the numpy arrays in the dict
            for variable in variables:
                for values in variable.previous_values().itervalues():
                    assert(isinstance(values,(ndarray,GenericVector)))
                    # Check the length is larger than 1
                    # Move the values one step forward by slice assign
                    values[0:-1] = values[1:]
            
            # Update value differently dependent on what type values has
            if isinstance(self._values_0,(ndarray,GenericVector)):
                self._values_0[:]  = self._values_1
                self._values_2[:]  = self._values_1
            else:
                self._values_0  = self._values_1
                self._values_2  = self._values_1
        
        self._converged = False

        # Call the subclass implementation of solve()
        self.solve(dt,iteration)

        # Broadcast the value to other variables
        self.send()

    def reset_after_non_convergence(self):
        " Reset variable after a step of non_convergence"
        from numpy import ndarray, absolute
        from dolfin.cpp import GenericVector
        if isinstance(self._values_0,(ndarray,GenericVector)):
            self._values_1[:]  = self._values_0
            self._values_2[:]  = self._values_0
        elif isinstance(self._values_0,(float,int)):
            self._values_1  = self._values_0
            self._values_2  = self._values_0
        else:
            raise TypeError, "Values can only be int, float or numpy arrays"
        
        self._converged    = False
    
    def active_variables(self):
        """ Return any active variables attached to this ContinousVariable

        This could for example be any active flux or fixed boundaries
        """
        return []
        
    def reset_converged(self):
        self._converged = False
        
    def converged(self):
        " Return True if the solver has converged"
        return self._converged
    
    def value(self, m = None, stat = None, copy = False):
        """ Returns the value of the continuous variable

        @param m    : A marker defining a subdomain of the field that is then returned
        @param stat : Return a statistic instead of an array. Available are:
                      MIN, MAX, MEAN, STD, VAR
        @param copy : If True a copy of the value are provided.
                      Is ignored if a stat is given
        """
        raise NotImplementedError(type(self).value)
    
    def _register_ContinuousVariable(self):
        """ Register a ContinuousVariable

        There is no need to call this method as it will be called by
        _registerDiffsimObjects.
        """
        if not self in self.__all_ContinuousVariables:
            self.__all_ContinuousVariables.append(self)
    
    @classmethod
    def clear_all(cls):
        " Clear all registered ContinuousVariables."
        cls.__all_ContinuousVariables = []
    
    @classmethod
    def objects(cls):
        " Return a shallow copy of all registered ContinuousVariables" 
        return cls.__all_ContinuousVariables[:]
    

class DiscreteVariable(DiffsimObject):
    """ A base class for a discrete variables
    
    This class implements functions and protected attributes for DiscreteVariables
    
    """
    
    __all_DiscreteVariables = []
    
    def __init__(self,name):
        """ Instantiate the Discrete Variable
        
        This function needs to be called by all derived classes.
        
        @param name : A string defining the name of the DiscreteVariable
        """
        super(DiscreteVariable,self).__init__(name)
        
        self._register_DiffsimObject(DiscreteVariable)
        
    def step(self):
        """ Step the DiscreteVariable
        
        This is used to update the internals of the DiscreteVariable when
        it is this variable that is "next event" in line.
        
        @return real_step  : If the step was "real" then True is returned
        """
        raise NotImplementedError(type(self).step)
    
    def next_event(self):
        " Return the time to the next event for this DiscreteVariable"
        raise NotImplementedError(type(self).next_event)
    
    def update_time(self,dt):
        """ Update the time
        
        This function is used to update the internal time of the DiscreteVariable
        when another DiscreteVariables is the "next event" in line.

        @param dt : The time to next event
        """
        raise NotImplementedError(type(self).update_time)

    def minimize_dt(self):
        """ Return True if an event will cause a minimization of the next dt

        Note: This function is different to the one in ContinuousVariable, which
        if called actually minimizes the next time step for that variable
        """
        return False
    
    def value(self,*args,**kwargs):
        " Return the value of the DiscreteVariable"
        raise NotImplementedError(type(self).value)
    
    def __lt__(self,other):
        assert(isinstance(other,DiscreteVariable))
        return self.next_event() < other.next_event()
    
    def __gt__(self,other):
        assert(isinstance(other,DiscreteVariable))
        return self.next_event() > other.next_event()
    
    def __le__(self,other):
        assert(isinstance(other,DiscreteVariable))
        return self.next_event() <= other.next_event()
    
    def __ge__(self,other):
        assert(isinstance(other,DiscreteVariable))
        return self.next_event() >= other.next_event()
    
    def _register_DiscreteVariable(self):
        """ Register a DiscreteVariable

        There is no need to call this method as it will be called by
        _registerDiffsimObjects.
        """
        if not self in DiscreteVariable.__all_DiscreteVariables:
            debug("Register: %s"%str(self))
            DiscreteVariable.__all_DiscreteVariables.append(self)
    
    @classmethod
    def clear_all(cls):
        " Clear all registered DiscreteVariables."
        cls.__all_DiscreteVariables = []
    
    @classmethod
    def objects(cls):
        " Return a shallow copy of all registered DiscreteVariables" 
        return cls.__all_DiscreteVariables[:]

# Placed here to avoid circular dependency
from diffsim.diffusiveligand       import (DiffusiveLigand, MEAN, MIN, MAX, STD, VAR,
                                           ALLOWED_STAT)
from diffsim.diffusiondomain       import DiffusionDomain
from diffsim.stochastictimestepper import DiscreteMarkovModel
debug("%s loaded"%__name__)

