__author__ = "Johan Hake (johan.hake@gmail.com)"
__copyright__ = "Copyright (C) 2008 Johan Hake"
__license__  = "GNU LGPL Version 3.0 or any later"
__date__ = "2008-11-18 -- 2010-02-02"

# Defining the import * statement
__all__ = ['Flux',
           'FluxBoundary',
           'LinearFluxBoundary',
           'ConcDepLinearFluxBoundary',
           'FixedBoundary'
           ]

# Importing DiffSim stuff
from diffsim.signal        import Signal
from diffsim.parameterdict import ParameterDict, TypeCheck, RangeCheck, OptionCheck
from diffsim.base          import DiffsimObject, check_ion_name, check_domain_name

# Initialize the logging system
from logger import *
init_local_logger(__name__,globals(),INFO)

# Importing numpy
import numpy

    
class Flux(DiffsimObject):
    " Base class for Fluxes"

    def __init__(self, name, domain_name, ion_name):
        """ Instantiate a Flux object

        Implements the interface needed for a Flux
        
        @param name        : The unique name of the Flux
        @param domain_name : The name of the diffusion domain the flux is defined on
        @param ion_name    : The name of the ion the flux is defined for
        """
        super(Flux,self).__init__(name)
        check_ion_name(ion_name)
        domain = check_domain_name(domain_name)

        self._nonlinear = False
        self._ligand    = domain.get_ligand(ion_name)
        self._symetric  = True
        self._register_DiffsimObject(Flux)
    
    # This base class implements an empty add_to_flux_matrix function
    def add_to_flux_matrix(self,A):
        """ Add itself to a flux matrix
        
        Add the contribution from this flux variable to the handed flux matrix.
        
        @param A : The flux matrix."""
        pass
    
    # This base class implements an empty add_to_source_vector function
    def add_to_source_vector(self,source):
        """ Add itself to a source vector
        
        Add the contribution from this flux variable to the handed source vector.
        
        @param b : The source vector."""
        pass

    def nonlinear(self):
        return self._nonlinear
    
    def activate(self, minimize=True):
        " Activate the Flux"
        self._ligand.activate(self, minimize)
    
    def deactivate(self, minimize=True):
        " Deactivate the Flux"
        self._ligand.deactivate(self, minimize)
    
    def is_symetric(self):
        " Return True if the flux matrix is symetric."
        return self._symetric

class FluxBoundary(Flux):
    """ A base class for flux boundaries
    
    The class implements an interface so it can be be activated by an external
    activator, through the send and receive function.
    """
    
    def __init__(self, name, domain_name, ion_name, marker, activator = None):
        """ Instantiate a FluxBoundary
        
        @param name        : The name of the FluxBoundary
        @param domain_name : The name of the diffusion domain the flux is defined on
        @param ion_name    : The name of the ion the flux is defined for
        @param marker      : An int that defines the subdomain of the boundary
        @param activator   : An class or an object that can be used to connect
                             the FluxBoundary.
                             If no activator is provided the Boundary is assumed
                             activated by default.
        """
        
        # Check the type of marker
        if not isinstance(marker,int):
            raise TypeError, "Please provide an int for the marker"
        
        marker_sufix = "_m%d"%marker
        # Check that marker_sufix is the last a part of the name otherwise make it
        if not(marker_sufix in name and\
               name[name.rfind(marker_sufix):] == marker_sufix):
            old_name = name
            name = "%s%s"%(name,marker_sufix)
            info("Renamed instant of %s from %s to %s"%
                 (type(self).__name__,old_name,name))
        
        # Initiate the super classes
        super(FluxBoundary,self).__init__(name, domain_name, ion_name)
        
        # If activator is provided
        if not activator is None:
            if isinstance(activator,type) and issubclass(activator,DiffsimObject):
                # If the activator is a subclass of DiffsimObject we assume
                # that an instant of this class will be used as an activator
                # and we try to instantiate an object with name and marker
                a = activator("%s_activator"%name,marker)
#                try:
#                    
#                except NameError, e:
#                    raise e
#                except TypeError:
#                    raise TypeError, """Please provide an activator type that is
#        instantiated by a string and an int as arguments"""
                activator = a
            elif isinstance(activator,DiffsimObject):
                # else if the activator is a DiffsimObject assume that hte object
                # should be used to activate and deactivate the FluxBoundary
                aname = activator.name()
                if not(marker_sufix in aname and\
                       aname[aname.rfind(marker_sufix):] == marker_sufix):
                    # If the name provided to activate the FluxBoundary does not
                    # include the marker sufix try to rename the object
                    if hasattr(activator,"rename"):
                        activator.rename("%s_activator"%name)
                    else:
                        raise AttributeError, "The activator does not have %s in the name, and it does not implement a rename function."%marker_sufix
            else:
                raise TypeError, \
            "The 'activator' must be either an instant or a subclass of DiffsimObject"
            self._activator = activator
            self._ext_vars[activator.name()] = activator.value()
        else:
            self._activator = None
            self.activate(self.minimize_from_start())

        # Take care of the previous activator value to facilitate non minimization of
        # dt during start.
        self._prev_activator_value = 0
        
        # Check the excistence of the markers on the domain
        self._ligand.domain().check_face_marker(marker)
        
        # Protected attributes
        self._marker      = marker

        # Register the boundary to the ligand
        self._ligand.register_boundary(self)
        
        # Register the instantiated object
        self._register_DiffsimObject(FluxBoundary)

    def minimize_from_start(self):
        return False

    # This base class implements an empty _update_changed_ext_val function
    def _update_changed_ext_val(self):
        """ Update internal variables
    
        This function implements internal changes to the FluxBoundary
        when there are changes in external variables
        """
        pass

    def marker(self):
        " Return the value of the marker"
        return self._marker
    
    def receive(self,sender):
        name = sender.name()
        debug("%s received %s"%(self.name(),sender.name()))
        if isinstance(sender, DiffusiveLigand):
            self._ext_vars[name] = sender.value(self._marker,
                                                 self._params.ligand_stat)
        else:
            self._ext_vars[name] = sender.value()
        
        # If the sender is the activator
        if sender is self._activator:
            # If the sent value is the same as already registered then exit
            if sender.value() == self._prev_activator_value:
                return
            self._prev_activator_value = sender.value()
            
            if sender.value() == 0.0:
                debug("Deactivating %s"%self.name())
                self.deactivate()
            else:
                debug("Activating %s"%self.name())
                self.activate()
        else:
            self._ext_vars_changed.append(name)
    
    @classmethod
    def default_params(cls):
        return ParameterDict(ligand_stat = OptionCheck(MEAN,list(ALLOWED_STAT)))

    # Set the docstrings from base classes 
    receive.__doc__              = DiffsimObject.receive.__doc__
    #default_params.__doc__       = DiffsimObject.default_params.__doc__

class LinearFluxBoundary(FluxBoundary):
    """ A base class for linear flux boundaries on the form
    
             du
       (-?)D*-- = alpha * u - c0
             dn

       A LinearFluxBoundary is defined on a subdomain on the boundary
    """
    
    def __init__(self, name, domain_name, ion_name, marker, alpha = 0.0, c0 = 0.0,
                 activator = None, standard_mass=True):
        """ Instantiate a LinearFluxBoundary
        
        param @name :
           The name of the LinearFluxBoundary
        param @domain_name :
           The name of the diffusion domain the flux is defined on
        param @ion_name :
           The name of the ion the flux is defined for
        param @marker :
           An int that defines the subdomain of the boundary
        param @alpha :
           A default value of alpha also setable through the parmeter system, float
        param @c0 :
           A default value of c0 also setable through the parmeter system. Can be a
           float or a numpy array. If the last is true it need to be as long as
           the number of vertices on the boundary
        param @activator :
           A class or an object that can be used to connect the LinearFluxBoundary.
           If no activator is provided the Boundary is assumed activated by default.
        @param standard_mass :
           If True use standard mass boundary matrix
        """
        
        # Initiate the super classes
        super(LinearFluxBoundary,self).__init__(name, domain_name, ion_name,
                                                marker, activator)
        
        # The domain
        domain = self._ligand.domain()
        
        # Protected attributes
        if standard_mass:
            # If using electric potential weighted value use standard mass matrix
            # FIXME: Always use standard mass matrix?
            self._MB       = self._ligand.domain().tensor("mass boundary", marker)
        else:
            # Potentially use epot_weighted mass matrix
            self._MB       = self._ligand.tensor("mass boundary", marker)
        
        self._area         = domain.domain_area(marker)

        self._dof_inds     = domain.get_vertex_indices(marker)
        self._dof_vec      = domain.get_vertex_vector(marker)
        self._params.alpha, self._params.c0 = self._check_and_set_alpha_c0(alpha,c0)

        self._source_base = (self._MB*self._dof_vec)[self._dof_inds]
            
        self._source_add  = numpy.zeros(len(self._dof_inds))

        # Register the instantiated object
        self._register_DiffsimObject(LinearFluxBoundary)
    
    def reset(self):
        self._check_and_set_alpha_c0(self._params.alpha,self._params.c0)
        self._update_c0()
        
    def add_to_flux_matrix(self,A):
        if self._ext_vars_changed:
            self._update_changed_ext_val()
            self._ext_vars_changed = []
        if self._alpha == 0.:
            return
        # Assume the A implement the in place addition function A.axpy
        A.axpy(self._alpha, self._MB, True)
    
    def add_to_source_vector(self,source):
        if self._ext_vars_changed:
            self._update_changed_ext_val()
            self._ext_vars_changed = []
        if isinstance(self._c0, float) and self._c0 == 0.:
            return
        source[self._dof_inds] += self._source_add

    @classmethod
    def default_params(cls):
        p = ParameterDict(alpha = TypeCheck(0.0),
                          c0 = 0.0,
                          **super(LinearFluxBoundary,cls).default_params())
        return p

    def _check_and_set_alpha_c0(self,alpha,c0):
        " Init the FluxBoundary with the handed c0 and alpha arguments"
        # Check alpha argument
        if not isinstance(alpha,(float,int)):
            raise TypeError, "Provide a scalar for 'alpha'"
        
        # Check c0 argument
        if not isinstance(c0,(float,int, numpy.ndarray,list,tuple)):
            raise TypeError, "Provide a 'scalar' or 'sequence' for c0"
        if isinstance(c0,(numpy.ndarray,list,tuple)):
            if len(c0) != len(self._dof_inds):
                raise ValueError, "The provided sequence for 'c0' must be of the same length as the number of dofs at the boundary: %d"%len(self._dof_inds)
            c0 = numpy.asarray(c0,'d')
        
        # Set the protected attributes
        self._alpha = float(alpha)
        self._c0    = c0
        
        # Return the checked arguments
        return self._alpha, self._c0
    
    def _update_c0(self):
        if isinstance(self._c0, float):
            self._source_add[:] = self._source_base
            self._source_add   *= self._c0
        else:
            # This is a generalised version of the one just above
            u0_vec = self._dof_vec.copy()
            u0_vec[self._dof_inds] *= self._c0
            self._source_add[:] = (self._MB*u0_vec)[self._dof_inds]
            
    # Set the docstrings from base classes 
    add_to_flux_matrix.__doc__   = Flux.add_to_flux_matrix.__doc__
    add_to_source_vector.__doc__ = Flux.add_to_source_vector.__doc__
    #default_params.__doc__       = DiffsimObject.default_params.__doc__

class ConcDepLinearFluxBoundary(LinearFluxBoundary):
    """
    The Concentration dependent flux is given on the following form.
    
        D (u-u0)/dn = alpha*[c] - c0

    where the rhs is the one that is used by the LinearFluxBoundary and
    the lhs is the one specified by the user, using this class.
    """
    def __init__(self, name, domain_name, ion_name, marker, u0=0.0, dn=1.0,
                 activator=None, standard_mass=False):
        """ Instantiate a Concentration dependent flux

        Implements the interface needed for a Flux

        @param name :
           The unique name of the Flux
        @param domain_name :
           The name of the diffusion domain the flux is defined on
        @param ion_name :
           The name of the ion the flux is defined for
        @param u0 :
           A concentration 'outside' the domain. Type, 'float' or numpy array.
        @param dn :
           The distance of the concentration gradient
        @param standard_mass :
           If True use standard mass boundary matrix
        """
        # Initiate the super classes
        super(ConcDepLinearFluxBoundary, self).\
            __init__(name, domain_name, ion_name, marker, activator = activator,
                     standard_mass=standard_mass)

        if not isinstance(u0,(float,int,tuple,list,numpy.ndarray)):
            raise TypeError, "Provide a 'scalar' or 'sequence' as 'u0'"
        
        if isinstance(u0,(tuple,list,numpy.ndarray)):
            u0 = numpy.asarray(u0,'d')
            if (u0<0).any():
                raise ValueError, "Provide only positive values for 'u0'"
        else:
            if u0<0:
                raise ValueError, "Provide a positive value for 'u0'"
        
        self._check_and_set_alpha_c0(*self.alpha_c0_from_u0_dn(u0,dn))
        self._params.u0 = u0
        self._params.dn = dn
        
        # Register the instantiated object
        self._register_DiffsimObject(ConcDepLinearFluxBoundary)
    
    def reset(self):
        self._check_and_set_alpha_c0(*self.alpha_c0_from_u0_dn(self._params.u0,
                                                              self._params.dn))
        self._update_c0()
        
    def alpha_c0_from_u0_dn(self, u0, dn):
        alpha = self._ligand.D() / dn
        return alpha, u0*alpha
    
    @classmethod
    def default_params(cls):
        p = ParameterDict(dn = RangeCheck(1.0,gt=0.0),
                          u0 = 0.0,
                          **super(ConcDepLinearFluxBoundary,cls).default_params())
        return p
    
    #default_params.__doc__       = DiffsimObject.default_params.__doc__
    
class FixedBoundary(DiffsimObject):
    """ A class for Fixed Boundaries
    
    This class implements an interface to apply dirichlet boundary conditions
    to a diffusive ligand.
    """
    
    def __init__(self, name, domain_name, ion_name, marker, value = 0.0):
        """ Instantiate a FixedBoundary
        
        param @name        : The unique name of the FixedBoundary
        param @ligand_name : The name of the ligand the flux is defined on
        param @marker      : An int that defines the subdomain of the boundary
        param @value       : The value which the dofs will take, also setable
                             through the parmeter system
        """
        
        # Initiate the super classes
        super(FixedBoundary,self).__init__(name)
        
        check_ion_name(ion_name)
        domain = check_domain_name(domain_name)
        
        # Check the type of marker
        if not isinstance(marker,int):
            raise TypeError, "Please provide an int for the marker"
        
        # Check the excistence of the markers on the domain
        domain.check_vertex_marker(marker)
        
        # Protected attributes
        self._ligand   = domain.get_ligand(ion_name)
        self._dof_inds = domain.get_vertex_indices(marker)
        self._marker   = marker
        self._params.value = value
        
        # Register the boundary to the ligand
        self._ligand.register_boundary(self)
        
        # Register the instantiated object
        self._register_DiffsimObject(FixedBoundary)

    def apply(self, dof_inds, values):
        """ Apply the FixedBoundary to two provided arrays
        
        @param dof_inds : A numpy array that will be extended with the
                          dof_inds of this boundary
        @param values   : A numpy array that contains the values of the dofs
        """
        from numpy import ndarray, append
        assert isinstance(values,ndarray)
        assert isinstance(dof_inds,ndarray) 
        assert dof_inds.dtype == 'int32'
        values[self._dof_inds] = self._value
        dof_inds = append(dof_inds, self._dof_inds)
        return dof_inds, values

    def activate(self):
        " Activate the boundary"
        self._ligand.activate(self)
    
    def deactivate(self):
        " Activate the boundary"
        self._ligand.deactivate(self)
        
    def marker(self):
        " Return the value of the marker"
        return self._marker

    @classmethod
    def default_params(cls):
        p = ParameterDict(value = RangeCheck(1.0,ge=0.0))
        return p
    
    # Set the __init__ docstring from DiffsimObject
    __init__.__doc__ = DiffsimObject.__init__.__doc__
    

# Placed here to avoid circular dependency
from diffsim.diffusiveligand     import (DiffusiveLigand, MEAN, MIN, MAX, STD,
                                         VAR, ALLOWED_STAT)
from diffsim.diffusiondomain     import DiffusionDomain
from diffsim.stochastictimestepper import DiscreteMarkovModel

debug("%s loaded"%__name__)
