__author__ = "Johan Hake (johan.hake@gmail.com)"
__copyright__ = "Copyright (C) 2008 Johan Hake"
__license__  = "GNU LGPL Version 3.0 or any later"
__date__ = "2008-02-19 -- 2010-01-18"

# Defining the import * statement
__all__     = ['DiffusionDomain']


# Importing diffsim stuff
from diffsim.base            import DiffsimObject
from diffsim.dolfingallery   import DolfinGallery
from diffsim.diffusiveligand import DiffusiveLigand
from diffsim.parameterdict   import ParameterDict, TypeCheck, RangeCheck, OptionCheck
from dolfin                  import Expression, interpolate
from dolfin.cpp              import GenericFunction

# Initialize the logging system
from diffsim.logger import *
init_local_logger(__name__,globals(),INFO)

# Importing numpy stuff
import numpy

class DiffusionDomain(DolfinGallery, DiffsimObject):
    """ Encapsulation of a DiffusionDomain
    
    A X{DiffusionDomain} is a place holder for common objects that are defined on a
    certain diffusion domain. It takes a name and a DOLFIN mesh as arguments and
    create variables that can be used by attached objects, i.e., L{DiffusiveLigands}
    """
    
    def __init__(self, name, mesh_file, \
                 vector_degree=1, \
                 same_boundary_tensor=None, \
                 electric_potential=None):
        """
        @param name :
            The name of the diffusive domain
        @param mesh_file :
            The mesh that the domain is defined on
        @param same_boundary_tensor :
            A list of markers of potential same boundary tensors
        @param electric_potential :
            Either a user defined electric potential, a dolfin.GenericFunction
            or just True to use Expression controlled by parameters
    
        """
        #if not self.__class__ is DiffusionDomain:
        #    raise TypeError, "DiffusionDomain cannot be inherited" # FIXME: Why?
        
        info("Loading Mesh and assembling a mass matrix:")
        
        # Initializing the base classes (hardcode scalar degree of 1)
        DolfinGallery.__init__(self, name, mesh_file, 1, vector_degree, same_boundary_tensor)
        DiffsimObject.__init__(self, name)

        # Check for handed electric potential
        self._user_defined_epot = False
        self._epot = None

        # Use epot?
        if electric_potential is not None and electric_potential is not False:
            # Check type
            if not isinstance(electric_potential, (bool, GenericFunction)):
                raise TypeError, "expected a bool or GenericFunction"
            self._params.use_epot = True

            if isinstance(electric_potential, GenericFunction):
                self._user_defined_epot = True
                self._epot = electric_potential
            else:
                self._epot = Expression(self._params.expression,
                                        self._params.defaults)
            
            self._epot_values = interpolate(self._epot, self.scalar())
        
        # Protected attributes
        self._ligands = {}
        
        self._register_DiffsimObject(DiffusionDomain)

    def reset(self):
        # Re initialize parameters for electric potential
        if not self._params.use_epot:
            self._epot = None
            return
        
        if not self._user_defined_epot:
            self._epot = Expression(self._params.expression,
                                    self._params.defaults)
        
        assert(self._epot is not None)

    def register_ligand(self,ligand):
        " Register a ligand"
        assert isinstance(ligand,DiffusiveLigand)
        if self._ligands.has_key(ligand.ion_name()):
            raise AttributeError, \
                  "The %s already have a ligand for ion % registered"%\
                    (self.name(),ligand.ion_name())
        self._ligands[ligand.ion_name()] = ligand

    def get_ligand(self,ion_name):
        " Return a ligand instance given a ligand_name"
        if not self._ligands.has_key(ion_name):
            raise AttributeError, \
                  "The %s does not have a ligand for ion % registered"%\
                    (self.name(),ion_name())
        return self._ligands[ion_name]

    def iter_ligands(self):
        " Return an iterator over the registered ligands"
        return self._ligands.itervalues()
        
    def rename(self,new_name):
        " Not able to rename DiffusonDomain"
        raise RuntimeError, "Not able to rename DiffusonDomain"

    def electric_potential(self):
        " Return the electric potential, if any, otherwise None"
        return self._epot
    
    @classmethod
    def default_params(cls):
        return ParameterDict(
            use_epot = TypeCheck(False),
            expression = TypeCheck("phi_0*exp(-kappa*x[2])"),
            defaults   = TypeCheck(
                dict(phi_0   = -2.2,
                     kappa   = 1.0)),
            )

debug("%s loaded"%__name__)
