__author__ = "Johan Hake (johan.hake@gmail.com)"
__copyright__ = "Copyright (C) 2008 Johan Hake"
__license__  = "GNU LGPL Version 3.0 or any later"
__date__ = "2008-02-19 -- 2010-02-02"

# Defining the import * statement
__all__     = ['DiffusiveLigand','MEAN', 'MIN', 'MAX', 'STD', 'VAR','ALLOWED_STAT']

# Importing dolfin stuff
from dolfin import Vector, KrylovSolver, GenericVector

# Importing numpy and math
import numpy
import math 

# Importing diffsim stuff
from diffsim.common        import VALENCE
from diffsim.base          import (ContinuousVariable, check_ion_name,
                                   check_domain_name, DiffsimObject)
from diffsim.parameterdict import ParameterDict, TypeCheck, RangeCheck, OptionCheck
from diffsim.processtime   import tic, toc
from diffsim.dolfingallery import tensor as _tensor

from dolfin                import Expression, interpolate, Constant

# Initialize the logging system
from diffsim.logger import *
init_local_logger(__name__,globals(),INFO)

# Output from DiffusiveLigand statistics
MEAN   = "mean"
STD    = "std"
VAR    = "var"
MIN    = "min"
MAX    = "max"

ALLOWED_STAT = set([MEAN,STD,VAR,MIN,MAX])

class DiffusiveLigand(ContinuousVariable):
    """ Provides functionality to simulate diffusion of a ligand
    
    """
    
    def __init__(self, domain_name, ion_name, init_values, D = None, use_time_stepper = True, transient=True):
        """ Instantiate a DiffusiveLigand

        @param domain_name : The name of the DiffusionDomain the ligand resides in
        @param ion_name    : The name of the ion
        @param init_values : A scalar or vector with initial values
        @param D           : The value of the diffusion coeffisient
        @param use_time_stepper : True if the adaptive time stepper should be employed
        @param transient   : True if using tranisent solution
        """
        if not self.__class__ is DiffusiveLigand:
            raise TypeError, "DiffusiveLigand cannot be inherited" # FIXME: Why?
        
        # Check domain_name
        check_ion_name(ion_name)
        domain = check_domain_name(domain_name)
        
        name = "%s_%s"%(domain_name,ion_name)

        # Store ion name, as it is used in calling default_params in super class
        # NOTE: No it is not called in super class. It is used by default_params.
        self._ion_name = ion_name
        
        # Initialize the DiffsimObject
        super(DiffusiveLigand,self).__init__(name)
        
        # Check the init_values
        if isinstance(init_values,(GenericVector, numpy.ndarray)):
            if (len(init_values) != domain.num_dofs()):
                raise TypeError, """expected an initial concentration that has the same length as the degrees of freedom. """
        elif not isinstance(init_values,(int,float,numpy.integer)):
            raise TypeError, "expected a scalar, dolfin.Vector or numpy.ndarray for init_values"
        
        self._init_values = Vector(domain.num_dofs())
        self._init_values[:] = init_values
        
        # Check the diffusion constant D
        D = D or DIFFUSION_CONSTANT[ion_name]

        # Register the handed values in the params
        self._params.D           = D
        self._params.init_values = init_values
        self._params.transient   = transient

        # Protected attributes
        self._tensors       = {}
        self._nonlinear     = False
        self._domain        = domain
        self._active_fluxes = []
        self._active_fixed  = []
        self._boundaries    = {}
        self._num_iter      = 3
        self._use_time_stepper = use_time_stepper and transient 

        self._dt            = DiffusiveLigandDt(self) if self._use_time_stepper else None
        
        # Protected attributes from ContinuousVariable
        self._values_0      = self._domain.zero_dofs(copy = True)
        self._values_1      = self._values_0.copy()
        self._values_2      = self._values_0.copy()

        # A try to speed up value delivery
        self._value_cache   = {}
        
        # Private attributes
        self.__numpy_dofs    = self._domain.numpy_dofs(copy=True)
        self.__numpy_dofs[:] = 0.0
        self.__sym_solver    = None 
        self.__asym_solver   = None 
        self.__x             = Vector(self._domain.num_dofs())
        self.__b             = self.__x.copy()
        self.__source_vector = self.__x.copy()

        if self._domain.electric_potential() is not None:
            # Create weight Expression
            a = Expression(file_to_str("EpotWeight.h"))
            a.epot = self._domain.electric_potential()
            a.valence = VALENCE[self._ion_name]

            self._epot_weight = a
            
            # Create intermediats Vectors that are used in solve
            a_int = interpolate(a, self._domain.scalar())
            a_int_array = a_int.vector().array()
            
            self._solve_values_0 = self._domain.zero_dofs(copy=True)
            self._solve_values_to_conc = self._solve_values_0.copy()
            self._solve_values_from_conc = self._solve_values_0.copy()
            
            self._solve_values_to_conc[:] = a_int_array
            self._solve_values_from_conc[:] = 1/a_int_array
            
        # Assign memory
        self.__flux_matrix   = self.tensor("stiffnes")
        
        domain.register_ligand(self)
        self._register_DiffsimObject(DiffusiveLigand)
    
    # Implement DiffsimObject interface ----------------------------------------
    def reset(self):
        """ Reset the DiffusiveLigand before s simulation."""
        p = self._solver_params

        self.__sym_solver  = KrylovSolver(p.sym_solver, p.prec)     
        self.__asym_solver = KrylovSolver(p.asym_solver, p.prec)

        # Set the parameters
        krylov_params = self.__sym_solver.parameters
        krylov_params.relative_tolerance = p.tol_rel
        krylov_params.absolute_tolerance = p.tol_abs
        krylov_params.monitor_convergence = p.monitor_convergence
        
        # If not nonlinear increase the maximum number of iterations
        krylov_params.maximum_iterations = p.max_iter if p.nonlinear else p.max_iter*4
        
        # Set info level of solver
        krylov_params.report = True #get_log_level() <= DEBUG
        
        # Update the asym solver parameters
        self.__asym_solver.parameters.update(krylov_params)

        # Set the intial values
        self._values_1[:] = self._init_values

        self._source_vec_prev = self.source_vector()
        self._flux_matrix_vec_prev = self._source_vec_prev.copy()
        self._one_vec = self._source_vec_prev.copy()
        self._one_vec[:] = 1.0

        # Reset dependent variables
        for f in self._active_fluxes:
            f.reset()
        for f in self._active_fixed:
            f.reset()

        if self._domain.electric_potential() is not None:
            cpparg = "exp(-valence*%s)"%self._domain.get_params().expression
            defaults = self._domain.get_params().defaults.copy()
            defaults["valence"] = VALENCE[self._ion_name]
            
            # Create weight Expression
            a = Expression(cpparg, defaults)
            a_int = interpolate(a, self._domain.scalar())
            a_int_array = a_int.vector().array()
            
            self._solve_values_to_conc[:] = a_int_array
            self._solve_values_from_conc[:] = 1/a_int_array
        
        if self._params.init_with_ss:
            cache = self._params.transient
            self._params.transient = False
            self.solve(0.0)
            self._params.transient = cache

    # Note no class method, as we need the ion_name
    def default_params(self):
        " Returns the default parameters" 
        return ParameterDict(D = RangeCheck(DIFFUSION_CONSTANT[self._ion_name],gt=0),
                             init_values  = 0.0,
                             init_with_ss = TypeCheck(True),
                             transient    = TypeCheck(True),
            )

    def nonlinear(self):
        " Return true if the DiffusiveLigand is nonlinear"
        # If it is defined as nonlinear
        if self._nonlinear:
            return True
        
        # If not check the active fluxes
        return any(flux.nonlinear() for flux in self._active_fluxes)

    def nonlinear_boundaries(self):
        " Return the markers for any nonlinears boundaries"
        nonlinear = [flux.marker() for flux in self._active_fluxes if flux.nonlinear()]
        nonlinear.sort()
        return nonlinear

    def valence(self):
        " Return the valence of the diffusive ligand"
        return VALENCE[self._ion_name]
    
    # Implement ContinuousVariable interface -----------------------------------
    def solve(self, dt, iteration=0):
        """ Solve a sparse linear system
        
        A sparse linear system is built from attached DiffsimObjects and
        solved depending on the parameters in _solver_params
        
        @params dt : The time step
        @params iteration : The fixpoint iteration number
        """
        # Clear value cache
        self._value_cache = {}
        
        # The solver parameters
        p = self._solver_params

        # Solve!
        self._values_1[:] = self._solve_with_method(dt, self._values_0)

        # Check for negative values (No we don't)
        #if (self._values_1 < 0).any():
        #    dt_acc = 0
        #    dt_loc = dt/2
        #    info("### Negative values reducing time step, new dt: %s"%str(dt_loc))
        #    while dt_acc < dt:
        #        self._values_1[:] = self._solve_method(dt_loc,self._values_0)
        #        if (self._values_1 < 0).any():
        #            dt_loc /= 2
        #            info("### Negative values reducing time step, new dt: %s"%str(dt_loc))
        #        else:
        #            dt_acc += dt_loc
        #            self._values_0[:] = self._values_1
                
        # Fix point iteration stop criteria
        # We need to have solved the problem twice
        # FIXME: Is this critera right?
        if iteration >= 1:
            e = self.__error(self._values_1, self._values_2)
            
            if e < p.nl_tol_abs:
                self._converged = True
                info("%s fix point iteration solved in: %d iterations with error: %e"%\
                     (self._name,iteration, e))
            else:
                info("%s fix point iteration not solved. Error: %e"%\
                     (self._name, e))
    
    def _solve_with_method(self, dt, x0):
    
        # The solver parameters
        p = self._solver_params

        # Set up system
        b = self.__b
        x = self.__x
        
        # Initialize the solution vector with the previous solution
        x[:] = x0

        # Gather source vector and flux matrix
        b[:] = self.source_vector()

        A = self.flux_matrix()
    
        # Check if we are not solving a transient solve
        if not self._params.transient:
            x = self._linear_solve(A, x, b)
            return x

        Mij = self.tensor("mass")
        
        # Choose numerical method for time
        if p.tr_bdf2:
            # First step: Trapezoidal method
            
            # Initialize the solution vector with the previous solution
            b *= dt/2
            A *= dt/2
            
            A *= 0.5
            b += (Mij-A)*x
            
            # Add the Mass matrix
            A.axpy(1, Mij, True)
            
            # Solve system
            x = self._linear_solve(A, x, b)

            # Update the solution values and broadcast them
            self._values_1[:] = x

            # If using epot
            if self._domain.electric_potential() is not None:
                raise NotImplementedError, "tr/bdf solver is not implemented for electric potentials"
                self._values_1 *= self._solve_values_to_conc    
                
            self.send()
            
            # Second step: Backward difference scheme
            
            # Collect the source term and the flux_matrix
            A = self.flux_matrix()
            A.axpy(1.5/(dt/2), Mij, True)
            
            b[:] = self.source_vector()
            b.axpy(2/(dt/2), Mij*x)
            b.axpy(-0.5/(dt/2), Mij*self._values_0)
            
            # Solve  system
            x = self._linear_solve(A, x, b)

        else:
            b *= dt
            A *= dt
            
            if p.crank_nich:
                A *= 0.5
                b += (Mij-A)*x
            else:
                b += Mij*x
            
            # Add the Mass matrix
            A.axpy(1, Mij, True)

            # Solve system
            x = self._linear_solve(A, x, b)

        # Return solution
        return x

    def send(self):
        self._value_cache   = {}
        DiffsimObject.send(self)
    
    def _linear_solve(self,A,x,b):
        " Solve the linear system A*x=b"
        
        # The solver parameters
        p = self._solver_params

        # Choose solver
        if self.__is_symetric():
            solver = self.__sym_solver
        else:
            solver = self.__asym_solver
        
        # Apply any fixed boundary conditions, i.e., Dirichlet BC
        A, b = self.__apply_fixed(A, b)
        
        tic()
        
        # First test the previous solution
        r = A*x
        r -= b
        r*=r
        r = math.sqrt(r.sum())
        b_sum = math.sqrt((b*b).sum())
        if b_sum == 0.0:
            rel_r = numpy.inf
        else:
            rel_r = r/b_sum
        # FIXME: use rel tolerance in addtion?
        debug("Initial residual for linear problem, abs: %s rel: %s"%(str(r),str(rel_r)))
        if r < p.tol_abs or rel_r < p.tol_rel:
            info("No solve for '%s', solution in steady state. This took %ss."%(self.name(),str(toc())))
        else:
            self._num_iter = solver.solve(A, x, b)
            info("It took %ss to solve '%s' using %d iterations, by a %s solver with '%s' preconditioner."%\
                 (str(toc()),self._name, self._num_iter, type(solver).__name__, p.prec.lower()))
        
        return x
    
    def value(self, m = None, stat = None, copy = False, epot_weighted=True):
        """ Returns the values of the diffusion field

        @param m :
           A marker defining a subdomain of the field that is then returned
        @param stat :
           Return a statistic instead of an array. Available are:
           MIN, MAX, MEAN, STD, VAR
        @param copy :
           If True a copy of the field values are provided. Is ignored if a
           stat is given
        @param epot_weighted:
           If true we weight the values with the electric potential
        """

        epot_weighted *= self._domain.electric_potential() is not None

        # If we need the MEAN we multiply the values with the weighted
        # mass matrix instead. Done later
        if epot_weighted and stat in [None, MIN, MAX]:
            values = self._values_1*self._solve_values_to_conc
        else:
            values = self._values_1
    
        # If only returning the field values
        if stat is None:
            if m is None or m == 0:
                if copy:
                    return values
                else:
                    return values
            else:
                return values[self._domain.get_vertex_indices(m)]
        # If returning some kind of statistics
        else:
            if not stat in ALLOWED_STAT:
                raise TypeError, "The only allowed statistics are: %s"%\
                                  (",".join(ALLOWED_STAT))
            # If we are only takeing the MAX or MIN value of the field
            if stat in [MIN,MAX]:
                stat_str  = ".%s()"%stat
                slice_str = ""
                if not m is None:
                    slice_str = "[self._domain.get_vertex_indices(m)]"
                return eval("values%s%s"%(slice_str,stat_str))
            
            # If we ned to take one of VAR, STD or MEAN
            if m is None or m == 0:
                if 0 not in self._value_cache:
                    self._value_cache[0] = {}
                    self._value_cache[0][epot_weighted] = {}
                    
                # We take the stat from the whole domain
                if epot_weighted:
                    mass_matrix = self.tensor("mass")
                else:
                    mass_matrix = self._domain.tensor("mass")
                domain_size = self._domain.domain_volume()
                values      = values.copy()
            else:
                if m not in self._value_cache:
                    self._value_cache[m] = {}
                    self._value_cache[m][epot_weighted] = {}
                
                # We take the stat from a boundary subdomain
                if epot_weighted:
                    mass_matrix = self.tensor("mass boundary", m, reduced=True)
                else:
                    mass_matrix = self._domain.tensor("mass boundary", m, reduced=True)
                domain_size = self._domain.domain_area(m)
                values      = values[self._domain.get_vertex_indices(m)]
            
            # Calculate the mean
            self._value_cache[m][epot_weighted][MEAN] = self._value_cache[m].\
                                        get(MEAN,(mass_matrix*values).sum()/\
                                        domain_size)
            if stat == MEAN:
                return self._value_cache[m][epot_weighted][MEAN]
            
            if epot_weighted:
                values = self._values_1*self._solve_values_to_conc
            
            # The variance
            values -= self._value_cache[m][epot_weighted][MEAN]
            values *= values
            
            self._value_cache[m][epot_weighted][VAR] = self._value_cache[m]\
                            [epot_weighted].get(VAR,(mass_matrix*values).sum()\
                                                /domain_size)
            if stat == VAR:
                return self._value_cache[m][epot_weighted][VAR]
            
            # The standard deviation
            return math.sqrt(self._value_cache[m][epot_weighted][VAR])

    # Special functions for DiffusiveLigand ------------------------------------
    def activate(self, obj, minimize=True):
        """ Activate the provided object
        
        This function is called by a Flux or a FixedBoundary instant, passing them
        self, to the DiffusiveLigand. Depending of the type of the calling instant,
        the object is attached to either _active_fluxes or _active_fixed
        
        @param obj        : The instant that are activated
        @raise TypeError  : If the handed instant is not a Flux or FixedBoundary
        """
        if not isinstance(obj, (Flux, FixedBoundary)):
            raise TypeError, \
                  "The passed instant need to be either a Flux or a FixedBoundary"
        
        if isinstance(obj, Flux):
            # If obj is a Flux and not allready active, activate it
            if not obj in self._active_fluxes:
                self._active_fluxes.append(obj)
        else:
            # If obj is a FixedBoundary and not allready active, activate it
            if not obj in self._active_fixed:
                self._active_fixed.append(obj)
        
        if minimize:
            self.minimize_dt()
        
    def deactivate(self, obj, minimize=True):
        """ Deactivate the provided object
        
        This function is called by a Flux or a FixedBoundary instant, passing them
        self, to the DiffusiveLigand. Depending of the type of the calling instant,
        the object is detached from either _active_fluxes or _active_fixed
        
        @param obj        : The instant that are activated
        @raise TypeError  : If the handed instant is not a Flux or FixedBoundary
        """
        if not isinstance(obj,(Flux, FixedBoundary)):
            raise TypeError, \
                  "The passed instant need to be either a Flux or a FixedBoundary"
        if isinstance(obj, Flux):
            # If obj is a Flux and activated, deactivate it
            if obj in self._active_fluxes:
                self._active_fluxes.remove(obj)
        else:
            # If obj is a FixedBoundary and activated, deactivate it
            if obj in self._active_fixed:
                self._active_fixed.remove(obj)

        if minimize:
            self.minimize_dt()


    def minimize_dt(self):
        " Triggers a minimization of dt"
        if self._use_time_stepper:
            self._dt.minimize()
    
    def rename(self,new_name):
        " Not able to rename DiffusiveLigand"
        raise RuntimeError, "Not able to rename DiffusiveLigand"

    # Accessor functions -------------------------------------------------------
    def x(self):
        return self.__x
    
    def D(self):
        " Return the diffusion constant"
        return self._params.D

    def ion_name(self):
        return self._ion_name
    
    def domain(self):
        " Return the DiffusionDomain of the ligand"
        return self._domain

    def num_iter(self):
        " Return the number of iterations of the last call to solve"
        return self._num_iter

    def active_variables(self):
        " Return the active variables"
        variables = []
        variables.extend(self._active_fluxes)
        variables.extend(self._active_fixed)
        return variables
    
    def flux_matrix(self):
        " Returns the contribution to the system wide flux matrix"
        # Create a new matrix and add the other weighted matrices to this
        A = self._params.D * self.tensor("stiffnes")
        for flux in self._active_fluxes:
            flux.add_to_flux_matrix(A)
        return A

    # FIXME: Is this depricated?
    def register_boundary(self, boundary):
        " Register a boundary to the ligand"
        # FIXME: This functions should be able to outsourse to a generic Boundary
        assert(hasattr(boundary,"marker"))
        if self._boundaries.has_key(boundary.marker()):
            raise ValueError, "A boundary with marker %d allready excist for %"%\
                  (boundary.marker(), self.name())
        self._boundaries[boundary.marker()] = boundary

    def tensor(self, tensor_name, marker=0, copy=False, reduced=False):
        """ Return a tensor

        If the domain is defined with an electric potential then give
        the weighted version of the tensor
        """
        if self._domain.electric_potential() is None:
            return self._domain.tensor(tensor_name, marker, copy, reduced)
        tensor_name += " with electric potential"
        
        #FIXME: Caching breaks for several DiffusiveLigands with different VALENCE 
        return self._domain.tensor(tensor_name, marker=marker, \
                                   copy=copy, reduced=reduced, \
                                   coefficients= [self._epot_weight])
            
    def source_vector(self):
        "Returns the contribution to the system wide source vector"
        source = self.__numpy_dofs
        source.fill(0.0)
        for flux in self._active_fluxes:
            flux.add_to_source_vector(source)
       
        self.__source_vector[:] = source
        return self.__source_vector
    
    # Private functions ------------------------------------------------------
    def __error(self, val1, val2):
        " Return the norm of the difference between to two arrays"
        r = val1 - val2
        r *= r
        return math.sqrt(r.sum())

    def __frac_below_zero(self):
        " Return the fraction of negative values"
        return float((self._values_1 < 0).sum())/self._domain.num_dofs()
                
    def _re_assemble_matrices(self):
        " re-assemble diffusion matrices"
        # If no electric potential no re-assemble
        if self._domain.electric_potential() is None:
            return
            
    def __apply_fixed(self,A,b):
        " Apply active fixed boundary condition."
        #if self._active_fixed:
        if False: # FIXME: Apply the dolfin way of doing this!
            fixed_dof_inds = numpy.array([],'i')
            fixed_values   = self._domain.zero_dofs(copy = True)
            for fixed in self._active_fixed:
                fixed_dof_inds,fixed_values = fixed.apply(fixed_dof_inds,fixed_values)
            A, BC, C = dirichlet_boundarycondition(A, fixed_dof_inds)
            # No need for this line as C's zero fields are identical
            # with fixed_values's, but for safety we keep it.
            fixed_values = C*fixed_values
            b *= BC
            b -= fixed_values
        return A, b

    def __is_symetric(self):
        " Return True if all of the active fluxes is symetric"
        fluxes = self._active_fluxes[:]
        while fluxes:
            if not fluxes.pop().is_symetric():
                return False
        return True

# Placed here to avoid circular dependency
from diffsim.boundaries        import Flux, FixedBoundary
from diffsim.diffusiondomain   import DiffusionDomain
from diffsim.common            import (DIFFUSION_CONSTANT)
from diffsim.diffusiveliganddt import DiffusiveLigandDt
from diffsim.io                import file_to_str

debug("%s loaded"%__name__)
