__author__ = "Johan Hake (johan.hake@gmail.com)"
__copyright__ = "Copyright (C) 2008 Johan Hake"
__license__  = "GNU LGPL Version 3.0 or any later"
__date__ = "2010-01-16 -- 2010-01-29"

# Defining the import * statement
__all__     = ['DiffusiveLigandDt']

# Importing dolfin stuff
from dolfin import Vector, KrylovSolver, GenericVector

# Importing numpy
import numpy

# Importing diffsim stuff
from diffsim.base          import (DiscreteVariable, Singleton,
                                   check_ion_name, check_domain_name)
from diffsim.diffusiveligand import DiffusiveLigand
from diffsim.parameterdict import ParameterDict, TypeCheck, RangeCheck, OptionCheck
from diffsim.base          import ContinuousVariable

# Initialize the logging system
from diffsim.logger import *
init_local_logger(__name__,globals(),INFO)

class DiffusiveLigandDt(DiscreteVariable, Singleton):
    """ An Heuristic Time step controller for a Diffusive Ligand
    
    """

    def __init__(self, ligand):
        if not isinstance(ligand, DiffusiveLigand):
            raise TypeError, "expected a DiffusiveLigand"
        
        super(DiffusiveLigandDt, self).__init__(ligand.domain().name() + "_Dt")
        
        self._domain = ligand.domain()
        self._estimate_cache = None
        self._dt_prev = 1.0
        self._estimate_prev_0 = 1.0
        self._estimate_prev_1 = 1.0
        self._previous_values = {}
        self._num_low_rel_std = 0
        self._num_high_rel_std = 0
        self._register_DiffsimObject(DiffusiveLigandDt)
        
    @classmethod
    def default_params(cls):
        " Returns the default parameters" 
        return ParameterDict(
            weight_max   = RangeCheck(1.e10, gt=0,lt=1e15),
            weight_start = RangeCheck(5., gt=0,lt=1e5),
            num_previous_values = RangeCheck(3, ge=1,le=10),
            start        = RangeCheck(1e-10, gt=0,lt=10),
            minimal      = RangeCheck(1e-10, gt=0,lt=10),
            use_scale    = TypeCheck(True),
            dt_scale     = RangeCheck(2.0, gt=1,lt=10),
            weight_scale = RangeCheck(2.0, gt=1,lt=10),
            value_stat   = OptionCheck("max",["max","mean"]),
            )
    
    def estimate(self):
        " Calculate an estimate of next dt"
        if self._estimate_cache is None:
            calc_estimates = []
            for ligand in self._domain.iter_ligands():
                rel_der = self.rel_der(ligand)
                
                #print "Sum absolute derivative:",der.sum()
                #print "Max absolute derivative:",der.max()
                #print "Mean absolute derivative:",der.mean()
                
                #print "Sum relative derivative: %e"%rel_der.sum()
                #print "Max relative derivative: %e"%rel_der.max()
                #print "Mean relative derivative: %e"%rel_der.mean()
                #print "Ratio Mean/Max %e"%(rel_der.mean()/rel_der.max())
                #calc_estimates.append(1/rel_der.mean())#/d.domain_volume())
                debug("%s %s 1/relative derivative: %e"%(\
                    self._params.value_stat.capitalize(),
                    ligand.ion_name(),
                    1/rel_der))
                self._previous_values[ligand.ion_name()][-1] = 1/rel_der
                
                calc_estimates.append(self._previous_values[ligand.ion_name()].mean())
            
            # Save the minimal value
            self._estimate_cache = min(calc_estimates)
            #debug("ESTIMATE: %e, PREV: %e"%((self._estimate_cache + self._estimate_prev_0)/2,
            #                                (self._estimate_prev_0 + self._estimate_prev_1)/2))
            debug("dt estimate: %s, weight: %s"%(str(self._estimate_cache),str(self._weight)))
            debug("num low rel std: %d, num high rel std: %d"%(self._num_low_rel_std,self._num_high_rel_std))
            
        return self._estimate_cache

    def rel_der(self, ligand):
        " Return the relative derivative for a certain ligand."
        # Do this suing the numpy interface
        val = numpy.absolute((ligand.domain().tensor("mass")*ligand.value()).array())
        der = numpy.absolute((ligand.source_vector()-ligand.flux_matrix()*ligand.x()).array())
        
        rel_der = der/val
        
        if self._params.value_stat == "max":
            return rel_der.max()
        
        if self._params.value_stat == "mean":
            return rel_der.mean()
    
    def reset(self):
        self._dt_prev = self._params.start
        self._dt      = self._params.start
        self._weight  = self._params.weight_start
        for l in self._domain.iter_ligands():
            self._previous_values[l.ion_name()] = \
                 numpy.array([1/self.rel_der(l).max()]*ContinuousVariable.\
                                get_solver_params().num_previous_values)
        self._estimate_cache  = None
        self._estimate_pregv_0 = self._dt
        self._estimate_prev_1 = self._dt
        self._using_dt = True
        self._num_low_rel_std = 0
        self._num_high_rel_std = 0
    
    def minimize(self):
        for l in self._domain.iter_ligands():
            self._previous_values[l.ion_name()] = \
                    numpy.array([1/self.rel_der(l).max()]*ContinuousVariable.\
                    get_solver_params().num_previous_values)
        self._dt     = self._params.minimal
        self._weight = self._params.weight_start
        self._using_dt = True
        self._num_low_rel_std  = 0
        self._num_high_rel_std = 0
        
    def update_time(self,dt):
        " This function is called in the end of a time step and will reset the cache."
        # Set estimate_cache to None, forcing a recalculation of it
        for values in self._previous_values.itervalues():
            assert(isinstance(values,numpy.ndarray))
            values[0:-1] = values[1:]
        self._estimate_cache = None
    
    def step(self):
        """ An heuristic implementation of an expanding and collapsing time step

        The logic:
        The time step can either be controlled by an internal 'dumb' time step:
        self._dt, or on a more educated guess: estimate(). The dt from estimate()
        is based on weak formulation of the diffusion equation.
        To be continued...

        """
        # FIXME: Write out the whole doc string documenting what I am doing.
        max_iterations = 0
        for ligand in self._domain.iter_ligands():
            max_iterations = max(max_iterations,ligand.num_iter())
        if max_iterations == 0:
            self._weight = min(self._params.weight_max,
                               self._weight*self._params.weight_scale)
            debug("Solved in 0 iterations, expanding weight: %s"%str(self._weight))
            return True
        
        rel_std_estimate = []
        for values in self._previous_values.itervalues():
            assert(isinstance(values,numpy.ndarray))
            rel_std_estimate.append(values.std()/values.mean())
        
        max_iterations = self._params.num_previous_values
        
        if max(rel_std_estimate) < 0.5:
            if max(rel_std_estimate) < 1.e-2:
                self._num_low_rel_std = max_iterations
            self._num_low_rel_std += 1
        
        if max(rel_std_estimate) > 0.7:
            self._num_high_rel_std += 1
            if max(rel_std_estimate) > 1.0:
                self._num_high_rel_std = max_iterations +1
        
        if self._num_low_rel_std > max_iterations:
            self._weight = min(self._params.weight_max,self._weight*self._params.weight_scale)
            self._num_low_rel_std = 0
            debug("Expanding weight: %s"%str(self._weight))
        
        if self._num_high_rel_std > max_iterations:
            self._num_low_rel_std = 0
            self._num_high_rel_std = 0
            self._weight = max(self._params.weight_start,self._weight/(self._params.weight_scale *2))
            debug("Decreasing weight: %s"%str(self._weight))
        
        debug("rel_std_estimate: %s"%str(rel_std_estimate))
        
        # If the estimate is larger or less than 5% smaller than the estimate_prev
        # expand dt
        #estimate = (self.estimate() + self._estimate_prev_1)/2
        #estimate_prev = (self._estimate_prev_0 + self._estimate_prev_1)/2
        #if  -0.05 <= (1-estimate_prev/estimate):
        #    if self._using_dt:
        #        debug( "Expanding dt")
        #        self._dt *= self._params.dt_scale
        #    else:
        #        self._weight = min(self._params.weight_max,self._weight*self._params.weight_scale)
        #        debug( "Expanding weight: %s"%str(self._weight))
        #elif estimate <= estimate_prev/2:
        #    if self._using_dt:
        #        debug( "Reducing dt")
        #        self._dt /= (self._params.dt_scale*2)
        #    else:
        #        self._weight /= (self._params.weight_scale*4)
        #        debug( "Reducing weight: %s"%str(self._weight))
        #
        #self._estimate_prev_0 = self._estimate_prev_1
        #self._estimate_prev_1 = self.estimate()
        
        return True
    
    def next_event(self):
        #debug("Next event: %s"%str(self.estimate()*self._params.weight))
        dt = self.estimate()*self._weight
        #self._using_dt = self._dt < self.estimate()*self._weight
        #if self._using_dt:
        #    return self._dt
        #else:
        #    return self.estimate()*self._weight


        #dt = min(self.estimate()*self._weight,self._dt)
        #if not self._params.use_scale:
        #    self._dt = self.estimate()*self._params.weight
        #dt = self.estimate()*self._weight
        return dt
    
debug("%s loaded"%__name__)
