#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Johan Hake (johan.hake@gmail.com)"
__copyright__ = "Copyright (C) 2008 Johan Hake"
__license__  = "GNU LGPL Version 3.0 or any later"
__date__ = "2008-10-24 -- 2010-01-13"

from cPickle import Unpickler, Pickler
from os      import path, listdir, makedirs
from glob    import glob
from time    import strftime

import copy
import re

TIME_FORMAT = "%Y.%m.%d-%H.%M.%S"

__all__ = ['load','save','present_time_str','read_data']

def present_time_str():
    return strftime(TIME_FORMAT)

def find_data_filename(datafiles):
    basestr = 'data%.3d.cpickle'
    i = 1
    filename = basestr%i
    while filename in datafiles:
        i += 1
        filename = basestr%i
    return filename

def save(filename = None, with_time = False, directory = "data", **kwarg):
    """ Save data using cPickle

    @param filename  : The name of the file to save the data in, .cpickle
                       will be appended to the file name if not provided
    @param with_time : If True, an extra str with the present time is added
    @param directory : The name of the directory the data shall be save in,
                       defaults to 'data'
    @param **kwarg   : The actuall data to be saved. 

    """
    # If no data to save
    if len(kwarg) == 0:
        return
    
    # If the directory is nonexcisting create one
    if not path.isdir(directory):
        makedirs(directory)
    
    # The absolute path to the data directory
    datadir = path.join(path.abspath(path.curdir),directory)

    # If no filename is provided find the first data001.cpickle file
    # that do not excist in directory
    if filename is None:
        datafiles = [path.basename(s) for s in glob(path.join(datadir,'data*.cpickle'))]
        datafiles.sort()
        print datafiles
        filename = find_data_filename(datafiles)
        filename = path.abspath(path.join(datadir,filename))
        print filename
    elif with_time:
        base_filename = filename.split('.cpickle')[0]
        base_filename += "_" + present_time_str()
        filename = path.join(datadir,base_filename+'.cpickle')
    else:
        base_filename = filename.split('.cpickle')[0]
        filename = path.join(datadir,base_filename+'.cpickle')
        
    # Open the file for writing
    print "Saveing data in %s"%path.join(directory,path.basename(filename))
    f = open(filename,'w')
    p = Pickler(f)

    # Dump the dictionary kwarg
    p.dump(kwarg)
    f.close()

def load(filename, directory = "data", ns = None):
    """ Load a saved dictionary

    @param filename  : The name of the file where the data is stored
    @param directory : The name of the directory where the file resides
    @param ns        : A namespace that if provided will be updated with
                       the loaded data
    """
    if not isinstance(filename, str):
        raise TypeError, "expected a str as filename"
    
    if not isinstance(directory, str):
        raise TypeError, "expected a str as directory"
    
    if not (isinstance(ns, dict) or ns is None):
        raise TypeError, "expected a dict as ns"
    
    # If the directory is nonexcisting
    if not path.isdir(directory):
        raise IOError, "The directory '%s' does not excist"%directory

    base_filename = path.abspath(path.join(directory,filename.split('.cpickle')[0]))
    filename = base_filename + ".cpickle"
    
    # If the file is nonexcisting
    if not path.isfile(filename):
        # Next we check if the user accidentially passed the directory twice
        if not (filename.split(path.sep)[-3] == filename.split(path.sep)[-2] and filename.split(path.sep)[-2] == directory):
            raise IOError, "The file '%s' does not excist"%filename
        another_filename = filename.replace(directory + path.sep, "", 1)
        if not path.isfile(another_filename):
            raise IOError, "The file '%s' does not excist neither does %s"%(filename,another_filename)
        filename = another_filename
    
    # Open file
    f = open(filename,"r")
    un = Unpickler(f)
    
    # Read data
    print "Loading data from %s"%path.join(directory, path.basename(filename))

    data = un.load()
    f.close()

    if "diffsim_data" in data:
        data = data["diffsim_data"]
    
    if ns is None:
        return data
    
    assert(isinstance(ns,dict))
    
    ns.update(data)
    return data

# In memory cache
_data = {}

def compare_dicts(p1,p2):
    "Recursively compares a dict of values"
    import numpy
    assert(isinstance(p1,dict))
    assert(isinstance(p2,dict))
    try:
        ret = p1 == p2
    except ValueError, e:
        ret = True
        for key, value in p1.iteritems():
            if isinstance(value, numpy.ndarray):
                ret = ret and (value == p2[key]).all()
            elif isinstance(value, dict):
                ret = ret and compare_dicts(value,p2[key])
            else:
                ret = ret and value == p2[key]
    return ret

def read_data(casename):
    global _data
    assert(isinstance(casename,str))
    if _data.has_key(casename):
        print "Loading %s data from memory cache"%casename
        return _data[casename]
    
    data_files = [path.basename(f) for f in glob(path.join("data",casename) + "*")]
    if len(data_files) == 0:
        raise IOError, "No file with name " + path.join("data",casename) + "* excists."
    data_files.sort()
    params = None
    data = None
    for f in data_files:
        data_loc = load(f)
        if not params:
            params = copy.deepcopy(data_loc['params'])
        elif not compare_dicts(params,data_loc['params']):
            print "Not the same parameters, skipping this data file"
            continue
        if not data:
            data = data_loc
        else:
            for key in data.iterkeys():
                if hasattr(data[key],'extend'):
                    data[key].extend(data_loc[key])
    
    _data[casename] = data
    return data

