__author__ = "Johan Hake (johan.hake@gmail.com)"
__copyright__ = "Copyright (C) 2008 Johan Hake"
__license__  = "GNU LGPL Version 3.0 or any later"
__date__ = "2008-01-08 -- 2010-01-29"

# Defining the import * statement
__all__     = ['DolfinGallery']

try:
    from ufl           import *
except ImportError:
    raise ImportError, "Please install ufl on your system"
except e:
    raise e

try:
    from dolfin        import (Mesh, MeshFunction, assemble, IntersectionDetector,\
                               Vector, GenericVector, File, FunctionSpace, parameters, \
                               VectorFunctionSpace, Point, jit, has_la_backend)
    
    from dolfin        import Function as dFunction
except ImportError,e:
    raise ImportError, "Please install a proper PyDOLFIN on your system:"+str(e)
except e:
    raise e

if not has_la_backend("PETSc"):
    raise ImportError, "install PyDOLFIN with support for 'PETSc' linear algebra backend"

import numpy
import math

from diffsim.attrdict    import AttrDict
from diffsim.processtime import *
from diffsim.logger      import *

# Importing debugging stuf
init_local_logger(__name__,globals())
set_log_level(INFO,__name__)
set_log_level(INFO,"UFL")
set_log_level(INFO,"FFC")

# Set some parameters for the form generation and la backend
diffsim_backend = "PETSc"#"Epetra"

parameters.optimize = True
parameters.linear_algebra_backend = diffsim_backend

def _compile_forms(forms):
    for form_name, form in forms.iteritems():
        (compiled_form, module, form_data) = jit(form)
        forms[form_name] = compiled_form

def dim_to_str(dim):
    if dim == 3:
        return "tetrahedron"
    else:
        return "triangle"
    

# Usefull function for defining forms
def _init_forms(scalar, vector, constant):

    if vector.degree() == 1:
        vector1 = vector
    else:
        vector1 = VectorElement("Lagrange", dim_to_str(vector.cell().d), 1)
    vector0 = VectorElement("Discontinuous Lagrange", \
                            dim_to_str(vector.cell().d), 0)
    u      = TrialFunction(scalar)
    v      = TestFunction(scalar)
    c      = Function(constant)
    field0 = Function(vector0)
    field  = Function(vector)
    f      = Function(scalar)
    h      = Function(constant)
    
    field_dir  = Function(vector)
    stab0  = Function(constant)
    stab1  = Function(scalar)
    stab2  = Function(vector0)
    stab3  = Function(vector)
    
    forms  = {}

    # Mass forms
    forms["mass"]           = u*v*dx
    forms["mass boundary"]  = u*v*ds

    forms["mass with electric potential"]     = f*u*v*dx
    forms["mass boundary with electric potential"] = f*u*v*ds
    
    forms["mass with scalar field"]           = f*u*v*dx
    forms["mass boundary with scalar field"]  = f*u*v*ds

    # Partial integrated forms that assume no flux over boundaries
    forms["stiffnes"]                   = dot(grad(u),grad(v))*dx
    forms["stiffnes with scalar field"] = f*dot(grad(u),grad(v))*dx
    forms["stiffnes with electric potential"] = f*dot(grad(u),grad(v))*dx
    
    forms["advection"]      = -u*dot(field,grad(v))*dx
    
    # Streamline Upwind / Petrov Galerkin stabilizer
    #
    # The term
    # s  = \tau*(h/(2|E|)*dot(E,grad(v))
    #
    # where \tau need to be multiplied to the matrix after assembly
    
    advection                = dot(field, grad(u)) + u * div(field)
    advection                = div(field*u)
    advection2               = dot(field, grad(u))
    
    SUPG0                    = h*stab0*dot(field_dir, grad(v))
    SUPG1                    = h*stab1*dot(field_dir, grad(v))
    SUPG2                    = dot(stab2, grad(v))
    SUPG3                    = dot(stab3, grad(v))
    
    forms["SUPG advection0"] = advection*SUPG0*dx
    forms["SUPG advection1"] = advection*SUPG1*dx
    forms["SUPG advection2"] = advection*SUPG2*dx
    forms["SUPG advection3"] = advection*SUPG3*dx
    
    forms["SUPG mass0"]      = u*SUPG1*dx
    forms["SUPG mass1"]      = u*SUPG1*dx
    forms["SUPG mass2"]      = u*SUPG2*dx
    forms["SUPG mass3"]      = u*SUPG3*dx

    forms["load with scalar field"]           = f*v*dx
    forms["load with electric potential"]     = f*v*dx

    
    #_compile_forms()
    return forms

def _load_mesh(mesh_file):
    # FIXME:
    # Clean up after rework of extraction of marker informations!
    tic()
    
    # Load the mesh
    mesh = Mesh(mesh_file)
    mesh.order()
    mesh.init(2,0)
    debug("Loading and initialize Mesh took: %0.2fs"%toc())
    
    # Extract file name information
    base_name = mesh_file[:mesh_file.rfind('.xml')]
    mesh_suffix = mesh_file[mesh_file.rfind('.xml'):]

    # Populate the markers dictionary
    markers = AttrDict()
    for top_type in ["face", "vertex"]:
        # Try open the mesh function file
        try:
            file_name = base_name + "_%s_markers"%top_type + mesh_suffix
            
            mf = MeshFunction("uint",mesh,file_name)
            debug("Loading %s MeshFunction took: %0.2fs"%(top_type,toc()))
        except IOError, e:
            warning("Error while trying to open %s: %s"%(file_name,e))

        # Initialize the marker dictionary
        markers[top_type]         = AttrDict()
        markers[top_type]["vec"]  = dict()
        markers[top_type]["ind"]  = dict()
        markers[top_type]["size"] = dict()
        markers[top_type]["val"]  = dict()
        
        # Try extract the markers
        try:
            markers_val = mf.values().copy()
            debug("Extracting %s markers took: %0.2fs"%(top_type,toc()))
            if top_type == "face":
                unique_markers = list(set(markers_val))
                if 0 in unique_markers:
                    unique_markers.remove(0)
                unique_markers = numpy.array(unique_markers)
            vec = numpy.zeros(len(markers_val),'i')
            for m in unique_markers:
                if top_type == "vertex":
                    # Fill the vector with zeros
                    vec[:] = 0
                    conn20 = mesh.topology()(2,0)
                    for face_ind in markers["face"].ind[m]:
                        vec[conn20(face_ind)] = 1
                    markers[top_type].vec[m] = Vector(len(vec))
                    markers[top_type].vec[m][:] = numpy.array(vec,'d')
                else:
                    vec[:] = markers_val == m
                markers[top_type].ind[m] = vec.nonzero()[0]
                markers[top_type].size[m]= len(vec.nonzero()[0])
            debug("Extracting %s ind and vectors took: %0.2fs"%(top_type,toc()))
            markers[top_type].val = unique_markers
        except Exception, e:
            warning("An error occured while loading %s markers, %s"%(top_type,str(e)))
    #if not markers.face.val == markers.vertex.val:
    #    warning("Facet and vertex markers do not correspond") 
    debug("Loding mesh information took : %0.2fs"%toc(from_tic = True))
    
    # Return the mesh and the markers
    return mesh, markers

# Unbound method, to be bound!
def tensor(self, tensor_name, marker=0, copy=False, reduced=False, coefficients=None):
    " Return a tensor with form name 'tensor_name'"
    if not (self._tensors.has_key(tensor_name) and \
        self._tensors[tensor_name].has_key(marker)):
        self.assemble_tensor(tensor_name, marker=marker,
                             coefficients=coefficients)

    if reduced:
        # Compute the reduced matrix if not excisting
        if not self._tensors[tensor_name][marker].has_key("reduced"):
            M = self._tensors[tensor_name][marker]["full"]
            inds = self.get_vertex_indices(marker)
            ret_tensor = M[inds,inds]
            self._tensors[tensor_name][marker]["reduced"] = ret_tensor
        else:
            ret_tensor = self._tensors[tensor_name][marker]["reduced"]
    else:
        ret_tensor =  self._tensors[tensor_name][marker]["full"]

    if copy:
        return ret_tensor.copy()
    
    return ret_tensor

class DolfinGallery(object):
    """ A Gallery of tensors from predefined forms
    
    """

    def __init__(self, name, mesh_file, scalar_degree = 1, vector_degree = 1, same_boundary_tensor = None):
        """ Initialise the DolfinGallery

        @param name : The name of the DolfinGallery
        @param mesh_file : The name of the mesh_file
        @param scalar_degree : The degree of the scalar element
        @param vector_degree : The degree of the vector element
        
        """
        
        if not isinstance(name,str):
            raise TypeError, "Please provide a string as name"

        # Load mesh and markers from file
        mesh, markers = _load_mesh(mesh_file)
        
        if scalar_degree == 1:
            info("Num dofs: %d"%mesh.num_vertices())

        assert(isinstance(scalar_degree,int))
        assert(isinstance(vector_degree,int))
        
        if not same_boundary_tensor is None:
            assert(isinstance(same_boundary_tensor,list))
        
        # Load dimension from mesh
        cell_shape = dim_to_str(mesh.topology().dim())
        
        # Some functional spaces
        scalar    = FunctionSpace(mesh, "Lagrange", scalar_degree)
        vector    = VectorFunctionSpace(mesh, "Lagrange", vector_degree)
        scalar0   = FunctionSpace(mesh, "Discontinuous Lagrange", 0)
        
        # Protected attributes
        self._mesh       = mesh
        self._markers    = markers
        self._cell_shape = cell_shape
        self._scalar     = scalar
        self._scalar0    = scalar0
        self._vector     = vector
        self._vector0    = VectorFunctionSpace(mesh, "Discontinuous Lagrange", 0)
        self._tensors    = {}
                
        # Private attributes
        self.__name                  = name
        self.__scalar_field_function = dFunction(scalar)
        self.__vector_field_function = dFunction(vector)
        self.__scalar_field          = self.__scalar_field_function.vector()
        self.__vector_field          = self.__vector_field_function.vector()
        self.__forms                 = _init_forms(scalar.ufl_element(),
                                                   vector.ufl_element(),
                                                   scalar0.ufl_element())
        
        self.__default_matrix        = None
        self.__int_det               = IntersectionDetector(mesh)
        self.__same_boundary_tensor  = same_boundary_tensor or []
        
        # Init some attributes (previously in DiffusionDomain)
        
        # The number of degrees of freedom
        num_dofs  = self._mesh.num_vertices()
        
        # Array of ones and zeros with length num_dofs
        assert(num_dofs == self.scalar_field().size())
        self._one_dofs   = self.scalar_field().copy()
        self._zero_dofs  = self.scalar_field().copy()
        
        self._one_dofs[:] = 1.0
        self._zero_dofs[:] = 0.0
        
        # A temporary numpy array used when dealing with dolfin-numpy-dolfin stuff
        self._numpy_dofs = numpy.zeros(num_dofs,'d')
        self._num_dofs  = num_dofs
        
        self._domain_volume = (self.tensor("mass")*self._one_dofs).sum()
        self._domain_area   = dict()
        self._midpoints     = dict()
        
        # Check any handed boundary markers
        assert(all(v not in self.__same_boundary_tensor for v in [0,1,2,3]))
        
        for m in self.__same_boundary_tensor:
            self.check_face_marker(m)
            self.check_vertex_marker(m)
        
        # Instantiate the mass matrix
        self.assemble_tensor("mass")
        debug("Leaving dolfingallery...")
        
    def cell_shape(self):
        return self._cell_shape

    def scalar(self):
        return self._scalar

    def scalar0(self):
        return self._scalar0

    def vector(self):
        return self._vector

    def vector0(self):
        return self._vector0

    def scalar_field_function(self):
        return self.__scalar_field_function
    
    def vector_field_function(self):
        return self.__vector_field_function

    def scalar_field(self):
        return self.__scalar_field
    
    def vector_field(self):
        return self.__vector_field

    def mesh(self):
        " Return the dolfin mesh"
        return self._mesh

    def points_inside(self,points):
        """ Check if an array of points is inside the domain

        Return False if any points are outside the domain
        """
        assert isinstance(points,numpy.ndarray), "Provide a numpy array with points"
        if len(points.shape) == 1:
            assert points.shape[0] == 3, \
                   "Provide eiter a vector of size three or a matrix of 3D points"
            points.shape = (1,3)
        elif len(points.shape) == 2:
            assert points.shape[1] == 3, \
                   "Provide eiter a vector of size three or a matrix of 3D points"
        else:
            raise TypeError,\
                  "Provide eiter a vector of size three or a matrix of 3D points"
        size  = cells.size()
        for i in xrange(points.shape[0]):
            p = Point(points[i,0],points[i,1],points[i,2])
            cells = self.__int_det.intersection(p)
            if len(cells) == 0:
                return False
        return True
    
    def save_field(self,file_name, field = None):
        """ Save a field to a dolfin supported xml file format"""
        # FIXME: Some faulty logic in this function...
        if field is not None:
            assert isinstance(field,(numpy.ndarray,GenericVector)), "Can only save numpy arrays and dolfin.GenericVectors"
            if not(len(field) == self.__scalar_field.size() or \
                   len(field) == self.__vector_field.size()):
                raise TypeError, "Cannot save field of length %d."%len(field)
            
            # Create the file
            if not (".xml" in file_name):
                file_name = "%s.xml"%file_name
            
            f = File(file_name)
            
            # Is it a scalar or vector field we shall save
            if len(field) == self.__scalar_field.size():
                self.__scalar_field[:] = field
                f << self.__scalar_field
            else:
                self.__vector_field[:] = field
                f << self.__vector_field
        
    def assemble_tensor(self, form_name, \
                        coefficients=None, \
                        return_tensor=False,\
                        reassemble=False, \
                        marker=0,\
                        form_compiler_options=None):
        """ Assembles a given tensor
        
        A call to this function assembles a tensor with correspondning with the
        given form name.
        
        @param form_name     : The name of the form
        @param coeffisients  : A list of coeffisients passed to the dolfin assembler
        @param return_tensor : Return the tensor if True
        @param reassemble    : Reassembles tensor if True
        @param marker        : Assembles a tensor over a specific subdomain
        @param form_compiler_options : Options passed direct to the jit compiler

        @raise AttributeError : If no form with with the passed name excist
        """
        assemble_str = "assemble %s"%form_name
        info(assemble_str)
        tic(assemble_str)
        if not self.__forms.has_key(form_name):
            raise AttributeError, "No form with name %s registered"%form_name

        (compiled_form, module, form_data) = jit(self.__forms[form_name], \
                        form_compiler_options)
        
        tensor_name = "matrix"
        if compiled_form.rank() == 0:
            tensor_name = "scalar"
        elif compiled_form.rank() == 1:
            tensor_name = "vector"

        # If the tensor allready excist and we shall not reassemble the matrix
        if self._tensors.has_key(form_name) and \
               self._tensors[form_name].has_key(marker) and \
               not reassemble:
            M = self._tensors[form_name][marker]["full"]
        
        # If the tensor is not present in the tensor cache
        else:
            if not self._tensors.has_key(form_name):
                self._tensors[form_name] = {}
            
            cell_domains           = None
            exterior_facet_domains = None
            interior_facet_domains = None
            
            # Assemble over subdomain?
            if marker > 0:
                assert (compiled_form.num_cell_integrals() + \
                        compiled_form.num_exterior_facet_integrals()) == 1,\
                       "The form can only provide one integral."
                
                # Any exterior facet integrals
                if compiled_form.num_exterior_facet_integrals() == 1:
                    self.check_face_marker(marker)
                    # Initialize a face MeshFunction
                    mf = MeshFunction("uint", self._mesh,2)
                    
                    # Fill it with 1
                    mf.values()[:] = 1
                    
                    # Set zero at the indices of all same_boundar_tensor markers 
                    if marker in self.__same_boundary_tensor:
                        for m in self.__same_boundary_tensor:
                            mf.values()[self._markers.face.ind[m]] = 0
                    else:
                        mf.values()[self._markers.face.ind[marker]] = 0
                    exterior_facet_domains = mf
                else:
                    # FIXME: Implement?
                    raise NotImplementedError, \
                          "Integral over cell domains not implemented."
            # If we have not assembled a tensor before, assemble a new one
            if self.__default_matrix is None:
                # Assemble the matrix
                M = assemble(compiled_form, \
                             coefficients           = coefficients, \
                             function_spaces        = self._scalar, \
                             cell_domains           = cell_domains, \
                             exterior_facet_domains = exterior_facet_domains, \
                             interior_facet_domains = interior_facet_domains)
                # FIXME: What about vectors? 
                self.__default_matrix = M.copy() # Store the matrix
            else:
                # Initialize the matrix
                M = self.__new_matrix()
                M.zero()
                M = assemble(compiled_form, \
                             function_spaces        = self._scalar, \
                             coefficients           = coefficients, \
                             cell_domains           = cell_domains, \
                             exterior_facet_domains = exterior_facet_domains, \
                             interior_facet_domains = interior_facet_domains, \
                             reset_sparsity         = False, \
                             tensor                 = M)

            # Store the tensor
            if marker in self.__same_boundary_tensor:
                for m in self.__same_boundary_tensor:
                    self._tensors[form_name][m] = {}
                    self._tensors[form_name][m]["full"] = M
            else:
                self._tensors[form_name][marker] = {}
                self._tensors[form_name][marker]["full"] = M
            info("Assembling %s %s took: %0.2fs"%(form_name, tensor_name,\
                                                  toc(assemble_str)))
        if return_tensor:
            return M
        
    def num_dofs(self):
        return self._num_dofs
        
    def domain_area(self, marker = None):
        " Return the area of the whole domain or from a sub domain"
        m = marker or 0
        if m > 0:
            self.check_face_marker(m)
        if not self._domain_area.has_key(m):
            if m == 0:
                mass_matrix = self.tensor("mass boundary", m)
                dofs = self._one_dofs
            else:
                mass_matrix = self.tensor("mass boundary", m, reduced=True)
                dofs = self._one_dofs[self.get_vertex_indices(m)]
            self._domain_area[m] = (mass_matrix*dofs).sum()
        return self._domain_area[m]

    def midpoint(self, marker):
        " Return a midpoint of a domain" 
        self.check_face_marker(marker)
        if not self._midpoints.has_key(marker):
            self._midpoints[marker] = self._mesh.coordinates()\
                                       [self.get_vertex_indices(marker)].mean(0)
        return self._midpoints[marker]
            
    def domain_volume(self):
        " Return the volume of the whole domain"
        return self._domain_volume
    
    def one_dofs(self, copy = False):
        if copy:
            return self._one_dofs.copy()
        else:
            return self._one_dofs
    
    def zero_dofs(self, copy = False):
        if copy:
            return self._zero_dofs.copy()
        else:
            return self._zero_dofs
        
    def numpy_dofs(self, copy = False):
        if copy:
            return self._numpy_dofs.copy()
        else:
            return self._numpy_dofs
        
    def check_face_marker(self,marker):
        " Check if a marker is a registered face marker"
        if not marker in self._markers.face.val:
            raise RuntimeError, \
                  "The domain %s does not include the face marker %d."%\
                  (self.__name,marker)
    
    def check_vertex_marker(self,marker):
        " Check if a marker is a registered vertex marker"
        if not marker in self._markers.vertex.val:
            raise RuntimeError, \
                  "The domain %s does not include the vertex marker %d."%\
                  (self.__name,marker)
    
    def get_face_indices(self, marker):
        " Return the face indices for marker m"
        self.check_face_marker(marker)
        return self._markers.face.ind[marker]
    
    def get_vertex_indices(self, marker):
        " Return the vertex indices for marker m"
        self.check_vertex_marker(marker)
        return self._markers.vertex.ind[marker]
    
    def get_vertex_vector(self, marker):
        " Return the vertex indices for marker m"
        self.check_vertex_marker(marker)
        return self._markers.vertex.vec[marker]

    # Bind tensor method
    tensor = tensor
    
    def __new_matrix(self):
        M = self.__default_matrix.copy()
        return M

    def L2_norm(self,e):
        "Return the L2 norm of e"
        assert(isinstance(e,(numpy.ndarray,GenericVector)))
        return math.sqrt((self.tensor("mass")*(e*e)).sum())
    
    def L2_norm_w(self,e):
        "Return the L2 norm of e weighted with the volume "
        assert(isinstance(e,(numpy.ndarray,GenericVector)))
        return math.sqrt((self.tensor("mass")*(e*e)).sum()/self.domain_volume())

    def normalized_L2_norm(self,u,analytic):
        "Return a normalized L2 norm of u-analytic"
        assert(isinstance(u,(numpy.ndarray,GenericVector)))
        assert(isinstance(analytic,(numpy.ndarray,GenericVector)))
        e = u-analytic
        return self.L2_norm(e)/self.L2_norm(analytic)

    def L2_boundary_norm(self,e,marker = None):
        " Return the L2 norm of e on the boundary, or sub domain of the boundary"
        assert(isinstance(e,(numpy.ndarray,GenericVector)))
        m = marker or 0
        self.check_face_marker(m)
        # Nullify the other entries
        e *= self._markers.vertex.vec[m]
        return math.sqrt((self.tensor("mass boundary",m)*(e*e)).sum())

    def L2_boundary_norm_w(self,e,marker = None):
        " Return the L2 norm of e on the boundary, or sub domain of the boundary weighted with the area"
        assert(isinstance(e,(numpy.ndarray,GenericVector)))
        m = marker or 0
        self.check_face_marker(m)
        # Nullify the other entries
        e *= self._markers.vertex.vec[m]
        return math.sqrt((self.tensor("mass boundary",marker)*(e*e)/
                     self.domain_area(m)).sum())

    def Max_norm(self,e):
        "Return the Max norm of e"
        assert(isinstance(e,(numpy.ndarray,GenericVector)))
        if isinstance(e,GenericVector):
            e = e.array()
        return ((self.tensor("mass")*(numpy.absolute(e))).sum())
        
    def Max_norm_w(self,e):
        "Return the Max norm of e weighted with the volume "
        assert(isinstance(e,(numpy.ndarray,GenericVector)))
        if isinstance(e,GenericVector):
            e = e.array()
        return math.sqrt((self.tensor("mass")*(numpy.absolute(e))).sum()/self.domain_volume())

    def Max_boundary_norm(self,e,marker = None):
        " Return the Max norm of e on the boundary, or sub domain of the boundary"
        assert(isinstance(e,(numpy.ndarray,GenericVector)))
        if isinstance(e,GenericVector):
            e = e.array()
        m = marker or 0
        self.check_face_marker(m)
        return math.sqrt((self.tensor("mass boundary",m)*(numpy.absolute(e))).sum())

    def Max_boundary_norm_w(self,e,marker = None):
        " Return the Max norm of e on the boundary, or sub domain of the boundary weighted with the area"
        assert(isinstance(e,(numpy.ndarray,GenericVector)))
        if isinstance(e,GenericVector):
            e = e.array()
        m = marker or 0
        self.check_face_marker(m)
        # Nullify the other entries
        e *= self._markers.vertex.vec[m]
        return math.sqrt((self.tensor("mass boundary",marker)*(numpy.absolute(e))/
                     self.domain_area(m)).sum())

debug("%s loaded"%__name__)
