__author__ = "Johan Hake (johan.hake@gmail.com)"
__copyright__ = "Copyright (C) 2008 Johan Hake"
__license__  = "GNU LGPL Version 3.0 or any later"
__date__ = "2008-09-04 -- 2010-02-02"

__all__ = ['ConstantChannelFlux',
           'LCCVoltageDepFlux',
           'LCCVoltageDepNonlinearFlux',
           'NaVoltageDepFlux']

from diffsim.base            import ContinuousVariable
from diffsim.boundaries      import LinearFluxBoundary
from diffsim.diffusiondomain import DiffusionDomain
from diffsim.diffusiveligand import DiffusiveLigand
from diffsim.common          import (FARADAY,UNIVERSAL_GAS,T_ABS,VALENCE,MILLI,
                                     MICRO,NANO,PICO,NM3_per_DM3,ALLOWED_IONS,Na,Ca)
from diffsim.parameterdict   import ParameterDict, TypeCheck, RangeCheck, OptionCheck

import numpy

# Initialize the logging system
from diffsim.logger import *
init_local_logger(__name__,globals(),INFO)

class ConstantChannelFlux(LinearFluxBoundary):
    """ A class to represent a constant channel flux
    """
    # FIXME: Implement an ion layer on top of diffusive ligand
    def __init__(self, name, domain_name, ion_name, marker, amp, activator = None):
        """ Instantiate a ConstantChannelFlux
        
        param @name        : The name of the LinearFluxBoundary
        param @ligand_name : The name of the ligand the flux is defined on
        param @marker      : An int that defines the subdomain of the boundary
        param @ion         : An ion type (for the valence value)
        param @amp         : The amplitude of the current in pA
        param @activator   : A class or an object that can be used to connect
                             the LinearFluxBoundary.
                             If no activator is provided the Boundary is assumed
                             activated by default.
        """
        # Initiate the super classes
        super(ConstantChannelFlux,self).__init__(name, domain_name, ion_name,
                                                 marker, activator = activator,
                                                 standard_mass=True)
        # Set the valence of the ion
        if ion_name.capitalize() in ALLOWED_IONS:
            self._z = VALENCE[ion_name.capitalize()]
        else:
            raise TypeError, "The provided ion needs to be one of %s"%\
                  (", ".join(s.capitalize() for s in ALLOWED_IONS))
        
        self._params.amp = amp

        print "calc c0:", self._calc_c0()
        self._check_and_set_alpha_c0(0.0, self._calc_c0())
        self._update_c0()
        
        print "sum b:", self._source_add.sum()
        print "sum nonzero b:", (self._source_add > 0.0).sum()
        
        # Register the instantiated object
        self._register_DiffsimObject(ConstantChannelFlux)
        
    def reset(self):
        " Update the the alpha and c0 before simulation"
        self._check_and_set_alpha_c0(0.0, self._calc_c0())
        self._update_c0()

    def _calc_c0(self):
        " Calculate the c0 value for the linear boundary"
        F = FARADAY*MICRO        # Faradays constant
        z = self._z              # Valence
        nm3_per_dm3 = 1e24 # Misc. conversion constants
        ms_per_s = 1 / MILLI
        amp  = self._params.amp
        area = self._area
        
        # And finally setting the actuall values of the ion fluxes,
        # corresponding to a general Robin BC
        self._dCa_dt = -amp / (z * F * ms_per_s ) * PICO
        return self._dCa_dt / area * NM3_per_DM3
       
    @classmethod
    def default_params(cls):
        p = ParameterDict(amp = RangeCheck(0.0, ge=-100.0 ,le=100),
                          **super(ConstantChannelFlux,cls).default_params())
        return p
    
    def dCa_dt(self):
        " Return"
        return self._dCa_dt

    def minimize_from_start(self):
        return True
        
class LCCVoltageDepFlux(LinearFluxBoundary):
    """ A class to represent Voltage dependent flux using constant field approximation

    This is a Goldman-Hodgkin-Katz model of the Ca flux.
    
    """
    def __init__(self, domain_name, marker, activator = None):
        # Initiate the super classes
        super(LCCVoltageDepFlux, self).__init__("LCC", domain_name, Ca, marker, \
                                                activator=activator,
                                                standard_mass=True)

        # A place holder for saveing the current
        self._current_values = []
        
        # Set the voltage dependency
        self._ext_vars["V"] = -90.0

        # Register the instantiated object
        self._register_DiffsimObject(LCCVoltageDepFlux)
    
    @classmethod
    def default_params(cls):
        return ParameterDict(save_current= OptionCheck(0,[0,1]),
                             CaO      = RangeCheck(2.e3,gt=0),
                             CaO_part = RangeCheck(1.0,gt=0),
                             default  = ParameterDict(
                                           amp = RangeCheck(-0.12,lt=0),
                                           V   = RangeCheck(0.0,ge=-120,le=60),
                                           c0  = RangeCheck(570.,gt=0,le=2000)
                                           ),
                             **super(LCCVoltageDepFlux,cls).default_params())
    
    _params = 0
    _permeability  = None
    _CaO = None
    _default_c0 = None
    
    def reset(self):
        LCCVoltageDepFlux._permeability = None
        LCCVoltageDepFlux._CaO = self._params.CaO
        LCCVoltageDepFlux._default_c0 = self._params.default.c0

        if self._params.save_current:
            self._current_values.append([])
        self._update_changed_ext_val()
        
    def add_to_source_vector(self, source):
        if self._params.save_current:
            #c = self._ligand.value(self._marker, "mean", epot_weighted=True)
            c = self._ligand.value(self._marker, "mean")
            print "Concentration:", c
            amp = self.calculate_current(self._ext_vars["V"], c)
            amp *= self.permeability()
            if self._current_values:
                self._current_values[-1].append(amp)
            info("LCC %d current: %s, alpha: %s, c0: %s"%(self._marker, str(amp), str(self._alpha), str(self._c0)))
        LinearFluxBoundary.add_to_source_vector(self, source)

    @classmethod
    def calculate_current(cls, V, c):
        """ Return the value of an ordinary GHK current equation

        It uses the values for CaO and P given in the params
        """
        from math import exp, pi
        params = cls._params or cls.default_params()
        F = FARADAY        # Faradays constant        [C/mole]
        R = UNIVERSAL_GAS  # Universal gas constant [J/(mole*K)]
        T = cls.get_T()
        permeability = 1       # nm/ms
        z = VALENCE["Ca"]      # Ca Valence
        CaO = LCCVoltageDepFlux._CaO
        
        # FIXME:
        # A nasty hack which does not scale with different channel areas
        area = pi#self._ligand.domain().domain_area(self._marker)
        ms_per_s = 1 / MILLI
        
        V_scaled = V/(R*T/(z*F*MILLI))
        if V_scaled != 0.0:
            alpha = V_scaled/(1-exp(-V_scaled))
            c0 = V_scaled/(exp(V_scaled)-1)*CaO
        else:
            alpha = 1
            c0 = CaO

        # Unit of alpha and c0: uM*nm/ms == D*(c-c0)/nm
        
        dCa_dt = permeability*(alpha*c - c0) * area / NM3_per_DM3 # umole/ms
        return dCa_dt*MICRO*(z * F * ms_per_s )/PICO # pA
    
    @classmethod
    def permeability(cls, V=None, c0=None, amp=None): # nm/ms
        # Calculate the permiability based on the given V, c0 and amp
        if cls._permeability is None:
            params = cls._params or cls.default_params()
            V = V or params.default.V
            c0 = c0 or LCCVoltageDepFlux._default_c0
            info("C0:%s"%str(c0))
            amp = amp or params.default.amp
            
            cls._permeability =  amp / cls.calculate_current(V, c0)
        
        return cls._permeability
    
    def _update_changed_ext_val(self):
        from math import exp
        F = FARADAY            # Faradays constant        [C/mole]
        R = UNIVERSAL_GAS      # Universal gas constant [J/(mole*K)]
        T = self.get_T()
        z = VALENCE["Ca"]      # Ca Valence
        CaO = LCCVoltageDepFlux._CaO
        
        V_scaled = self._ext_vars["V"]/(R*T/(z*F*MILLI))
        print "Update_changed_ext_val:"
        print "V_scaled:", V_scaled
        if V_scaled != 0.0:
            self._alpha = V_scaled/(1-exp(-V_scaled))
            self._c0 = V_scaled/(exp(V_scaled)-1)*CaO
        else:
            self._alpha = 1
            self._c0 = CaO
        
        self._alpha *= self.permeability()
        self._c0    *= self.permeability()#*numpy.exp(-2*-2.2*numpy.exp(-2.0))
        self._update_c0()
        if get_log_level <= DEBUG:
            amp = self.permeability()*self.calculate_current(\
                                self._ext_vars["V"],\
                                self._ligand.value(self._marker,"mean")*\
                                numpy.exp(2*-2.2*numpy.exp(-2.0)))
            debug( "LCC %d Ca amp [pA]: %s"%(self._marker,str(amp)))
    
    def finalize(self):
        from numpy import array
        if self._params.save_current:
            for i, values in enumerate(self._current_values):
                self._current_values[i] = array(values)
            return self._current_values
        
        return None

        
class LCCVoltageDepNonlinearFlux(LinearFluxBoundary):
    """ A class to represent Voltage dependent flux using constant field approximation

    This calculate a fixed current for each timestep, based on the mean value of the
    [Ca] at the channel.

    Parameter values from:
    
    Luo, C. H. & Rudy, Y.
    A dynamic model of the cardiac ventricular action potential. I.
    Simulations of ionic currents and concentration changes.
    Circ. Res., 1994, 74, 1071-1096
    """
    def __init__(self, domain_name, marker, activator = None):
        # Initiate the super classes
        super(LCCVoltageDepNonlinearFlux,self).__init__("LCC", domain_name, \
                                                        Ca, marker, \
                                                        activator=activator,
                                                        standard_mass=True)
        # Set the voltage dependency
        self._ext_vars["Dyadic_cleft_Ca"] = self._ligand.value(marker, "mean")
        self._ext_vars["V"] = -90.0
        self._ext_vars["time"] = 0.0
        
        # Store previous values of Ca
        self.reset_previous_values()

        self._nonlinear = 1

        self._added_present_current_value_once = False
        
        # Register the instantiated object
        self._register_DiffsimObject(LCCVoltageDepNonlinearFlux)
    
    # A place holder for saveing the current for all LCCs
    _present_current_value = 0.0
    _current_values = []
    _time_values = []
    
    # Initiate class storage of parameters
    _params = 0
    _permeability = None
    
    def activate(self):
        "Called by diffusiveligand when activating the flux"
        self.reset_previous_values()
        super(LCCVoltageDepNonlinearFlux, self).activate()

    def reset_previous_values(self):
        self._previous_Ca = [self._ext_vars["Dyadic_cleft_Ca"]]*\
                            ContinuousVariable.get_solver_params().num_previous_values
        

    @classmethod
    def default_params(cls):
        return ParameterDict(save_current=OptionCheck(0, [0, 1]),
                             CaO=RangeCheck(2.e3, gt=0),
                             deactivate_at_start=OptionCheck(1, [0, 1]),
                             default=ParameterDict(\
                                 amp = RangeCheck(-0.2, lt=0),\
                                 V   = RangeCheck(0.0, ge=-100, le=40),\
                                 c0  = RangeCheck(560., gt=0, le=1000)),\
                             **super(LCCVoltageDepNonlinearFlux, cls).default_params())
    
    def reset(self):
        # Deactivate if it is an active flux
        if self._params.deactivate_at_start:
            self.deactivate(False)
        LCCVoltageDepNonlinearFlux._permeability = None
        self.reset_previous_values()

        self._update_changed_ext_val()

        if self._params.save_current:
            # Only append if there has not been appended a list before
            if len(LCCVoltageDepNonlinearFlux._current_values) == 0 or \
                   LCCVoltageDepNonlinearFlux._current_values[-1][-1] != 0.0:
                LCCVoltageDepNonlinearFlux._current_values.append([LCCVoltageDepNonlinearFlux._present_current_value])
                LCCVoltageDepNonlinearFlux._time_values.append([0.0])
                LCCVoltageDepNonlinearFlux._present_current_value = 0.0

    def receive(self, sender):
        # If receiving the time for the "first" time in a time step
        if not self._params.save_current:
            super(LCCVoltageDepNonlinearFlux, self).receive(sender)
            return
        
        if sender.name() == "time":
            # Reset counter for adding current values
            self._added_present_current_value_once = False

            # If we have not already registered the collected current for this time step
            if sender.value() != LCCVoltageDepNonlinearFlux._time_values[-1][-1]:

                # Add time and current value
                LCCVoltageDepNonlinearFlux._time_values[-1].append(sender.value())
                LCCVoltageDepNonlinearFlux._current_values[-1].append(LCCVoltageDepNonlinearFlux._present_current_value)
                
                debug("'%s' received time."%str(self.__class__.__name__))
                debug("present current value: %s"%str(LCCVoltageDepNonlinearFlux._present_current_value))
                
                # Reset current values before next timestep
                LCCVoltageDepNonlinearFlux._present_current_value = 0.0
            
                
        super(LCCVoltageDepNonlinearFlux, self).receive(sender)

    def add_to_source_vector(self, source):
        if self._params.save_current:
            amp = self.calculate_current(self._ext_vars["V"], self._ligand.value(self._marker, "mean"))
            amp *= self.permeability()

            # If we have not added the current already for this boundary
            if not self._added_present_current_value_once:
                self._added_present_current_value_once = True
                LCCVoltageDepNonlinearFlux._present_current_value += amp
            
            info("LCC %d current: %s, alpha: %s, c0: %s"%(self._marker, str(amp), str(self._alpha), str(self._c0)))
        LinearFluxBoundary.add_to_source_vector(self, source)

    @classmethod
    def calculate_current(cls, V, c):
        """ Return the value of an ordinary GHK current equation

        It uses the values for CaO and P given in the params
        """
        from math import exp, pi
        params = cls._params or cls.default_params()
        F = FARADAY        # Faradays constant        [C/mole]
        R = UNIVERSAL_GAS  # Universal gas constant [J/(mole*K)]
        T = cls.get_T()
        permeability = 1       # nm/ms
        z = VALENCE["Ca"]      # Ca Valence
        CaO = params.CaO
        
        # FIXME:
        # A nasty hack which does not scale with different channel areas
        area = pi#self._ligand.domain().domain_area(self._marker)
        ms_per_s = 1 / MILLI
        
        V_scaled = V/(R*T/(z*F*MILLI))
        if V_scaled != 0.0:
            alpha = V_scaled/(1-exp(-V_scaled))
            c0 = V_scaled/(exp(V_scaled)-1)*CaO
        else:
            alpha = 1
            c0 = CaO

        # Unit of alpha and c0: uM*nm/ms == D*(c-c0)/nm
        dCa_dt = permeability*(alpha*c - c0) * area / NM3_per_DM3 # umole/ms
        return dCa_dt*MICRO*(z * F * ms_per_s )/PICO # pA
    
    @classmethod
    def permeability(cls, V=None, c0=None, amp=None): # nm/ms
        # Calculate the permiability based on the given V, c0 and amp
        if cls._permeability is None:
            params = cls._params or cls.default_params()
            V = V or params.default.V
            c0 = c0 or params.default.c0
            info("C0:%s"%str(c0))
            amp = amp or params.default.amp
            
            cls._permeability =  amp / cls.calculate_current(V, c0)
        
        return cls._permeability
    
    def _update_changed_ext_val(self):
        F = FARADAY*MICRO      # Faradays constant        [C/mole]
        z = VALENCE["Ca"]         # Ca Valence
        area = self._area
        ms_per_s = 1 / MILLI

        # Register new values of Ca
        self._previous_Ca.append(max(self._ext_vars["Dyadic_cleft_Ca"],0.0))
        
        # pop old values
        self._previous_Ca.pop(0)
        
        # Take the mean of the previous registered values
        c = sum(self._previous_Ca)/len(self._previous_Ca)
        V = self._ext_vars["V"]
        
        amp = self.calculate_current(V, c)
        amp *= self.permeability()
        
        # And finally setting the actuall values of the ion fluxes,
        # corresponding to a general Robin BC
        self._dCa_dt = -amp / (z * F * ms_per_s ) * PICO
        self._c0 = self._dCa_dt / area * NM3_per_DM3
        
        self._update_c0()
    
    def finalize(self):
        from numpy import array
        if self._params.save_current and LCCVoltageDepNonlinearFlux._current_values is not None:
            return_values = LCCVoltageDepNonlinearFlux._current_values
            LCCVoltageDepNonlinearFlux._current_values = None
            for i, values in enumerate(return_values):
                return_values[i] = array(values)
            return return_values
        
        return None


class NaVoltageDepFlux(LinearFluxBoundary):
    """ A class to represent Voltage dependent flux using constant field approximation

    Parameter values from: Conductance from Patlak 1988 ()
    
    """
    def __init__(self, domain_name, marker, g_Na = 20.0, activator = None):
        # Initiate the super classes
        super(NaVoltageDepFlux,self).__init__('Na', domain_name, Na, marker, \
                                               activator = activator)
        # Set the valence of the ion
        self._z = VALENCE[Na]
        
        # Set the voltage dependency
        self._ext_vars["V"] = -90.0
        self._ext_vars["Dyadic_cleft_Na"] = self._ligand.value(marker)
        self._nonlinear = 1
        self._fudge = 1.0
        self._previous_values = {Na:numpy.array([self._ext_vars[self._ligand.name()]]*\
             ContinuousVariable.get_solver_params().num_previous_values)}
        
        # Register the instantiated object
        self._register_DiffsimObject(NaVoltageDepFlux)
    
    @classmethod
    def default_params(cls):
        return ParameterDict(g_Na = RangeCheck(25.0,gt=0),         # pS
                             NaO  = RangeCheck(140e3,gt=0),        # uM
                             **super(NaVoltageDepFlux,cls).default_params())
    
    def reset(self):
        self._previous_values[Na][:] = self._ext_vars[self._ligand.name()]
        self._update_changed_ext_val()
    
    def _update_changed_ext_val(self):
        from numpy import log
        R = UNIVERSAL_GAS      # Universal gas constant [J/(mole*K)]
        T = self.get_T()
        z = self._z      
        NaO = self._params.NaO
        area = self._area
        ms_per_s = 1 / MILLI

        # Prevent zero concentration resulting in a NaN in the numpy.log
        if self._ext_vars["Dyadic_cleft_Na"] > 0:
            Nai = self._ext_vars["Dyadic_cleft_Na"]
        else:
            # A small but not zero number
            Nai = 1e-9

        #raise RuntimeError,"FIXME: _previous_values"
        # Take the mean of the previous registered values
        self._previous_values[Na][-1] = Nai
        Nai = self._previous_values[Na].mean()
        debug("Nai : %s"%str(Nai))
        
        # The Nernst Potential
        ENa  = R*T/(z*FARADAY)*log(NaO/Nai) # V
        i_Na = -self._params["g_Na"] * (self._ext_vars["V"]*MILLI-ENa)    # pA
        debug("ENa : %s"%str(ENa))
        
        #print "i_Na:",i_Na," Na:",self._ext_vars["Dyadic_cleft_Na"]
        
        # And finally setting the actuall values of the ion fluxes,
        # corresponding to a general Robin BC
        self._dNa_dt = i_Na / (z * FARADAY * MICRO * ms_per_s ) * PICO
        self._c0 = self._dNa_dt / area * NM3_per_DM3
        self._alpha = 0.
        self._update_c0()
        debug("Nai flux (c0): %s"%str(self._c0))
        
    def dNa_dt(self):
        " Return"
        return self._dNa_dt
        
    
debug("%s loaded"%__name__)
