#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""A module to produce a 3D mesh of the dyadic cleft."""

__author__ = "Johan Hake (johan.hake@gmail.com)"
__copyright__ = "Copyright (C) 2008 Johan Hake"
__license__  = "GNU LGPL Version 3.0 or any later"
__date__ = "2007-10-10 -- 2010-01-18"

from tritetmesh.tetmesh import PLCEditor

import glob
import gzip
import numpy
import math

from numpy.random import random, permutation

from diffsim.parameterdict import *

__all__ = ['CleftMesh3D']

pi     = numpy.pi
arange = numpy.arange
array  = numpy.array
zeros  = numpy.zeros
ones   = numpy.ones
ceil   = numpy.ceil
floor  = numpy.floor
sin    = numpy.sin
cos    = numpy.cos

def circular_evendistr(N,R,no_remove=None,extend=10):

    if hasattr(no_remove,"__len__") and len(no_remove) == 0:
        no_remove = None
    
    assert(extend >= 1)
    def randomR(R0,N):
        " Return a random distribution of points"
        import numpy.random
        theta = numpy.random.random(N)*2*numpy.pi;
        R = numpy.sqrt(numpy.random.random(N))*R0;
        x0 = R*numpy.cos(theta);
        y0 = R*numpy.sin(theta);
        x0.shape = y0.shape = (N,1)
        return numpy.concatenate((x0,y0),1)

    # Create the random points
    points = randomR(R,N*extend)
    
    if not no_remove is None:
        #print "no_remove:",no_remove.shape,"points:",points.shape
        assert(isinstance(no_remove,numpy.ndarray))
        assert(len(no_remove.shape)==2 and no_remove.shape[1]==2)
        points = numpy.concatenate((no_remove,points),0)
        N_no_remove = no_remove.shape[0]
    else:
        N_no_remove = 0

    N_ext = points.shape[0]
    
    onevec = numpy.ones((1,N_ext))
    M      = numpy.zeros((N_ext,N_ext,2))
    MT     = numpy.zeros((N_ext,N_ext,2))
    for i in xrange(2):
        coord_vec = points[:,i]
        coord_vec.shape = (coord_vec.shape[0],1)
        M[:,:,i] =  coord_vec*onevec
        MT[:,:,i] = numpy.transpose(M[:,:,i])
        
    ds = numpy.triu(numpy.sqrt(((M-MT)**2).sum(2)))

    MAX = 10000
    ds[ds==0] = MAX
    
    total_Nout = N *(extend-1)
    
    indout = []
    indkeep = range(N_no_remove,N_ext)
    #print "Indkeep:",indkeep
    #print "total_Nout:",total_Nout
    while len(indkeep) > N:
        indmin = numpy.argmin(ds)
        colmin = indmin%N_ext
        if colmin >= N_no_remove:
            #print "Removing",colmin
            indkeep.remove(colmin)
        ds[:,colmin] = MAX
        ds[colmin,:] = MAX
    
    return points[indkeep,:]

def rectangle_collar(npoints, start_vertex = 0, offset = 0):
    """ Makes a collar of length npoints using rectangles
    
    @param npoints      : The number of points in one circle
    @param start_vertex : Defines the number of the first vertex in the collar.
    @param offset       : The differences in points between the two circles
    """
    for i in xrange(npoints):
        yield [i + start_vertex,
               (i+1) % npoints + start_vertex,
               ((i + npoints + 1 ) % npoints) + npoints + offset + start_vertex,
               i + npoints + offset + start_vertex]

def triangle_collar(npoints, start_vertex = 0, offset = 0):
    """ Makes a collar of length npoints using triangles
    
    @param npoints      : The number of points in one circle
    @param start_vertex : Defines the number of the first vertex in the collar.
    @param offset       : The differences in points between the two circles
    """
    for i in xrange(npoints):
        yield [[i + start_vertex,
                (i+1) % npoints + start_vertex,
                ((i + npoints + 1 ) % npoints) + npoints + offset + start_vertex],
               
               [i + start_vertex,
                i + npoints + offset + start_vertex,
                ((i + npoints + 1 ) % npoints) + npoints + offset + start_vertex] ]

def channel_collar(p0, channel_area, ds, height = 0.0, part = 1.0, weight = 2):
    """ Return three arrays of coordinate points, used to build a channel
    
    Given a center point, a channel area and an approximate circumfence distance
    this function returns three arrays describing a channel collar, with the given
    properties. The channel is situated in the xy plane.

    @param p0           : The coordinate of the center of the channel
    @param channel_area : The area of the channel
    @param ds           : An approximate distance of the length of the chords
                          circumfencing the channel
    @param height        : The height of the channel
    @param part         : If height is nonzero, then part gives the partial distance
                          in z coordinates between z0 and the height
    @param weight       : If height is nonzero the weigth times the ds adds to the
                          radius to give the radius of the channel base
    
    """
    
    x0, y0, z0 = p0
    R_tmp = math.sqrt(channel_area/pi)
    height_sign = 1 if height > 0 else -1
    #print "Height sign:",height_sign
    height = abs(height)
    if height != 0.:
        # If the channel is raised from the membrane we need to
        # have constraint on ds: FIXME: Why?
        if ds / R_tmp < 0.05:
            ds = 0.05*R_tmp
        small_radius = ds * weight/4
        z = part*height
        #print "small radius:",small_radius,"z:",z, "part:",part
        if z > height:
            z_loc = z-height
            offset = z_loc
        if z > height - small_radius:
            z_loc = z-(height - small_radius)
            #print "z_loc:",z_loc, "sqrt:",1-(z_loc/small_radius)**2
            offset = small_radius*math.sqrt(abs(1-(z_loc/small_radius)**2))
        elif z > small_radius:
            z_loc = z - small_radius
            offset = small_radius + weight * ds * (1-z_loc/(height-2*small_radius))
        else:
            offset = small_radius + weight * ds + small_radius*(1-z/small_radius)
    else:
        offset = 0.0
        
    dphi    = 2*pi/ceil(2*pi*R_tmp/ds)
    phi_vec = arange(0,2*pi, dphi, 'd')

    # Calculate the radius that give the exact right area
    N_seg = len(phi_vec)
    R     = math.sqrt(2*channel_area / (N_seg*sin(dphi))) + offset
    
    sin_phi = sin(phi_vec)
    cos_phi = cos(phi_vec)

    ret = (cos_phi*R + x0, sin_phi*R + y0, ones(N_seg,'d') * (z0 + part*height*height_sign))
    #print "Height of channel part,", part,",",ret[-1][0]
    return ret

class CleftMesh3D:
    def __init__(self):
        self._p = self.default_params()
        
        # Initialize coordinate arrays
        self._xvec = array([])
        self._yvec = array([])
        self._zvec = array([])

        self._editor = PLCEditor()

    @classmethod
    def default_params(cls):
        return ParameterDict(geom = ParameterDict(
                                R                = RangeCheck(100,gt=0),
                                h                = RangeCheck(15.,gt=0),
                                cleft_partitions = [],
                                vol_attributes   = [],
                                ),
                         
                         channel = ParameterDict(
                                ds               = RangeCheck(0.1,gt=0),
                                area             = RangeCheck(pi*1.**2,gt=0),
                                height            = RangeCheck(0.,ge=0),
                                weight           = RangeCheck(2.0,gt=0),
                                ),
                        
                         LCC = ParameterDict(
                                hard_coded_placement = [0,6,13], # The number of the opposing RyR...
                                marker_start     = 10,            
                                xyz_coord        = [(0.,0.,0.)],    
                                max_area         = None,             
                                smothing         = False,            
                                random_placement = True,     
                                RyR_per_LCC      = 5,             
                                ),
                  
                         RyR = ParameterDict(
                                marker_start = 100,
                                distribution = False,
                                num = 5,
                                dist = 31.,
                                height = 5.0,
                                use = True,
                                max_area = None,
                                ),

                         Na  = ParameterDict(
                                num = 0,
                                frac_radius = 0.4,
                                marker_start = 5,
                                max_dist_ncx = -1.0,
                                max_area = None,
                                ),
                         
                         NCX = ParameterDict(
                                use = False,
                                sigma = 400, # in #/um²
                                marker_start = 300,
                                frac_border = 0.7,
                                max_area = None,
                                random_placement = True,     
                                ),
                         
                         TT =  ParameterDict(
                                max_area = None,
                                marker = 1,
                                ),
                         
                         SR =  ParameterDict(
                                max_area = None,
                                marker = 2,
                                ),
                         
                         Cytosole = ParameterDict(
                                marker = 3,
                                N_points = 32,
                                max_area = None,
                                extension = ParameterDict(
                                     use = False,
                                     R = 200,
                                     N_points = 16),
                                ),
    
                         points = ParameterDict(
                                marker_start = 200,
                                xyz_coord = [],
                                ),
    
                         tritetmesh   = ParameterDict(
                                save_tetgen   = False,
                                name          = "cleft_mesh",
                                radius_edge_ratio    = 1.8,
                                max_vol       = None,
                                compress      = True,
                                )
                         )

    def get_params(self):
        return self._p

    def clear(self):
        self._xvec = array([])
        self._yvec = array([])
        self._zvec = array([])
        self._editor.clear()

    def add_boundary_and_partitions(self):
        # Start with the TT facet and add the facets together with the coordinates 
        # for each part in cleft_partitions.

        p = self._p

        dtheta = 2*pi/p.Cytosole.N_points
        theta = arange(0,2*pi,dtheta,'d')
        sin_theta = sin(theta)
        cos_theta = cos(theta)

        # Clear facet_numbers
        self._facet_numbers = {}
        
        # Fill up the self._cleft_partions
        self._cleft_partitions = p.geom.cleft_partitions[:]
        self._cleft_partitions.insert(0,0.)
        self._cleft_partitions.append(1.0)
        #print "cleft_partitions:",self._cleft_partitions
        
        Cytosole_max_area = p.Cytosole.max_area or -1.0
        for part in self._cleft_partitions:
            
            # Register the number of added vertices
            nrv = self._num_registered_vertices()
            
            # If we are on the SR facet
            if part != 1.0:
                # If not on the SR facet add cytosolic collar facets

                # Set markers if not using Cytosole extension
                m = 0 if p.Cytosole.extension.use else p.Cytosole.marker
                for polygon in rectangle_collar(p.Cytosole.N_points, nrv):
                    self._editor.add_facet(polygon, m, Cytosole_max_area)
                
                # If on the T-tubule facet
                if part == 0.0:
                    facet_marker   = p.TT.marker   or 0
                    facet_max_area = p.TT.max_area or -1.0
                
                # Else we are on a shadow facet
                else:
                    facet_marker   = 0
                    facet_max_area = -1.0
                
            else:
                # If we are on the SR facet
                facet_marker   = p.SR.marker   or 0
                facet_max_area = p.SR.max_area or -1.0
                
            # Add facet
            self._editor.add_facet(arange(p.Cytosole.N_points) + nrv,
                             facet_marker, facet_max_area)
                
            # Add the coordinates for cytosolic vertices
            self._add_coordinates((cos_theta * p.geom.R,
                                   sin_theta * p.geom.R,
                                   ones(p.Cytosole.N_points,'d') * p.geom.h * part))

            if not p.Cytosole.extension.use:
                # Add any markers for the cytosolic vertices
                self._editor.add_vertex_marker(p.Cytosole.marker,
                                               arange(p.Cytosole.N_points) + nrv)
            
            # Store the facet number
            self._facet_numbers[part] = self._editor.num_facets()-1

    def add_cytosolic_extension(self):
        p = self._p

        arc = 2*pi
        dtheta = arc/p.Cytosole.N_points
        theta = arange(0, arc, dtheta, 'd')
        sin_theta = sin(theta)
        cos_theta = cos(theta)

        arc = pi/2
        dphi = 0.5*pi/p.Cytosole.extension.N_points
        phi = arange(0, arc, dphi, 'd') + dphi
        Rs = sin(phi)*p.Cytosole.extension.R + p.geom.R
        heights = (1-cos(phi))*p.Cytosole.extension.R + p.geom.h
        
        for i, Rh in enumerate(zip(Rs, heights)):
            nrv = self._num_registered_vertices()
            
            for p1, p2 in triangle_collar(p.Cytosole.N_points, nrv - p.Cytosole.N_points):
                self._editor.add_facet(p1)
                self._editor.add_facet(p2)
            
            R, h = Rh
            self._add_coordinates((cos_theta*R,
                                   sin_theta*R,
                                   ones(p.Cytosole.N_points,'d') * h))
        
        nrv = self._num_registered_vertices()
        self._editor.add_facet(arange(p.Cytosole.N_points)+nrv)
        self._editor.add_polygon_to_last_facet(arange(p.Cytosole.N_points)+nrv-p.Cytosole.N_points, (0, 0, h))
        
        self._add_coordinates((cos_theta*(R + p.geom.h),
                               sin_theta*(R + p.geom.h),
                               ones(p.Cytosole.N_points,'d') * h))
        
        for polygon in rectangle_collar(p.Cytosole.N_points, nrv):
            self._editor.add_facet(polygon, p.Cytosole.marker, p.Cytosole.max_area or -1.0)
        nrv += p.Cytosole.N_points
        
        self._editor.add_facet(arange(p.Cytosole.N_points)+nrv)
        self._editor.add_polygon_to_last_facet(arange(p.Cytosole.N_points), (0, 0, 0))
        
        self._add_coordinates((cos_theta*(R + p.geom.h),
                               sin_theta*(R + p.geom.h),
                               zeros(p.Cytosole.N_points,'d')))

    def add_channel(self, facet, p0, marker, max_area = None, area = None,
                    ds = None, height = None, weight = None):
        p = self._p
        
        max_area = max_area or -1
        area     = area     or  p.channel.area
        ds       = ds       or  p.channel.ds
        height    = height if not height is None else p.channel.height
        weight   = weight   or  p.channel.weight
        
        assert(facet in self._facet_numbers.values())
        assert(marker >= 0)
        assert(area > 0)
        assert(ds > 0)

        # assert that we are on the TT facet if we have a height 
        if height > 0:
            assert(facet == self._facet_numbers[0])

        if height < 0:
            assert(facet == self._facet_numbers[1.0])
        
        def add_collar(channel_facet, part=-1):
            # FIXME: This function is very unintuitive!!!!
            
            # Get the coordinates for the channel facet
            if channel_facet:
                channel_part = 1.0
            else:
                if height > 0:
                    channel_part = part * p.geom.h / height
                else:
                    channel_part = 0.0

            #print "channel_part:",channel_part
            # Get the coordinates of the collar
            collar_coords = channel_collar(p0, area, ds, height, channel_part, weight)
            num_collar_vertices = len(collar_coords[0])
            nrv = self._num_registered_vertices()
            #print "collar coords:",collar_coords
            vertices = arange(num_collar_vertices) + nrv

            # If we have a height and part == 0 or if we do not have height and
            # we are adding the channel facet
            if (not channel_facet) or (channel_facet and height == 0.0):
                # Register the hole in the facet
                if channel_facet:
                    hole_in_facet = facet
                else:
                    hole_in_facet = self._facet_numbers[part]
                    
                #print "hole_in_facet:",hole_in_facet,":", p0
                self._editor.add_polygon_to_facet(vertices, hole_in_facet)
                self._editor.add_facet_hole(p0, hole_in_facet)

            if not channel_facet:
                #print "registering the slope"
                # Register the slope collar facets
                for p1, p2 in triangle_collar(num_collar_vertices, nrv):
                    self._editor.add_facet(p1)
                    self._editor.add_facet(p2)
            else:
                #print "adding facet with marker:", marker
                # Add the facet and marker information to the editor
                self._editor.add_facet(vertices, marker, max_area)
                self._editor.add_vertex_marker(marker, vertices)
            
            # Register the coordinates 
            self._add_coordinates(collar_coords)

        # Add the different collars depending on the 
        if height != 0.0:
            # FIXME: Add possible smoothing?
            # If no extra cleft partions is added.
            if len(self._cleft_partitions[:-1]) == 2:
                add_collar(True)
            else:
                if height > 0:
                    for part in self._cleft_partitions[:-1]:
                        add_collar(False, part)
                else:
                    add_collar(False,1.0)
                
                add_collar(True)
                
        else:
            # Add the channel facet
            add_collar(True)
                
    def add_RyR(self):
        p = self._p
        
        b = 2*p.geom.R/math.sqrt(2)
        Nx_RyR = max(int(floor(b/p.RyR.dist)),1)
        offset = (b-Nx_RyR*p.RyR.dist)/2 + p.RyR.dist/2 - b/2
        RyR_p0 = []
        print "Num RyR:",Nx_RyR*Nx_RyR
        for i in xrange(Nx_RyR):
            for j in xrange(Nx_RyR):
                RyR_p0.append([offset+i*p.RyR.dist,offset+j*p.RyR.dist,p.geom.h])
        self._RyR_p0 = array(RyR_p0)
        #print SR_facet, RyR_p0
        # FIXME: Put this into a generic add channel function

        for i, p0 in enumerate(self._RyR_p0):
            self.add_channel(self._facet_numbers[1.0], p0, p.RyR.marker_start + i, p.RyR.max_area, height = -p.RyR.height)
            
    
    def add_LCC(self):

        p = self._p
        # Add information of any channel to the facets
        max_area   = p.LCC.max_area or -1.0

        if p.LCC.random_placement:
            N_LCC_RANDOM = int(floor(len(self._RyR_p0)/p.LCC.RyR_per_LCC))
            print "Num LCC:",N_LCC_RANDOM
            # NOT so random after all... :)
            ind_LCC = permutation(len(self._RyR_p0))[:N_LCC_RANDOM]
            #ind_LCC = array(p.LCC.hard_coded_placement)
            self._LCC_p0 = self._RyR_p0[ind_LCC,:]
            self._LCC_p0[:,-1] = 0.0
        else:
            self._LCC_p0 = p.LCC.xyz_coord
        for i, p0 in enumerate(self._LCC_p0):
            # If the channel has height add corresponding holes and collar facets to the editor
            if p.channel.height != 0.0:
                # Iterate over the horisontal partitions in the cleft
                # First add the 0 partition, i.e., the TT facet
                part = 0
                    
                collar_coords = channel_collar(p0, p.channel.area, p.channel.ds,
                                    p.channel.height, part*p.geom.h/p.channel.height, p.channel.weight)
                
                num_collar_vertices = len(collar_coords[0])
                nrv = self._num_registered_vertices()
                
                # Register the hole in the facet
                self._editor.add_polygon_to_facet(arange(num_collar_vertices) + nrv, self._facet_numbers[part])
                self._editor.add_facet_hole(p0, self._facet_numbers[part])
                
                # Register the slope collar facets
                for p1, p2 in triangle_collar(num_collar_vertices, nrv):
                    self._editor.add_facet(p1)
                    self._editor.add_facet(p2)
                    
                # Register the coordinates 
                self._add_coordinates(collar_coords)
                
                if p.LCC.smothing:
                    # Add smothing for the border close to the TT
                    for part in array([0.5,1.0])*(p.channel.ds*p.channel.weight/4)/p.channel.height:
                        
                        collar_coords = channel_collar(p0, p.channel.area, p.channel.ds,
                                            p.channel.height, part,p.channel.weight)
                        num_collar_vertices = len(collar_coords[0])
                        nrv = self._num_registered_vertices()
                        
                        # Register the hole in the facet
                        #editor.add_polygon_to_facet(arange(num_collar_vertices) + nrv, self._facet_numbers[part])
                        #editor.add_facet_hole(p0, self._facet_numbers[part])
                        
                        # Register the slope collar facets
                        for p1, p2 in triangle_collar(num_collar_vertices, nrv):
                            self._editor.add_facet(p1)
                            self._editor.add_facet(p2)
                            
                        # Register the coordinates 
                        self._add_coordinates(collar_coords)
                    
                for part in self._cleft_partitions[1:-1]:
                    # p0, p.channel.area, ds, height = 0.0, part = 0.0, weight = 2
                    
                    collar_coords = channel_collar(p0, p.channel.area, p.channel.ds,
                                        p.channel.height, part*p.geom.h/p.channel.height,p.channel.weight)
                    num_collar_vertices = len(collar_coords[0])
                    nrv = self._num_registered_vertices()
                    
                    # Register the hole in the facet
                    self._editor.add_polygon_to_facet(arange(num_collar_vertices) + nrv, self._facet_numbers[part])
                    self._editor.add_facet_hole(p0, self._facet_numbers[part])
                    
                    # Register the slope collar facets
                    for p1, p2 in triangle_collar(num_collar_vertices, nrv):
                        self._editor.add_facet(p1)
                        self._editor.add_facet(p2)
                        
                    # Register the coordinates 
                    self._add_coordinates(collar_coords)
                
                if p.LCC.smothing:
                    # Add smothing for the border close to the TT
                    for part in ((p.channel.height-p.channel.ds*p.channel.weight/4 + array([0.1,0.4,0.7])*p.channel.ds*p.channel.weight/4))/p.channel.height:
                        collar_coords = channel_collar(p0, p.channel.area, p.channel.ds,
                                            p.channel.height, part,p.channel.weight)
                        num_collar_vertices = len(collar_coords[0])
                        nrv = self._num_registered_vertices()
                        
                        # Register the hole in the facet
                        #editor.add_polygon_to_facet(arange(num_collar_vertices) + nrv, self._facet_numbers[part])
                        #editor.add_facet_hole(p0, self._facet_numbers[part])
                        
                        # Register the slope collar facets
                        for p1, p2 in triangle_collar(num_collar_vertices, nrv):
                            self._editor.add_facet(p1)
                            self._editor.add_facet(p2)
                            
                        # Register the coordinates 
                        self._add_coordinates(collar_coords)
                    
            # Get the coordinates for the channel facet
            collar_coords = channel_collar(p0, p.channel.area, p.channel.ds, p.channel.height)
            
            nrv = self._num_registered_vertices()
            num_collar_vertices = len(collar_coords[0])
            
            if p.channel.height == 0.0:
                # Register the hole in the facet
                self._editor.add_polygon_to_facet(arange(num_collar_vertices) + nrv, self._facet_numbers[0])
                self._editor.add_facet_hole(p0, self._facet_numbers[0])
            
            # Add the facet and marker information to the editor
            self._editor.add_facet(arange(num_collar_vertices) + nrv,
                             p.LCC.marker_start + i, max_area)
            
            self._editor.add_vertex_marker(p.LCC.marker_start + i, arange(num_collar_vertices) + nrv)
            
            # Register the coordinates 
            self._add_coordinates(collar_coords)


    def add_NCX(self):
        p = self._p
        
        if p.Na.num < 1:
            raise TypeError, "If including NCX also include Na"
        
        #assert(len(self._cleft_partitions)==2), "Cannot have cleft partitions together with Na and NCX, for now..."
        # We will here calculate positions of any included NCX and Na
        # stochastically.
        num_NCX = round(p.NCX.sigma*p.geom.R**2*pi*1e-6)
        print "Num NCX:",num_NCX
        xy_LCC = numpy.array([[x,y] for x,y,z in self._LCC_p0])
        if not p.NCX.random_placement:
            assert(num_NCX + self._LCC_p0.shape[0]<=self._RyR_p0.shape[0])
            # NOT so random after all... :)
            ind_NCX = list(permutation(len(self._RyR_p0)))
            for ind_LCC in p.LCC.hard_coded_placement:
                ind_NCX.remove(ind_LCC)
            ind_NCX = ind_NCX[:num_NCX]
            xy_NCX = self._RyR_p0[ind_NCX,:-1]
            N_new_points = p.Na.num
            max_radius_for_random_points = p.geom.R*p.Na.frac_radius
            no_remove = numpy.concatenate((xy_NCX,xy_LCC))
            extend = 30
        else:
            max_radius_for_random_points = p.geom.R*p.NCX.frac_border
            N_new_points = num_NCX + p.Na.num
            no_remove = xy_LCC
            extend = 10

        Na_num = p.Na.num
        
        # If not random place them a max dist from an NCX
        enough_in_the_middle = False
        while not enough_in_the_middle:
            xy_random = circular_evendistr(N_new_points,max_radius_for_random_points,no_remove=no_remove,extend=extend)            
            
            # Pick an area in the midle of the cleft where we want to put the
            # Na and the NCX channels
           
            R_inside = p.Na.frac_radius*p.geom.R
            ind_random_middle = (numpy.sqrt((xy_random**2).sum(1))<R_inside).nonzero()[0]
            if len(ind_random_middle) == 0:
                continue
            enough_in_the_middle = True
            #ind_LCC_middle = (numpy.sqrt((xy_LCC**2).sum(1))<R_inside).nonzero()[0] # Not used!?
        
        # Convert random picked NCX points to Na points
        ind_Na_in_middle_ind = permutation(len(ind_random_middle))[:Na_num]
        ind_Na = ind_random_middle[ind_Na_in_middle_ind]
        xy_Na = xy_random[ind_Na]
        if p.NCX.random_placement:
            ind_NCX = range(len(xy_random))
            for ind in ind_Na:
                ind_NCX.remove(ind)
            
            xy_Na = xy_random[ind_Na]
            xy_NCX = xy_random[ind_NCX]
        #else:
        #    print "xy_random", xy_random
        #    print "xy_Na", xy_Na
        #    assert((xy_random == xy_Na).all()) # Note does not work if they are rearranged
        
        # If using maximal distance betwen Na and NCXs
        if p.Na.max_dist_ncx > 0.0:
            if p.Na.max_dist_ncx < 2*numpy.sqrt(p.channel.area/numpy.pi):
                raise TypeError, "Minimal distance between Na and NCX is smaller than twice the radius of a channel"
            dist_Na = []
            min_dist = p.geom.R
            ind_min_Na  = 0
            ind_min_NCX = 0
            for i in xrange(p.Na.num):
                # FIXME: Occasionally this raise line raise an
                # IndexError: index out of bounds
                # Check and fix!
                dist = numpy.sqrt(((xy_NCX - xy_Na[i])**2).sum(1))
                if dist.min() < min_dist:
                    ind_min_Na  = i
                    ind_min_NCX = dist.argmin()
                    min_dist    = dist.min()
            if min_dist > p.Na.max_dist_ncx:
                dir_vec = xy_NCX[ind_min_NCX] - xy_Na[ind_min_Na]
                unit_dir_vec = dir_vec/numpy.sqrt((dir_vec**2).sum())
                xy_Na[ind_min_Na] = xy_NCX[ind_min_NCX] + unit_dir_vec*p.Na.max_dist_ncx
                print "Marker Na:", p.Na.marker_start + ind_min_Na
                print "Marker NCX:", p.NCX.marker_start + ind_min_NCX
                
        # Add the information of the Na channels
        for i, xy in enumerate(xy_Na):
            # Get the coordinates for the channel facet
            p0 = (xy[0],xy[1],0.0)
            self.add_channel(self._facet_numbers[0], p0, p.Na.marker_start + i, p.Na.max_area)
            
        # Add the information of the NCX channels
        for i, xy in enumerate(xy_NCX):
            p0 = (xy[0],xy[1],0.0)

            self.add_channel(self._facet_numbers[0], p0, p.NCX.marker_start + i, p.NCX.max_area)
           

    def create_mesh(self, name):
        self.clear()
        
        p = self._p

        # Check the given channel height
        if p.channel.height > 0:
            # If using height partitions
            if len(p.geom.cleft_partitions) > 0 and p.channel.height < p.geom.cleft_partitions[-1]*p.geom.h:
                # pick the largest partition and compare with the channel height
                raise RuntimeError, "Specify a channel height that is larger than the largest cleft partition"
            if p.channel.height > p.geom.h:
                raise RuntimeError, "Specify a channel height that is smaller than the cleft height"

        # Compute the boundaries and any height partitions and add them to the editor
        self.add_boundary_and_partitions()

        print "num vertices after boundaries", len(self._xvec)

        # If using Cytosolic extensions
        if p.Cytosole.extension.use:
            self.add_cytosolic_extension()

        print "num vertices after extensions", len(self._xvec)
        
        # Register any RyR
        if p.RyR.use:
            print "Adding RyR"
            self.add_RyR()

        print "num vertices after RyR", len(self._xvec)
        # Register any LCC, these relates to the RyR so they need to be added after these
        print "Adding LCC"
        self.add_LCC()
        
        print "num vertices after LCC", len(self._xvec)
        # If any NCXs 
        if p.NCX.use > 0:
            print "Adding NCX and Na"
            self.add_NCX()

        # Add volume attributes
        for xyz, max_vol, attr in p.geom.vol_attributes:
            self._editor.add_region(xyz,attr,max_vol)
        
        # Add points with markes to the vertices
        # (Note: This can impare the final result)
        if p.points.xyz_coord != []:
            nrv = self._num_registered_vertices()
            new_x = numpy.zeros(len(p.points.xyz_coord))
            new_y = new_x.copy()
            new_z = new_x.copy()
            for i, xyz in enumerate(p.points.xyz_coord):
                new_x[i],new_y[i],new_z[i] = xyz
                self._editor.add_vertex_marker(p.points.marker_start+i,
                                         [nrv + i])
            self._add_coordinates((new_x, new_y, new_z))
        
        # Register the final vertices
        self._editor.register_vertices(*self._registerd_vertex_coordinates())
        self._editor.set_radius_edge_ratio( p.tritetmesh.radius_edge_ratio)
        if not p.tritetmesh.max_vol is None:
            self._editor.set_uniform_volume_constraint(p.tritetmesh.max_vol)

        # Save the plc, for debug information
        if p.tritetmesh.save_tetgen:
            self._editor.save_plc(name)
        if 1:
            self._editor.build(name)
            if p.tritetmesh.compress:
                from subprocess import Popen
                filenames = glob.glob(name + '*.xml')
                processes = []
                for fn in filenames:
                    #processes(Popen("gzip %s"%fn,shell = True))
                    Popen("gzip -f %s"%fn,shell = True)
            
    def _add_coordinates(self,coord):
        " Add a tuple coordinates to the stored vertex coordinates" 
        assert(isinstance(coord,tuple))
        assert(len(coord) == 3)
        assert(len(coord[0]) == len(coord[1]) and len(coord[0]) == len(coord[2]))
        
        self._xvec = numpy.concatenate((self._xvec,coord[0]))
        self._yvec = numpy.concatenate((self._yvec,coord[1]))
        self._zvec = numpy.concatenate((self._zvec,coord[2]))

    def _num_registered_vertices(self):
        " Return the number of registered vertices" 
        return len(self._xvec)

    def _registerd_vertex_coordinates(self):
        " Return the registered vertex coordinates"
        return self._xvec, self._yvec, self._zvec
        
if __name__ == '__main__':
    cm = CleftMesh3D()
    print cm.get_params()
