__author__ = "Johan Hake (johan.hake@gmail.com)"
__copyright__ = "Copyright (C) 2008 Johan Hake"
__license__  = "GNU LGPL Version 3.0 or any later"
__date__ = "2008-09-09 -- 2010-01-25"

from diffsim.base          import check_domain_name, DiffsimObject, check_ion_name
from diffsim.boundaries    import Flux, ConcDepLinearFluxBoundary
from diffsim.dolfingallery import dim_to_str
from diffsim.common        import VALENCE
from diffsim.parameterdict import TypeCheck, RangeCheck, OptionCheck, inf, ParameterDict

from dolfin                import DOLFIN_EPS, Expression

from diffsim.logger import *

import numpy

__all__ = ['StaticField','CytosolicStaticFieldFlux']

init_local_logger(__name__,globals())
set_log_level(INFO,__name__)

class CytosolicStaticFieldFlux(ConcDepLinearFluxBoundary):
    """ This class provides a Cytosolic flux based on static field present in the dyad
    
    """
    def __init__(self, static_field, ion_name, marker, u0, dn = 1.0):
        if not isinstance(static_field, StaticField):
            raise TypeError, "Provide a 'StaticField' for the static_field parameter"
        if not isinstance(u0,(int,float)):
            raise TypeError, "Provide a 'float' for 'u0'"
        # Get the domain
        domain = static_field.domain()
        z_vec = domain.mesh().coordinates()[:,-1][domain.get_vertex_indices(marker)]
        check_ion_name(ion_name)
        efield_params = static_field.get_params()
        u0 = u0*numpy.exp(-VALENCE[ion_name]*efield_params.phi_0*numpy.exp(-efield_params.kappa*z_vec))
        super(CytosolicStaticFieldFlux,self).__init__(
            "CytosolicStaticFieldFlux_"+ion_name,
            domain.name(), ion_name, marker, u0, dn)
        
        # Register the instantiated object
        self._register_DiffsimObject(CytosolicStaticFieldFlux)
        
class StaticField(DiffsimObject):
    """ A class that implements a static electric field in the cleft"""
    def __init__(self, name, domain_name, kappa = 1.0):
        
        super(StaticField,self).__init__(name)
        domain = check_domain_name(domain_name)

        # Create the electric field
        
        # Note double negation!
        # FIXME: What double negation?
        efield_expr = ("0.0","0.0","phi_0*valence*kappa*D*exp(-kappa*x[2])")
        efield      = Expression(efield_expr)
        
        efield.kappa   = kappa
        efield.phi_0   = 1.0
        
        # Ligand specific parameters
        efield.valence = 1.0 # Use a dummy valence and D of 1 for this base matrix
        efield.D       = 1.0
        
        A_adv = domain.assemble_tensor("advection",\
                                       coefficients = [efield],\
                                       return_tensor = True)
        
        # The advection matrix is substracted from the ordinary stiffnes matrix
        #A_adv *= -1
        
        # Store protected variables
        self._efield       = efield
        self._A_adv        = A_adv
        
        # Initialize the FieldFluxes
        self._field_fluxes = {}
        self._params.kappa       = kappa
        for ligand in domain.iter_ligands():
            self._field_fluxes[ligand.ion_name()] = StaticFieldFlux(ligand,self)

        self._domain = domain
        
        # Register the instantiated object
        self._register_DiffsimObject(StaticField)

    @classmethod
    def default_params(self):
        p = ParameterDict(phi_0       = TypeCheck(-2.2),
                          kappa       = RangeCheck(1.0,gt=0,lt=100),
                          SUPG_type   = OptionCheck(-1,[-1,0,1,2,3]),
                          tau         = RangeCheck(1.0,gt=0,le=1))
        return p

    def domain(self):
        " Return the domain the static field is defined on"
        return self._domain
    
    def A_adv(self):
        " Return a copy of the original advection matrix"
        return self._A_adv.copy()

    def efield(self):
        return self._efield

class StaticFieldFlux(Flux):
    """ A class that implements a StaticFieldFlux
    
    This class is used by a StaticField to create the actuall flux that is attached
    to a diffusive ligand
    """
    
    def __init__(self, ligand, static_field):
        """ Instantiate a StaticFieldFlux

        @param ligand       : The actual ligand the flux is defined for
        @param static_field : The StaticField the flux will follow
        """
        name = "%s_%s_flux"%(static_field.name(),ligand.ion_name())
        super(StaticFieldFlux,self).__init__(name,ligand.domain().name(),\
                                             ligand.ion_name())
        
        # The contribution from the flux makes the total flux matrix non symetric
        self._symetric = False
        
        self._valence = VALENCE[ligand.ion_name()]
        self._static_field = static_field
        self._ligand = ligand
        self._A      = self._static_field.A_adv()
        self._A.zero()
        self._SUPG_cache = {}
        
        # A dummy ext_vars to create a dependency on the static_field
        #self._ext_vars[static_field.name()] = 0
        
        # Register the instantiated object
        self._register_DiffsimObject(StaticFieldFlux)

    def minimize_from_start(self):
        return True

    def reset(self):
        
        # Activate the flux
        self.activate()
        
        valence  = self._valence
        D        = self._ligand.D()
        phi_0    = self._static_field.get_params().phi_0

        debug("Reseting %s: D: %f, phi_0: %f"%(self.name(),D,phi_0))
        # Initialize the matrix
        self._A.zero()
        self._A.axpy(1.0,self._static_field.A_adv(),True)
        self._A *= valence * D *phi_0
        
        # Check for SUPG
        SUPG_type = self._static_field.get_params().SUPG_type
        if  SUPG_type < 0:
            return

        # Global tau
        tau      = self._static_field.get_params().tau

        # Check if the matrix has been assembled before
        unique_SUPG_str = "SUPG type: %d, D: %e, phi_0: %f"%\
                          (SUPG_type,D,phi_0)


        # Check if the matrix has been assembled before
        if self._SUPG_cache.has_key(unique_SUPG_str):
            debug("Using SUPG cache")
            self._A.axpy(tau,self._SUPG_cache[unique_SUPG_str],True)
            return 
        
        efield      = self._static_field.efield()
        domain      = self._ligand.domain()
        mesh        = domain.mesh()
        
        # Initialize the efield with ligand specific parameters
        efield.D       = D
        efield.phi_0   = phi_0
        efield.valence = valence

        def _file_string(file_name):
            from os import path
            return open(path.join(path.dirname(__file__),"cpp",file_name)).read()
        
        ffc_options    = {"representation":"quadrature",
                          "cpp optimize":True}
        
        # Test of different implementations of the SUPG method.
        if SUPG_type in [0,1]:
            
            # Add coefficients functions
            coefficients = [efield]
            raise NotImplemented, "Remove SUPG_type != 3"
            #coefficients += [MeshSize(mesh)]
            
            coefficients.append(Expression(_file_string("FieldDirection.h")))
            # Using different elements for Stab
            if SUPG_type == 0:
                V = domain.scalar0()
            else:
                V = domain.scalar()
            coefficients.append(Expression(_file_string("Stab.h")))
                
            # Set attributes in the two functions
            coefficients[-2].field = efield
            coefficients[-1].field = efield
            coefficients[-1].D     = D
            
            if SUPG_type == 0:
                A_SUPG_adv = domain.assemble_tensor("SUPG advection0",\
                                                    coefficients = coefficients,\
                                                    reassemble    = True,
                                                    return_tensor = True)
                coefficients.pop(0)
                M_SUPG = domain.assemble_tensor("SUPG mass0",\
                                                coefficients = coefficients,
                                                reassemble    = True,
                                                return_tensor = True)
            else:
                A_SUPG_adv = domain.assemble_tensor("SUPG advection1",\
                                                    coefficients = coefficients,\
                                                    reassemble    = True,
                                                    return_tensor = True)
                coefficients.pop(0)
                M_SUPG = domain.assemble_tensor("SUPG mass1",\
                                                coefficients = coefficients,
                                                reassemble    = True,
                                                return_tensor = True)
                
        elif SUPG_type in[2,3]:
            coefficients = [efield]
            # Using different elements for Stab2
            if SUPG_type == 2:
                V = domain.vector0()
            else:
                V = domain.vector()
            f = Expression(_file_string("Stab2.h"))
            coefficients.append(f)
            coefficients[-1].field = efield
            coefficients[-1].D     = D
            
            if SUPG_type == 2:
                A_SUPG_adv = domain.assemble_tensor("SUPG advection2",\
                                                    coefficients = coefficients,\
                                                    reassemble    = True,
                                                    return_tensor = True)
                coefficients.pop(0)
                M_SUPG = domain.assemble_tensor("SUPG mass2",\
                                                coefficients = coefficients,
                                                reassemble    = True,
                                                return_tensor = True)
            else:
                A_SUPG_adv = domain.assemble_tensor("SUPG advection3",\
                                                    coefficients = coefficients,\
                                                    reassemble    = True,
                                                    return_tensor = True)
                coefficients.pop(0)
                M_SUPG = domain.assemble_tensor("SUPG mass3",\
                                                coefficients = coefficients,
                                                reassemble    = True,
                                                return_tensor = True)
        
        self._SUPG_cache[unique_SUPG_str] = A_SUPG_adv
        debug("SUPG Method %d, tau: %f"%(SUPG_type,tau))
        self._A.axpy(tau,A_SUPG_adv,True)
        
        # Reset the efield
        efield.valence = 1.0
        efield.D       = 1.0
        efield.phi_0   = 1.0
    
    def add_to_flux_matrix(self,A):
        A.axpy(1.0,self._A,True)

    
debug("%s loaded"%__name__)
