# -*- coding: utf-8 -*-
__author__ = "Johan Hake (johan.hake@gmail.com)"
__copyright__ = "Copyright (C) 2008 Johan Hake"
__license__  = "GNU LGPL Version 3.0 or any later"
__date__ = "2008-09-04 -- 2010-01-14"

__all__ = ['LCCMarkovModel_Greenstein','LCCMarkovModel_Hake','RyRMarkovModel_Stern',
           'RyRMarkovModel_Dura','RyRMarkovModel_Polakova','NaMarkovModel_Clancy']

from diffsim.stochastictimestepper import DiscreteMarkovChain, DiscreteMarkovModel, mul
from diffsim.parameterdict         import (ParameterDict, TypeCheck,
                                           RangeCheck, OptionCheck, ParConstant)
from diffsim.common                import T_ABS

from diffsim.logger import *
init_local_logger(__name__,globals(),INFO)

import math
mexp = math.exp
mlog = math.log

def lcc_rate_func_generator_Greenstein():
    """This generator yields a two field tuple.
    
    The first field is a connection information.
    It is simply a tuple of two states, (from,to).
    
    The second field is a function that can be called with two parameters,
    both dictionaries. The first dictionary contains constant parameters,
    p, and the second dependent variables, var.
    """
    
    # A base string for the rate function
    rate_func_base_str = """
def _rate_func_%d(p, var):
    return %s
"""
    # Two base strings for voltage dependent forward and backward rates
    alpha_str = "p['T_scale']*%f*2*mexp(0.012*(var['V']-35))%s"
    beta_str = "p['T_scale']*%f*0.0882*mexp(-0.05*(var['V']-35))%s"
    for i in xrange(28):
        if i < 8:
            if i < 4:
                func_str   = alpha_str%((4-i),"")
                connection = (i,i+1)
            else:
                func_str   = alpha_str%((8-i),"*p['a']")
                connection = (i+2,i+3)
        elif i < 16:
            if i < 12:
                func_str   = beta_str%((i-7),"")
                connection = (i-7,i-8)
            else:
                func_str   = beta_str%((i-11),"/p['b']")
                connection = (i-5,i-6)
        elif i < 21:
            func_str   = "p['T_scale']*p['gamma0']*p['orig_Ca']/p['ref_Ca']*p['a']**%d*var['Dyadic_cleft_Ca']*p['scale']"%(i-16)
            #func_str   = "p['T_scale']*p['gamma0']*p['a']**%d*var['Dyadic_cleft_Ca']*p['scale']"%(i-16)
            connection = (i-16,i-10)
        elif i < 26:
            func_str   = "p['T_scale']*p['omega']/(p['b']**%d)"%(i-21)
            connection = (i-15,i-21)
        elif i < 27:
            func_str   = "p['T_scale']*p['f']"
            connection = (4,5)
        else:
            func_str   = "p['T_scale']*p['g']"
            connection = (5,4)
        exec(rate_func_base_str%(i,func_str))
        
        yield connection, eval("_rate_func_%d"%i)

def lcc_rate_func_generator_Hake():
    """This generator yields a two field tuple.
    
    The first field is a connection information.
    It is simply a tuple of two states, (from,to).
    
    The second field is a function that can be called with two parameters,
    both dictionaries. The first dictionary contains constant parameters,
    p, and the second dependent variables, var.
    """
    
    # A base string for the rate function
    rate_func_base_str = """
def _rate_func_%d(p, var):
    return %s
"""
    # Two base strings for voltage dependent forward and backward rates
    alpha_str = "%f*p['T_scale']*1.7/(1+mexp(-(var['V']-12)/12.))%s"
    beta_str = "%f*p['T_scale']*0.4*(mexp(-(var['V']+40)/20.)+0*mexp(-(var['V']+100)/7.5))%s"
    for i in xrange(28):
        if i < 8:
            if i < 4:
                func_str   = alpha_str%((4-i),"")
                connection = (i,i+1)
            else:
                func_str   = alpha_str%((8-i),"*p['a']")
                connection = (i+2,i+3)
        elif i < 16:
            if i < 12:
                func_str   = beta_str%((i-7),"")
                connection = (i-7,i-8)
            else:
                func_str   = beta_str%((i-11),"/p['b']")
                connection = (i-5,i-6)
        elif i < 21:
            func_str   = "p['T_scale']*p['gamma0']*p['orig_Ca']/p['ref_Ca']*p['a']**%d*var['Dyadic_cleft_Ca']*p['scale']"%(i-16)
            connection = (i-16,i-10)
        elif i < 26:
            func_str   = "p['T_scale']*p['omega']/(p['b']**%d)"%(i-21)
            connection = (i-15,i-21)
        elif i < 27:
            func_str   = "p['T_scale']*p['f']"
            connection = (4,5)
        else:
            func_str   = "p['T_scale']*p['g']"
            connection = (5,4)
        exec(rate_func_base_str%(i,func_str))
        
        yield connection, eval("_rate_func_%d"%i)

def lcc_y_state_rates_Greenstein():
    # Import the math version of exp in the local namespace
    from math import exp
    def y_inf(var):
        return 0.4/(1+ exp(var['V']+12.5)/5)+0.6
    
    def y_tau(var):
        return 340/(1+ exp(var['V']+30)/12)+60
    
    def rate_func_0(p,var):
        return p['T_scale']*y_inf(var)/y_tau(var)
    
    def rate_func_1(p,var):
        return p['T_scale']*(1-y_inf(var))/y_tau(var)
    
    return {(0,1):rate_func_0,(1,0):rate_func_1}

class LCCMainGate_Greenstein(DiscreteMarkovChain):
    _rate_func_map = dict(lcc_rate_func_generator_Greenstein())
    _params = ParameterDict(scale   = RangeCheck(2.0,gt = 0),
                            f       = RangeCheck(0.85 ,gt = 0),
                            g       = RangeCheck(2.0,gt = 0),
                            a       = RangeCheck(2.0,gt = 0),
                            b       = RangeCheck(1.9356,gt = 0),
                            gamma0  = RangeCheck(4.4e-4,gt = 0),
                            orig_Ca = ParConstant(15.),       # This value is the original [Ca]_ss level for one open LCC after
                                                              # a step to 0 mV.
                            ref_Ca  = RangeCheck(300.,gt = 0),# This value needs to be set by the user and corresponds to the [Ca] level at LCC after a step to 0 mV
                            omega   = RangeCheck(0.02158,gt = 0),
                            T_scale = RangeCheck(1.0,gt = 0, lt = 100),
                            )
    _ext_vars = {
        "V":[-90,(0,1,2,3,4,6,7,8,9,10)],
        "Dyadic_cleft_Ca":[0.1,(0,1,2,3,4)],
        }
    
    _state_values = {5:1}

class LCCMainGate_Hake(DiscreteMarkovChain):
    _rate_func_map = dict(lcc_rate_func_generator_Hake())
    _params = ParameterDict(scale   = RangeCheck(1.0,gt = 0),
                            f       = RangeCheck(1.5 ,gt = 0),
                            g       = RangeCheck(2.0,gt = 0),
                            a       = RangeCheck(2.0,gt = 0),
                            b       = RangeCheck(1.9356,gt = 0),
                            gamma0  = RangeCheck(4.4e-4,gt = 0),
                            orig_Ca = ParConstant(15.),       # This value is the original [Ca]_ss level for one open LCC after
                                                              # a step to 0 mV.
                            ref_Ca  = RangeCheck(300.,gt = 0),# This value needs to be set by the user and corresponds to the [Ca] level at LCC after a step to 0 mV
                            omega   = RangeCheck(0.02158,gt = 0),
                            T_scale = RangeCheck(1.0,gt = 0, lt = 100),
                            )
    _ext_vars = {
        "V":[-90,(0,1,2,3,4,6,7,8,9,10)],
        "Dyadic_cleft_Ca":[0.1,(0,1,2,3,4)],
        }
    _state_values = {5:1}

class LCCInactivationGate_Greenstein(DiscreteMarkovChain):
    _rate_func_map = lcc_y_state_rates_Greenstein()
    _params = ParameterDict(T_scale = RangeCheck(1.0,gt = 0, lt = 100))
    _ext_vars = {"V":[-90,(0,1)],}
    _state_values = {1:1}

class LCCMarkovModel_Greenstein(DiscreteMarkovModel):
    """ LCC markov model

    Model and rated taken from:
    Greenstein, J. L. & Winslow, R. L.
    An integrative model of the cardiac ventricular myocyte incorporating local control of Ca$^2+$ release.
    Biophys. J., 2002, 83, 2918-2945

    Based on the model from:
    
    Jafri, M. S.; Rice, J. J. & Winslow, R. L.
    Cardiac Ca$^2+$ dynamics: the roles of ryanodine receptor adaptation
    and sarcoplasmic reticulum load.
    Biophys. J., 1998, 74, 1149-68

    The rates are all fitted for 37°C
    
    """
    _DiscreteMarkovChains = [LCCMainGate_Greenstein,LCCInactivationGate_Greenstein]
    _dmc_combine = mul
    _params = ParameterDict(T_ref = ParConstant(37+T_ABS),
                            Q10   = RangeCheck(1.8,ge=1,le=5))

class LCCMarkovModel_Hake(DiscreteMarkovModel):
    """ LCC markov model

    This model is based on the model from:    

    Greenstein, J. L. & Winslow, R. L.
    An integrative model of the cardiac ventricular myocyte incorporating local control of Ca$^2+$ release.
    Biophys. J., 2002, 83, 2918-2945

    which also is based on the model from:
    
    Jafri, M. S.; Rice, J. J. & Winslow, R. L.
    Cardiac Ca$^2+$ dynamics: the roles of ryanodine receptor adaptation
    and sarcoplasmic reticulum load.
    Biophys. J., 1998, 74, 1149-68

    The alpha and beta functions are fitted to the data presented in:

    Poláková, E.; Zahradníková, A.; Pavelková, J.; Zahradník, I. & Zahradníková, A.
    Local calcium release activation by DHPR calcium channel openings in rat cardiac myocytes.
    J Physiol 2008, 586, 3839-3854
    
    Especially the low beta rate at -40 and -80 mV letting the channel reopen.
    The two nP0 values for the two voltage steps to 60 mV, one for 1.5 ms and one
    for 5.0 ms made me implement a constant activation rate and deactivation rate.
    Also the sharp transition from close to open while pertaining the
    
    The rates are all fitted for 22°C
    
    """
    _DiscreteMarkovChains = [LCCMainGate_Hake,LCCInactivationGate_Greenstein]
    _dmc_combine = mul
    _params = ParameterDict(T_ref = ParConstant(22+T_ABS),
                            Q10   = RangeCheck(1.8,ge=1,le=5))

def RyR_Stern_rates():
    def alpha(p,var):
        return p['T_scale']*p["kO"]*(var["Dyadic_cleft_Ca"]*p["scale"])**2
    def beta(p,var):
        return p['T_scale']*p["kOM"]
    def gamma(p,var):
        return p['T_scale']*p["kIM"]
    def delta(p,var):
        return p['T_scale']*p["kI"]*var["Dyadic_cleft_Ca"]*p["scale"]
    return {(0,1):alpha,(1,0):beta,(2,3):alpha,(3,2):beta,
            (0,2):delta,(2,0):gamma,(1,3):delta,(3,1):gamma
            }

class RyRMarkovChain_Stern(DiscreteMarkovChain):
    _rate_func_map = RyR_Stern_rates()
    _params = ParameterDict(scale   = RangeCheck(2.0 , gt = 0),
                            kI      = RangeCheck(5e-4, gt = 0),
                            kIM     = RangeCheck(5e-3, gt = 0),
                            kO      = RangeCheck(3.5e-5, gt = 0),
                            kOM     = RangeCheck(0.06, gt = 0),
                            T_scale = RangeCheck(1.0,gt = 0, lt = 100))
    _ext_vars = {"Dyadic_cleft_Ca":[0.1,(0,1,2)]}
    _state_values = {1:1}

class RyRMarkovModel_Stern(DiscreteMarkovModel):
    """ Markov model for a Ryanodine receptor

    Model and rates taken from:

    Stern, M. D.; Song, L. S.; Cheng, H.; Sham, J. S.; Yang, H. T.; Boheler, K. R. & Ríos, E.
    Local control models of cardiac excitation-contraction coupling.
    A possible role for allosteric interactions between ryanodine receptors.
    J. Gen. Physiol., 1999, 113, 469-489 
    """
    _DiscreteMarkovChains = [RyRMarkovChain_Stern]
    _dmc_combine = mul
    _params = ParameterDict(T_ref = ParConstant(37+T_ABS),
                            Q10   = RangeCheck(2.,ge=1,le=5))


class RyRMarkovChain_Dura(DiscreteMarkovChain):
    _rate_func_map = {
        (0,1):lambda p,var:p['T_scale']*4*p["kon"]*var["Dyadic_cleft_Ca"],
        (1,2):lambda p,var:p['T_scale']*3*p["kon"]*var["Dyadic_cleft_Ca"],
        (2,3):lambda p,var:p['T_scale']*2*p["kon"]*var["Dyadic_cleft_Ca"],
        (3,4):lambda p,var:p['T_scale']*p["kon"]*var["Dyadic_cleft_Ca"],
        (1,0):lambda p,var:p['T_scale']*p["koff"],
        (2,1):lambda p,var:p['T_scale']*2*p["koff"],
        (3,2):lambda p,var:p['T_scale']*3*p["koff"],
        (4,3):lambda p,var:p['T_scale']*4*p["koff"],
        
        (4,5):lambda p,var:p['T_scale']*p["kC4O1"],
        (5,4):lambda p,var:p['T_scale']*p["kO1C4"],

        (4,6):lambda p,var:p['T_scale']*p["kC4O2"],
        (6,4):lambda p,var:p['T_scale']*p["kO2C4"],

        (7,5):lambda p,var:p['T_scale']*p["kC5O1"],
        (5,7):lambda p,var:p['T_scale']*p["kO1C5"],

        (7,6):lambda p,var:p['T_scale']*p["kC5O2"],
        (6,7):lambda p,var:p['T_scale']*p["kO2C5"],

        (7,8):lambda p,var:p['T_scale']*p["kC5I"],
        (8,7):lambda p,var:p['T_scale']*p["kIC5"]
        }
    _params = ParameterDict(scale   = RangeCheck(2.0 , gt=0),
                            kon     = RangeCheck(0.712,gt=0),
                            koff    = RangeCheck(3.,gt=0),
                            kC4O1   = RangeCheck(10.,gt=0),
                            kC4O2   = RangeCheck(1.e-3,gt=0),
                            kO1C4   = RangeCheck(0.5,gt=0),
                            kO2C4   = RangeCheck(5.e-4,gt=0),
                            kO1C5   = RangeCheck(2.e-3,gt=0),
                            kO2C5   = RangeCheck(3.,gt=0),
                            kC5O1   = RangeCheck(6.666e-4,gt=0),
                            kC5O2   = RangeCheck(0.1,gt=0),
                            kC5I    = RangeCheck(5.e-4,gt=0),
                            kIC5    = RangeCheck(1.5e-3,gt=0),
                            T_scale = RangeCheck(1.0,gt = 0, lt = 100))
    _ext_vars = {"Dyadic_cleft_Ca":[0.1,(0,1,2,3)]}
    _state_values = {5:1,6:1}

class RyRMarkovModel_Dura(DiscreteMarkovModel):
    """ Markov model for a Ryanodine receptor

    States in the model:
    0 ←→ 1 ←→ 2 ←→ 3 ←→ 4 ←→ 5
                             ↑↓   ↑↓
                              6 ←→ 7 ←→ 8

    C0 ←→ C1 ←→ C2 ←→ C3 ←→ C4 ←→ O1
                                 ↑↓    ↑↓
                                  O2 ←→ C5 ←→ I

    State 5 and 6 are the open states.

    Model and rates taken from:

    Dura, M.; Zahradník, I. & Zahradníková,
    A. Kinetics of cardiac RyR channel gating studied at high temporal resolution.
    Physiol. Res. 2003, 52, 571-578
    """
    
    _DiscreteMarkovChains = [RyRMarkovChain_Dura]
    _dmc_combine = mul
    _params = ParameterDict(T_ref = ParConstant(37+T_ABS),
                            Q10   = RangeCheck(2.,ge=1,le=5))

class RyRMarkovChain_PolakovaOld(DiscreteMarkovChain):
    _rate_func_map = {
        (0,1):lambda p,var:p['T_scale']/p["tau_act"]*\
                           4*var["Dyadic_cleft_Ca"]/p["ref_conc"],
        (1,2):lambda p,var:p['T_scale']/p["tau_act"]*\
                           3*var["Dyadic_cleft_Ca"]/p["ref_conc"],
        (2,3):lambda p,var:p['T_scale']/p["tau_act"]*\
                           2*var["Dyadic_cleft_Ca"]/p["ref_conc"],
        (3,4):lambda p,var:p['T_scale']/p["tau_act"]*\
                           var["Dyadic_cleft_Ca"]/p["ref_conc"],
        (1,0):lambda p,var:p['T_scale']  *p["koff"],
        (2,1):lambda p,var:p['T_scale']*2*p["koff"],
        (3,2):lambda p,var:p['T_scale']*3*p["koff"],
        (4,3):lambda p,var:p['T_scale']*4*p["koff"],
        }
    
    _params = ParameterDict(koff     = RangeCheck(0.0,ge=0),
                            tau_act  = ParConstant(0.62),
                            ref_conc = RangeCheck(1500.,gt=0),
                            T_scale  = RangeCheck(1.0,gt = 0, lt = 100))
    
    _ext_vars = {"Dyadic_cleft_Ca":[0.1,(0,)]}
    _state_values = {4:1}

class RyRMarkovChain_Polakova(DiscreteMarkovChain):
    _rate_func_map = {
        (0,1):lambda p,var:(p['scale']*p['T_scale']/p["tau_act"]*var["Dyadic_cleft_Ca"]/p["ref_conc"])**p['num_sub_units'],
        (1,0):lambda p,var:(p['T_scale'] *p["koff"])**p['num_sub_units'],
        }
    
    _params = ParameterDict(scale    = RangeCheck(1.,gt=0),
                            koff     = RangeCheck(200000.0,ge=0),
                            tau_act  = ParConstant(0.62),
                            num_sub_units = OptionCheck(4,range(1,5)),
                            ref_conc = RangeCheck(1500.,gt=0),
                            T_scale  = RangeCheck(1.0,gt = 0, lt = 100))
    
    _ext_vars = {"Dyadic_cleft_Ca":[0.1,(0,)]}
    _state_values = {1:1}

class RyRMarkovModel_Polakova(DiscreteMarkovModel):
    """ Markov model for a simplyfied Ryanodine receptor
    
    States in the model:
    0 ←→ 1 ←→ 2 ←→ 3 ←→ 4 ←→ 5
    
    State 5 is the open state

    Model and rates taken from:
    
    Poláková, E.; Zahradníková, A.; Pavelková, J.; Zahradník, I. & Zahradníková, A.
    Local calcium release activation by DHPR calcium channel openings in rat cardiac myocytes.
    J Physiol 2008, 586, 3839-3854
    
    In this study they measure the on rate to be: k_act = 1/tau_act = 1/0.62 ms.
    This rate can be disentangled into a konstant, kon, times the local [Ca]. The [Ca] is different
    for different geometries and efield configurations and kon therefore needs to be fitted.
    
    """
    
    _DiscreteMarkovChains = [RyRMarkovChain_Polakova]
    _dmc_combine = mul
    _params = ParameterDict(T_ref = ParConstant(22+T_ABS),
                            Q10   = RangeCheck(2.,ge=1,le=5))

def Na_Clancy_rates():
    #alpha_str = """def alpha%s(p,var):
    #return 3.802/(0.1027*mexp(-var["V"]/%s)+%s*mexp(-var["V"]/150.0))
    #"""
    #args = [("11","17.0","0.2"),
    #        ("12","15.0","0.23"),
    #        ("13","12.0","0.25"),]
    #
    #for arg in args:
    #    print alpha_str%arg
    #
    #beta_str = """def beta%s(p,var):
    #return %s*mexp(-(var["V"]-%s)/20.3)
    #"""
    #args = [("11","0.1917","0.0"),
    #        ("12","0.20","5.0"),
    #        ("13","0.22","10.0"),]
    #
    #for arg in args:
    #    print beta_str%arg
    
    def alpha11(p,var):
        return p['T_scale']*3.802/(0.1027*mexp(-var["V"]/17.0)+0.2*mexp(-var["V"]/150.0))
    
    def alpha12(p,var):
        return p['T_scale']*3.802/(0.1027*mexp(-var["V"]/15.0)+0.23*mexp(-var["V"]/150.0))
    
    def alpha13(p,var):
        return p['T_scale']*3.802/(0.1027*mexp(-var["V"]/12.0)+0.25*mexp(-var["V"]/150.0))
    
    def beta11(p,var):
        return p['T_scale']*0.1917*mexp(-(var["V"]-0.0)/20.3)
    
    def beta12(p,var):
        return p['T_scale']*0.20*mexp(-(var["V"]-5.0)/20.3)
    
    def beta13(p,var):
        return p['T_scale']*0.22*mexp(-(var["V"]-10.0)/20.3)
    
    def alpha2(p,var):
        return p['T_scale']*9.178*mexp(var["V"]/29.68)
    
    def beta2(p,var):
        return alpha13(p,var)*alpha2(p,var)*alpha3(p,var)/((beta13(p,var)*beta3(p,var)))
    
    def alpha3(p,var):
        return p['T_scale']*(3.7933e-7*mexp(-var["V"]/7.7))
    
    def beta3(p,var):
        return p['T_scale']*(0.0084 + 0.00002*var["V"])
    
    def alpha4(p,var):
        return alpha2(p,var)/100.0
    
    beta4 = alpha3
    
    def alpha5(p,var):
        return alpha2(p,var)/9.5e4
    
    def beta5(p,var):
        return alpha3(p,var)/50.0
    
    return {(5,6) : alpha11,
            (6,7) : alpha12,
            (7,8) : alpha13, 
            (0,1) : alpha11,
            (1,2) : alpha12, 
            (6,5) : beta11,
            (7,6) : beta12,
            (8,7) : beta13,
            (1,0) : beta11,
            (2,1) : beta12,
            (8,2) : alpha2,
            (2,8) : beta2,
            (2,7) : alpha3,
            (1,6) : alpha3,
            (0,5) : alpha3,
            (7,2) : beta3,
            (6,1) : beta3,
            (5,0) : beta3,
            (2,3) : alpha4,
            (3,2) : beta4,
            (3,4) : alpha5,
            (4,3) : beta5}

class NaMarkovChain_Clancy(DiscreteMarkovChain):
    _rate_func_map = Na_Clancy_rates()
    _params = ParameterDict(T_scale = RangeCheck(1.0,gt = 0, lt = 100))
    _ext_vars = {"V":[-70.0,tuple(xrange(9))]}
    _state_values = {8:1}

class NaMarkovModel_Clancy(DiscreteMarkovModel):
    """ Markov model for a transient Na channel

    States in the model:
     0 - 1 - 2 - 3 - 4
     |   |   | \
     5 - 6 - 7 - 8

     IC3 - IC2 - IF - IM1 - IM2
      |     |    |  \
     C3  - C2  - C1 -  O

    State 8 is an open state.

    Model and rates taken from:

    Clancy, C. E. & Rudy, Y.
    Na+ channel mutation that causes both Brugada and long-QT syndrome phenotypes:
    a simulation study of mechanism.
    Circulation, 2002, 105, 1208-1213
    
    """
    _DiscreteMarkovChains = [NaMarkovChain_Clancy]
    _dmc_combine = mul
    _params = ParameterDict(T_ref = ParConstant(37+T_ABS),
                            Q10   = RangeCheck(2.,ge=1,le=5))

debug("%s loaded"%__name__)
