# -*- coding: utf-8 -*-
__author__ = "Johan Hake (johan.hake@gmail.com)"
__copyright__ = "Copyright (C) 2008 Johan Hake"
__license__  = "GNU LGPL Version 3.0 or any later"
__date__ = "2008-11-30 -- 2009-01-24"

from diffsim.base            import (DiffsimObject, ContinuousVariable,
                                     check_domain_name, Q10_scale)
from diffsim.boundaries      import LinearFluxBoundary, FluxBoundary
from diffsim.common          import (VALENCE, Ca, Na, FARADAY, UNIVERSAL_GAS,
                                     MILLI, MICRO, PICO, MICRO, NM3_per_DM3, T_ABS)
from diffsim.parameterdict   import (TypeCheck, RangeCheck, OptionCheck,
                                     inf, ParameterDict)
from diffsim.diffusiveligand import (DiffusiveLigand, MEAN, MIN, MAX, STD,
                                     VAR, ALLOWED_STAT)

from diffsim.logger import *

import numpy

__all__ = ['NCX','NCXFlux']

init_local_logger(__name__,globals())
set_log_level(INFO,__name__)

class NCX(ContinuousVariable):
    """ A class that defines the Na/Ca exchanger current"""
    _first = None
    def __init__(self, domain_name, marker):
        # Init base class
        super(NCX,self).__init__("NCX_m%d"%marker)
        
        # Get the domain
        domain = check_domain_name(domain_name)
        
        # Check the marker
        domain.check_face_marker(marker)
        
        # Check registered DiffusiveLigands in the domain
        registerd_ions = [l.ion_name() for l in domain.iter_ligands()]
        assert(Ca in registerd_ions)
        assert(Na in registerd_ions)
        
        self._marker = marker

        # Register the domain, ligand, and external values for Na, Ca and V
        self._domain = domain
        self._ion_fluxes = {}
        self._ligands    = {}
        
        for ligand in domain.iter_ligands():
            if ligand.ion_name() in [Na, Ca]:
                self._ligands[ligand.ion_name()] = ligand
                self._ion_fluxes[ligand.ion_name()] = NCXFlux(ligand,self,self._marker)

        self._ext_vars = {domain.get_ligand(Na).name():self._ligands[Na].value(marker),
                          domain.get_ligand(Ca).name():self._ligands[Ca].value(marker),
                          "V":-75.0}
        
        npv = ContinuousVariable.get_solver_params().num_previous_values
        self._previous_values = {
            Na:numpy.array([self._ext_vars[self._ligands[Na].name()]]*npv),
            Ca:numpy.array([self._ext_vars[self._ligands[Na].name()]]*npv),
            'amp':numpy.array([0.0]*npv)}

        # A place holder for saveing the current
        self._current_values = []
        
        # Included to mimic Flux boundary interface, as we are reusing
        # its receive function, which uses self._activator
        self._activator = None

        self._amp_scale    = 1.0
        self._V_scale      = FARADAY*MILLI/(UNIVERSAL_GAS*self.get_T())
        
        # Protected attributes from ContinuousVariable
        self._values_0      = 0.0
        self._values_1      = 0.0
        self._values_2      = 0.0

        if NCX._first is None:
            NCX._first = marker
        
        # Register the instantiated object
        self._register_DiffsimObject(NCX)
        
    def reset(self):
        # Store the parameter dict as an ordinary dict
        from math import exp
        
        # Get the Na and Ca concentrations
        Nai = self._ext_vars[self._domain.get_ligand(Na).name()]
        Cai = self._ext_vars[self._domain.get_ligand(Ca).name()]

        num_previous_values = ContinuousVariable.get_solver_params().num_previous_values
        self._previous_values[Na][:] = Nai
        self._previous_values[Ca][:] = Cai

        # Build the namespace that will be used to execute the current statement
        self._current_ns = self._params.NCX.copy(to_dict=True)
        
        self._current_ns['Nai']         = Nai
        self._current_ns['Cai']         = Cai
        self._current_ns['exp']         = exp
        self._current_ns['result']      = 0.0
        self._current_ns['numerator']   = 0.0
        self._current_ns['denominator'] = 0.0
        
         # Constant scale factor (From A/F to pA)
        self._current_ns['amp_scale']   = self._params.C/self._params.sigma
        self._current_ns['T_scale']     = 1.0
        self._current_ns['fudge_scale'] = 1.0
        
        # Tune the NCX current so amp is 2.e-4 pA (0.2fA) see Hilgemann Biophysical Journal 1996(71):759
        # at [CaO] = 0mM, [NaO] = 40mM, [Cai] = 4mM, [Nai] = 40mM and V=0mV, at 35°C
        self._current_ns['CaO'] = 1.e-9
        self._current_ns['NaO'] = 40.e3
        self._V_scale = FARADAY*MILLI/(UNIVERSAL_GAS*(T_ABS+35.))
        self._current_ns['fudge_scale'] = -2.e-4/self._compute_current(40.e3,4.e3,0.0)

        # Reset the parameters after setting 'fudge_scale'
        self._V_scale               = FARADAY*MILLI/(UNIVERSAL_GAS*(self.get_T()))
        self._current_ns['T_scale'] = Q10_scale(self._params.Q10,self._params.T_ref)
        self._current_ns['CaO']     = self._params.NCX.CaO
        self._current_ns['NaO']     = self._params.NCX.NaO
        
        # Calculate the current
        amp = self._compute_current(Nai,Cai,self._ext_vars["V"])
        
        # Store the amp in a list keeping track of previous values
        self._previous_values['amp'][:] = amp
        
        debug("Temperature scaling for NCX: %s"%str(self._current_ns['T_scale']))
        debug("Fudge scaling for NCX: %s"%str(self._current_ns['fudge_scale']))
              
        # Update the fluxes
        self._ion_fluxes[Ca].update_flux(amp)
        self._ion_fluxes[Na].update_flux(amp)

        if self._params.save_current:
            self._current_values.append([])
            self._current_values[-1].append(amp)
        
    @classmethod
    def default_params(cls):
        return ParameterDict(
            save_current= OptionCheck(0,[0,1]),
            ligand_stat = OptionCheck(MEAN,list(ALLOWED_STAT)),
            tol_abs     = RangeCheck(0.01,gt=0),
            C           = RangeCheck(0.01, gt=0),  # pF/um^2
            Q10         = RangeCheck(2.50,gt=0),
            T_ref       = RangeCheck(T_ABS+35,ge=T_ABS+35,le=T_ABS+35),
            sigma       = RangeCheck(400.0,gt=0), # N/um^2
            Kd_Act      = RangeCheck(0.256,gt=0),# uM
            NCX         = ParameterDict(NaO    = RangeCheck(140.0e3,gt=0),
                                        CaO    = RangeCheck(2.0e3,gt=0),
                                        Vmax   = RangeCheck(9.0,gt=0),    # A/F
                                        KmCai  = RangeCheck(3.59,gt=0),   # uM
                                        KmCaO  = RangeCheck(1.3e3,gt=0),  # uM
                                        KmNai  = RangeCheck(12.29e3,gt=0),# uM
                                        KmNaO  = RangeCheck(87.5e3,gt=0), # uM
                                        ksat   = RangeCheck(0.27,gt=0),
                                        eta    = RangeCheck(0.35,ge=0,le=1),
                                        HNa    = RangeCheck(3,gt = 0),
                             )
            )

    _params = 0
    _current_code = compile("""
numerator    = exp(eta*V_scaled)*Nai**HNa*CaO
numerator   -= exp((eta-1)*V_scaled)*NaO**HNa*Cai
numerator   *= Ka*Vmax*amp_scale*T_scale*fudge_scale
denominator  = KmCai*NaO**HNa*(1+(Nai/KmNai)**HNa)
denominator += KmNai**HNa*CaO*(1+Cai/KmCai)
denominator += KmNaO**HNa*Cai
denominator += KmCaO*Nai**HNa
denominator += Nai**HNa*CaO
denominator += NaO**HNa*Cai
denominator *= 1+ksat*exp((eta-1)*V_scaled)
result       = numerator/denominator
""","","exec")

    def _compute_current(self,Nai=0.0,Cai=0.0,V=0.0):
        """ Return the NCX current in pA.
        
        Based on NCX model from:(wrong Q10)
        Shannon, T. R.; Wang, F.; Puglisi, J.; Weber, C. & Bers, D. M.
        A mathematical treatment of integrated Ca dynamics within
        the ventricular myocyte.
        Biophys. J., 2004, 87, 3351-3371
        
        For the temperature dependency look in:(NOT!)
        Puglisi, J. L., W. Yuan, J. W. M. Bassani, and D. M. Bers. 1999.
        Ca inﬂux through Ca channels in rabbit ventricular myocytes
        during action potential clamp: inﬂuence of temperature.
        Circ. Res. 85:e7–e16.
        """
        # NOTE: A positive current is letting 3 Na+ out 
        # while letting 1 Ca2+ in to the cell.
        # Update the namespace dictionary
        assert Cai>0 and Nai>0, "Concentrations need to be positive"
        self._current_ns['Cai'] = Cai
        self._current_ns['Nai'] = Nai
        self._current_ns['Ka'] = 1/(1+(self._params.Kd_Act/Cai)**3)
        self._current_ns['V_scaled'] = V*self._V_scale
        
        exec(self._current_code,{},self._current_ns)
        
        return self._current_ns['result']
    
    def finalize(self):
        from numpy import array
        if self._params.save_current:
            for i, values in enumerate(self._current_values):
                self._current_values[i] = array(values)
            return self._current_values
        
        return None

    def solve(self, dt, iteration=0):
        " Updates the current (not dt dependent)"
        
        # Update the different currents
        Cai = self._ext_vars[self._domain.get_ligand(Ca).name()]
        Nai = self._ext_vars[self._domain.get_ligand(Na).name()]

        # Register new values of Ca and Na
        # Do not allow negative concentrations
        self._previous_values[Ca][-1] = max(Cai,1e-9)
        self._previous_values[Na][-1] = max(Nai,1e-9)
        
        # Take the mean of the previous registered values
        Cai = self._previous_values[Ca].mean()
        Nai = self._previous_values[Na].mean()
        
        # Calculate the current
        amp = self._compute_current(Nai,Cai,self._ext_vars["V"])
        
        # Update the fluxes
        self._ion_fluxes[Ca].update_flux(amp)
        self._ion_fluxes[Na].update_flux(amp)
        
        if self._marker == NCX._first:
            #debug("Whole cell current NCX_m72: %s"%str(amp/scale))
            debug("NCX_m%d current: %s"%(self._marker,str(amp)))
            debug("V: %s"%str(self._ext_vars["V"]))
            debug("Nai: %s"%str(Nai))
            debug("Cai: %s"%str(Cai))
            debug("Na flux (c0): %s"%str(self._ion_fluxes[Na]._c0))
            debug("Ca flux (c0): %s"%str(self._ion_fluxes[Ca]._c0))
            debug("Kd_Act: %s"%str(1/(1+(self._params.Kd_Act/Cai)**3)))
        
        # Store the current
        self._values_1 = amp
        
        # Check for convergens
        if iteration >= 1:
            e = self._values_1-self._values_2
            # FIXME: Should probably combine with relative convergens here.
            if  e < self._solver_params.tol_abs:
                self._converged = True
                #info("%s fix point iteration solved in: %d iterations with error: %e"%\
                #     (self._name,self._fix_point_it, e))
        else:
            # Save the value from the first iteration...
            if self._params.save_current:
                self._current_values[-1].append(amp)
    
    # Reuse FluxBoundaries receive function
    receive = FluxBoundary.receive.im_func
    
    def active_variables(self):
        " The active variables of the ncx is its self! (FIXME: LOGIC!)"
        return [self]
        
class NCXFlux(LinearFluxBoundary):
    def __init__(self, ligand, ncx, marker):
        """ Instantiate a NCXFlux
        
        @param ligand  : The actual ligand the flux is defined for
        @param ncx     : The NCX instance that defines the current
        @param marker  : The marker of the boundary where the flux wille be applied
        """
        name = "%s_%s_%d_flux"%(ncx.name(),ligand.ion_name(),marker)
        super(NCXFlux,self).__init__(name,ligand.domain().name(),\
                                     ligand.ion_name(),marker)
        
        # If the ligand is Na we register a factor of -3, meaning that for each e going out
        # We loose 3 Na ions. The same for Ca but with reverse sign and different number
        self._factor    = -3 if ligand.ion_name() == Na else 1
        self._ncx       = ncx
        self._nonlinear = True
        self._alpha = 0.
        self._c0 = 0.
        
        # Register the instantiated object
        self._register_DiffsimObject(NCXFlux)

    def reset(self):
        " Ensure that the flux is activated"
        self._alpha = 0.
        self._c0 = 0.
        self.activate(True)
        
    _params = 0
    
    def update_flux(self,amp):
        """ Update the c0

        Called by the parent NCX instant.
        """
        F = FARADAY*MICRO        # Faradays constant
        nm3_per_dm3 = 1e24       # Misc. conversion constants
        ms_per_s = 1 / MILLI
        
        # And finally setting the actuall values of the ion fluxes,
        # corresponding to a general Robin BC
        self._dion_dt = amp * self._factor / (F * ms_per_s ) * PICO
        self._c0 = self._dion_dt / self._area * NM3_per_DM3
        self._update_c0()

    def dion_dt(self):
        " Return the ion rate in uM/ms"
        return self._dion_dt
    
debug("%s loaded"%__name__)
