__author__ = "Johan Hake (johan.hake@gmail.com)"
__copyright__ = "Copyright (C) 2008 Johan Hake"
__license__  = "GNU LGPL Version 3.0 or any later"
__date__ = "2008-08-19 -- 2010-01-29"

# Defining the import * statement
__all__     = ['Plotter',
               'ValueLogger',
               'PointLogger',
               'SCREEN', 'FILE', 'file_to_str'
               ]

# Importing diffsim stuff
from diffsim.base            import DiffsimObject
from diffsim.diffusiveligand import DiffusiveLigand, MEAN, ALLOWED_STAT
from diffsim.parameterdict   import ParameterDict, TypeCheck, RangeCheck, OptionCheck
from diffsim.disk            import save

# Initialize the logging system
from diffsim.logger import *
init_local_logger(__name__,globals(),INFO)

# Importing plotter
from viper import Viper

def file_to_str(filename):
    " Help function to a cpp file to a string" 
    from os import path
    return open(path.join(path.dirname(__file__), "cpp", filename)).read()
        

class Plotter(DiffsimObject):
    """ A simple plotter for DiffsimObjects
    """
    _count = 0
    def __init__(self, var_name, marker = None, plot_range = None, elevate = None,
                 azimuth = None, interactive = True, save_png = False, save_eps = False,
                 cutplane_origin = None, cutplane_normal = (0,0,1), below_zero = False,
                 spheres = None, sphere_radius = 2.0):
        """ Instantiate the plotter

        @param var_name   : The name of the logged variable (Must be a Diffusive ligand)
        @param marker     : An optional integer defining a subdomain that are going
                            to be plotted
        @param plot_range : An optional argument defining the plot range of the plot
        @param elevate    : Elevates the plot camera
        @param azimuth    : ???
        @param interative : Turn on interactive mode if True after a simulation
        @param save_png   : Save a png file of the plotter each turn, if scitools is
                            installed on the system a movie is also created
        @param save_eps   : Save a png file of the plotter each turn, if scitools is
                            installed on the system a movie is also created
        @param cutplane_origin : Coordinates for a point defining the origin of a
                                cut plane
        @param cutplane_normal : A vector defining the normal of a cut plane
        @param below_zero      : If true plot below zero values vs above zero values
        @param spheres         : A list with coordinates for spheres
        @param sphere_radius   : Radius for the added spheres
        """
        obj = DiffsimObject.get_object(var_name)
        if not isinstance(obj,DiffusiveLigand):
            raise TypeError, "Can only plot results from a diffusive ligand"
        if obj is None:
            raise AttributeError, "No DiffsimObject with name '%s' excists"%var_name
        name_list = [obj.name()]
        if not marker is None:
            name_list.extend(["-",str(marker)])
        name_list.append("-plotter")
        super(Plotter,self).__init__("".join(name_list) + "_%d"%Plotter._count)
        Plotter._count += 1
        
        if not hasattr(obj,"domain"):
            raise RuntimeError, "Can only plot a continuous variable that resides in a domain."
        self._ext_vars = {}.fromkeys([var_name])
        self._marker = marker
        self._interactive = interactive
        self._save_png = save_png
        self._save_eps = save_eps
        self._below_zero = below_zero
        
        mesh = obj.domain().mesh()
        if marker is None:
            self._values = obj.value().array()
        else:
            raise NotImplementedError, "Plotting of subdomains is not implemented"
            self._values = obj.value(marker)
        
        
        if self._below_zero:
            self._below_zero_values = (self._values < 0)*1.0
            self._viper = Viper(mesh = mesh, x = self._below_zero_values,
                                vmin = 0, vmax = 1,
                                cutplane_origin = cutplane_origin,
                                cutplane_normal = cutplane_normal)
        
        elif plot_range is None:
            self._viper = Viper(mesh = mesh, x = self._values,
                            cutplane_origin = cutplane_origin,
                            cutplane_normal = cutplane_normal)
        else:
            self._viper = Viper(mesh = mesh, x = self._values,
                            vmin = plot_range[0], vmax = plot_range[1],
                            cutplane_origin = cutplane_origin,
                            cutplane_normal = cutplane_normal)
        
        if not elevate is None:
            self._viper.elevate(elevate)
            
        if not azimuth is None:
            self._viper.azimuth(azimuth)
        
        if not spheres is None:
            assert isinstance(spheres,list)
            for point in spheres:
                assert len(point)==3, "Provide a list of points with length 3"
                self._viper.add_sphere(point,sphere_radius)
        
        self._viper.show_scalarbar()
        self._viper.update()

        self._base_name = None
        
        # Register the instantiated object
        self._register_DiffsimObject(Plotter)

        # Enable inspection before simulation
        if self._interactive:
            self._viper.interactive()
    
    def reset(self):
        if self._base_name is None:
            self.set_case_name()
        self._viper.update()
        
    def set_case_name(self,case_name = "simulation"):
        import os
        var_name = [ s for s in self._ext_vars.iterkeys()
                     if s != "time"][0]
        self._basename = "%s_%s"%(case_name,var_name)
        if os.path.isdir("figures"):
            self._basename = os.path.join("figures",self._basename)
            
        if self._save_png or self._save_eps:
            self._viper.init_writer(self._basename)
            
    def receive(self, sender):
        # If the sender is a DiffusiveLigand, then it wont send a value but it self!
        if not isinstance(sender,DiffusiveLigand):
            raise RuntimeError, "Can only plot results from a diffusive ligand"
        if not self._marker is None:
            raise NotImplementedError, "Plotting of subdomains is not implemented"
            self._values[:] = sender.value(self._markers)
        
        if self._below_zero:
            self._below_zero_values[:] = (self._values < 0)*1.0
        else:
            self._values[:] = sender.value().array()
        
        #debug("Viper update")
        
        self._viper.update()
        
        if self._save_png:
            self._viper.write_png()
    
        if self._save_eps:
            self._viper.write_ps()
    
    def finalize(self):
        if self._save_png:
            self._viper.movie()
        if self._interactive:
            self._viper.interactive()
    
class ValueLogger(DiffsimObject):
    """ A value logger

    An object of this fetches values from an other DiffsimObject and presents
    a formated output to the screen and/or store them to file.
    """
    def __init__(self, var_name, markers = None, stats = [MEAN], to_screen = True, to_file = True):
        """ Instantiate an object of ScreenLogger

        @param var_name : The name of the logged variable (Must be a Diffusive ligand)
        @param markers  : A list of markers of sudomains
        @param stats    : A list of stats, which are going to be presented,
                          allowed stats are MEAN, STD, VAR, MIN, MAX
        @param to_screen: Print the values to screen if true
        @param to_file  : Store the values to file if true
        """
        
        obj = DiffsimObject.get_object(var_name)
        if obj is None:
            raise AttributeError, "No DiffsimObject with name '%s' excists"%var_name
        if not set(stats).issubset(ALLOWED_STAT):
            raise RuntimeError, "Allowed format inputs: %s"%\
                  ("["+ ",".join([s.upper() for s in ALLOWED_STAT]) + "]")
        if not markers is None and not isinstance(markers,list):
            raise TypeError, "Provide a list of integers as markers"
        
        if not to_file and not to_screen:
            raise TypeError, "No point logging values if you won't present them or store them."
        
        # Constructing the name_list
        name_list = [var_name]
        name_list.append("value-logger")
        
        if not markers is None:
            if not hasattr(obj,"domain"):
                raise RuntimeError, "Can only specify markers if the value logged resides on a diffusion domain"
            name_list.insert(1,"-".join([str(i) for i in markers]))
        else:
            # Assigning the index 0 to the list of markers 
            markers = [0]
        
        super(ValueLogger,self).__init__("-".join(name_list))
        
        self._num_reg   = 0
        self._to_screen = to_screen
        self._to_file   = to_file
        self._markers   = markers
        self._ext_vars  = {}.fromkeys([var_name, 'time'])
        self._values    = None
        self._formats   = stats
        self._value_changed = True
        self._available_memory = 0
        
        # Initialize the to value collecter
        if to_file:
            self._values = []
        else:
            self._values = {}
            for m in self._markers:
                self._values[m] = {}
                for stat in self._formats:
                    self._values[m][stat] = 0.0

        # Register the instantiated object
        self._register_DiffsimObject(ValueLogger)
    
    def reset(self):
        " Reset the ValueLogger"
        from numpy import zeros
        # If we are writing to file and not reseting for the first time
        if self._to_file and self._num_reg != 0:
            self.__concatenate_memory()
            
        # If we are writing to file initialize the memory for the new run
        if self._to_file:
            self._values.append({})
            for m in self._markers:
                self._values[-1][m] = {}
                for stat in self._formats:
                    self._values[-1][m][stat] = zeros(self._params.chunk)
        
        self._available_memory = self._params.chunk
        self._num_reg = 0
        
    def receive(self, sender):
        # For now assume that it can only receive data from a DiffusiveLigand
        if sender.name() != 'time':
            if not isinstance(sender,(DiffusiveLigand)):
                raise TypeError, "Can only receive data from a DiffusiveLigand"
            self._ligand = sender
            self._value_changed = True
        else:
            time = sender.value()
            if self._value_changed:
                self._value_changed = False
                self.__collect_values()
                if self._to_screen:
                    # Collect lines
                    lines = []
                    lines.append("")
                    lines.append("%s at time: %6.3f"%(self._ligand.name(),time))
                    header  = " | ".join(["  m "] + [s.capitalize().center(9)
                                                     for s in self._formats])
                    lines.append(header)
                    lines.append("".join(['-']*(len(header)+3)))
                    for m in self._markers:
                        lines.append(" %3d | "%m + " | ".join(\
                            ["%9.3f"%(self.__get_value(m,s)) for s in self._formats]))
                    
                    lines.append("")

                    # Print it to screen
                    info("\n".join(lines))
                
                # Check for negative values
                if get_log_level() <= WARNING:
                    num_below_zero = (self._ligand.value()<0).sum()
                    if num_below_zero>0:
                        warning("%d values of %d are below zero."%\
                                (num_below_zero,self._ligand.domain().num_dofs()))
                
    def __collect_values(self):
        " Collects values and store them in the values list"
        # If writing to file
        if self._to_file:
            # Extend memory if needed
            if self._num_reg >= self._available_memory:
                self.__extend_memory()
            for m in self._markers:
                for stat in self._formats:
                    self._values[-1][m][stat][self._num_reg] = self._ligand.value(m,stat)
            self._num_reg += 1
        else:
            for m in self._markers:
                for stat in self._formats:
                    self._values[m][stat] = self._ligand.value(m,stat)

    def __get_value(self, m, stat):
        " Return the collected value"
        if self._to_file:
            return self._values[-1][m][stat][self._num_reg-1]
        else:
            return self._values[m][stat]

    def __extend_memory(self):
        " Extend the reserved memory"
        assert(self._to_file)
        from numpy import append, zeros
        for m in self._markers:
            for stat in self._formats:
                self._values[-1][m][stat] = append(self._values[-1][m][stat],
                                                   zeros(self._params.chunk))
        self._available_memory += self._params.chunk

    def __concatenate_memory(self):
        if len(self._values) > 0:
            for m in self._markers:
                for stat in self._formats:
                    self._values[-1][m][stat] = self._values[-1][m][stat][:self._num_reg]
    
    def finalize(self):
        # If we are writing to file
        if self._to_file:
            self.__concatenate_memory()
            values = self._values
            self._values = []
            return values
        
        return None
        
    @classmethod
    def default_params(cls):
        return ParameterDict(chunk = 50)

                
class PointLogger(DiffsimObject):
    """ A Point logger

    An object of this fetches values from a DiffusiveLigand and store values of
    the ligand in given points.

    Don't give too many points could be slow...
    """
    def __init__(self, var_name, points = None):
        """ Instantiate an object of PointLogger
        
        @param var_name : The name of the logged variable (Must be a Diffusive ligand)
        @param points   : The points you want to log, given as a Nx3 numpy array, of
                          coordinates, where N is the number of points logged
        """
        from numpy import zeros
        obj = DiffsimObject.get_object(var_name)
        if obj is None:
            raise AttributeError, "No DiffsimObject with name '%s' excists"%var_name

        if not isinstance(obj,DiffusiveLigand):
            raise TypeError, "The DiffsimObject '%s' is not a DiffusiveLigand"%var_name

        if points is None:
            self._num_val  = obj.domain().num_dofs()
            self._points   = None
        else:
            if not obj.domain().points_inside(points):
                raise ValueError, "One or more points are outside the domain"
            self._num_val  = points.shape[0]
            self._points   = points[:]
        
        # Constructing the name_list
        name_list = [var_name]
        name_list.append("point-logger")
        super(PointLogger,self).__init__("-".join(name_list))
        
        self._ext_vars   = {}.fromkeys([var_name, 'time'])
        self._times      = zeros(self._params.chunk)
        self._values     = zeros((self._params.chunk,self._num_val))
        self._num_reg    = 0
        self._field_func = obj.domain().scalar_field_function()
        self._file_name  = None
        
        # Register the instantiated object
        self._register_DiffsimObject(PointLogger)

    @classmethod
    def default_params(cls):
        return ParameterDict(chunk = 50)

    def __extend_memory(self):
        " Extend the reserved memory for the values and times"
        from numpy import append, zeros
        chunk        = self._params.chunk
        self._times  = append(self._times,zeros(chunk))
        self._values = append(self._values, zeros((chunk,self._num_val)),0)

    def reset(self):
        if self._file_name is None:
            self.set_case_name()
        
    def set_case_name(self,case_name = "simulation"):
        import os
        var_name = [ s for s in self._ext_vars.iterkeys() if s != "time"][0]
        self._file_name = "%s_%s.cpickle"%(case_name,var_name)
        #if os.path.isdir("data"):
        #    self._file_name = os.path.join("data",self._file_name)
            
    def file_name(self):
        " Return the file name used to to store the points"
        return self._file_name
    
    def receive(self, sender):
        # For now assume that it can only receive data from a DiffusiveLigand
        from numpy import zeros
        if sender.name() != 'time':
            if not isinstance(sender,(DiffusiveLigand)):
                raise TypeError, "Can only receive data from a DiffusiveLigand"
            self._ligand = sender
            self._value_changed = True
        else:
            time = sender.value()
            if self._value_changed:
                if self._num_reg == len(self._times):
                    self.__extend_memory()
                self._times[self._num_reg] = time
                if self._points is None:
                    self._values[self._num_reg,:] = self._ligand.value()
                else:
                    self._field_func.vector().set(self._ligand.value())
                    point = zeros(3)
                    value = zeros(1)
                    for i in xrange(self._num_val):
                        point[:] = self._points[i,:]
                        self._field_func.eval(value,point)
                        self._values[self._num_reg,i] = value[0]
                self._num_reg += 1
                self._value_changed = False
    
    def finalize(self):
        " Collect the data and return them in a dict"
        ret = dict(time = self._times[:self._num_reg],\
                   values = self._values[:self._num_reg,:],\
                   points = self._points)
        
        self._times      = zeros(self._params.chunk)
        self._values     = zeros((self._params.chunk,self._num_val))
        self._num_reg    = 0
        return ret
        
#class VTKLogger(object):
#    def __init__(self,instance, case_name = None):
#        self._demands = [instance.name]
#        self._marker = marker
#        name_list = [instance.name]
#        if not case_name is None:
#            if not isinstance(case_name,str):
#                raise RuntimeError, "Pass string as case_name" 
#            name_list.append(case_name)
#        name_list.append("-VTKLogger")
#        self._name = "".join(name_list)
#        if not hasattr(instance,"domain"):
#            raise RuntimeError, "Can only plot a continuous variable that resides in a domain."
#        self.save_field = instance.domain.gal.save_field
#        self._values = instance.values()
#    
#    @property
#    def name(self):
#        return self._name
#    
#    @property
#    def demands(self):
#        return self._demands
#    
#    def reset(self):
#        self._p.update()
#    
#    def fetch_values(self,value,var = None):
#        # If the sender is a DiffusiveLigand, then it wont send a value but it self!
#        if isinstance(value,DiffusiveLigand):
#            ligand = value
#            if not self._marker is None:
#                raise NotImplementedError, "Plotting of subdomains is not implemented"
#                self._values[:] = ligand.values(markers)
#            self._p.update()
#            
#            if self._save_png:
#                self.save_png()
#            
#        else:
#            raise RuntimeError, "Can only plot results from a diffusive ligand"
#
#from time import strftime
#
#class FileLogger(object):
#    def __init__(self, instance, file_name):
#        raise NotImplemented
#        # TODO: Finish this. Add simulation end to TimeHandler and find a way to
#        #       save multiple runs. Maybee a final simulation end, and for each
#        #       reset we add a dictionary of values to a list or something.
#        self._file_name = file_name
#        self._demands = [instance.name, 'time', 'simulation end']
#        
#        name_list = self._demands[:]
#        name_list.append("file-logger")
#        self._name = "-".join(name_list)
#        self._values = instance.values()[:]
#        self._value_changed = False
#    
#    @property
#    def name(self):
#        return self._name
#    
#    @property
#    def demands(self):
#        return self._demands
#    
#    def reset(self):
#        if self._file_name is None:
#            file_str = "-".join([self._name,strftime('%y.%m.%d-%H.%M.%S')])
#        else:
#            file_str = self._file_name
#        self._file = file(file_str,"w")
#        def writeln(s):
#            self._file.write(s)
#            self._file.write('\n')
#        self._print = writeln
#                
#    def fetch_values(self,value,var):
#        # If the sender is a DiffusiveLigand, then it wont send a value but it self!
#        if isinstance(value,DiffusiveLigand):
#            ligand = value
#            self._values[:] = ligand.values()
#            self._value_changed = True
#        elif var != 'time':
#            self._values[var] = value
#            self._value_changed = True
#        else:
#            time = value
#            if self._value_changed:
#                var = self._demands[0]
#                value_list = ["%7.2f"%eval("_N.%s(self._values[var])"%s) for s in self._formats if s != ALL]
#                value_list.insert(0,"%6.3f"%time)
#                if self._marker is None:
#                    value_list.insert(0,"Whole domain: " + str(self._marker))
#                else:
#                    value_list.insert(0,"marker:%4d"%self._marker)
#                self._print(", ".join(value_list))
#                self._value_changed = False
#                

debug("%s loaded"%__name__)
