#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Johan Hake (johan.hake@gmail.com)"
__copyright__ = "Copyright (C) 2008 Johan Hake"
__license__  = "GNU LGPL Version 3.0 or any later"
__date__ = "2008-06-05 -- 2009-09-18"

"""A simple wrapper around the logging module in python"""

# Import the loggin module
import logging as _logging

# Import sys
import sys

# Assign the loging functions to the ROOT loggers functions
log, critical, error, warning, info, debug = \
     _logging.log,     _logging.critical, _logging.error, \
     _logging.warning, _logging.info,     _logging.debug

# Assign log levels 
CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET = \
    _logging.CRITICAL, _logging.ERROR, _logging.WARNING,\
    _logging.INFO,     _logging.DEBUG, _logging.NOTSET

# The default log level
_default_log_level = INFO

# Define the formating of the log messages
_logging.basicConfig(
    format = "%(levelname)-10s %(name)-12s %(message)s",
    level  = _default_log_level
    )

# Some constants
SCREEN, FILE, SILENT, ROOT, ALL = "SCREEN", "FILE", "SILENT", "ROOT", "ALL"

# Using the getLogger function from the logging module
get_logger = _logging.getLogger

def present_logger():
    """ Return the name of the present logger

    If no local logger is registered the "ROOT" logger is returned.
    """
    if hasattr(debug,"im_self"):
        return debug.im_self.name
    else:
        return "ROOT"

def get_log_level(logger = None):
    """ Return the log level of the present logger.

    If no local logger is registered return the level of the root logger

    @param logger : If not None, the level of this logger is returned
    """
    if logger is None:
        # Check if there is a registered local logger,
        # i.e., checking if the debug method belongs to an actuall logger instance
        if hasattr(debug,"im_self"):
            return debug.im_self.level
        else:
            # Return the level on the root logger
            return get_logger().level
    elif logger.upper() == "ROOT":
        return get_logger().level
    elif get_logger().manager.loggerDict.has_key(logger):
        return get_logger(logger).level
    else:
        raise RuntimeError, "No logger with name %s"%str(logger)
    
def set_log_level(level, logger = None):
    """ Set the level of a logger.
    
    @param level  : The level to be set
    @param logger : If given this logger's level is set.
                    If None the local logger is set.
                    If no local logger is found the level of the root logger is set
    """
    if logger is None:
        # Check if there is a registered local logger,
        # i.e., checking if the debug method belongs to an actuall logger instance
        if hasattr(debug,"im_self"):
            debug.im_self.setLevel(level)
        else:
            # Set the level on the root logger
            get_logger().setLevel(level)
    elif logger.upper() == ROOT:
        get_logger().setLevel(level)
    elif get_logger().manager.loggerDict.has_key(logger):
        get_logger(logger).setLevel(level)
    elif logger.upper() == ALL:
        for l in all_loggers():
            if l not in ["UFL", "FFC", "instant"]:
                get_logger(l).setLevel(level)
    else:
        raise RuntimeError, "No logger with name %s"%str(logger)

def format_log_messages(format_str):
    """ Format the log messages.

    The format strings applies to all streams. (Only have one for now...)
    """
    raise NotImplementedError, "Implement me!"
    for stream in get_logger().handlers():
        stream.setFormatter
    
def set_log_destination(dest = None, file_name = None):
    """ Set the log destination

    The destination can be set to either SCREEN, FILE, or SILENT

    @param dest :     The destination, if none is provided the screen
                      and stderr is default.
    @param file_name: An optional name of a logger file. If None is provided
                      "ouput.log" is choosen.
    """
    # Remove any registered handlers
    if not dest in [None, SCREEN, FILE, SILENT]:
        raise RuntimeError, "Invalid destination."
    
    handlers = get_logger().handlers
    while handlers:handlers.pop()
    
    if dest is None or (dest.upper() == SCREEN):
        handlers += [_logging.StreamHandler(sys.stdout)]
    
    if dest.upper() == FILE:
        file_name = file_name or "output.log"
        handlers += [_logging.FileHandler(file_name)]
        
def init_local_logger(logger, ns = None, level = None):
    """ Add a logger with the specified name.
    
    If a name space dict is provided the logging functions: "log", "critical",
    "error", "warning", "info", "debug" are added to it.

    If the name is an allready registered logger then no new logger is created.

    @param logger : The name of the logger
    @param ns     : A name space dict which will be populated with the log functions
    @param level  : The level the logger is initialized with
    """
    logger = get_logger(logger)
    if not level is None:
        logger.setLevel(_default_log_level)
    if not ns is None:
        if not isinstance(ns,dict):
            raise RuntimeError, "Provide a name space dict"
        ns["log"]             = logger.log
        ns["critical"]        = logger.critical
        ns["error"]           = logger.error
        ns["warning"]         = logger.warning
        ns["info"]            = logger.info
        ns["debug"]           = logger.debug   
    return logger

def all_loggers():
    """ Return a list of all registered loggers.
    """
    # TODO: print out the attached loggers in an indented way.
    # Showing the interdependency of the loggers
    return get_logger().manager.loggerDict.keys()

# Set the log destination to stdout
set_log_destination(SCREEN)

if __name__ == "__main__":
    init_local_logger(__name__,globals(),INFO)
    info("This is a simple info message")
    debug("This debug messages will not show")
    set_log_level(DEBUG)
    debug("This debug messages will show")
    another_logger = init_local_logger(__name__ + ".another")
    another_logger.info("This message is from another logger")
    set_log_level(CRITICAL,__name__)
    info("This should not show")
    another_logger.info("nor should this message, as long that the level is NOTSET")
    
