from numpy import inf
from string import ljust, rjust, center

# Initialize the logging system
from diffsim.logger import *
init_local_logger(__name__,globals(),INFO)

VALUE_JUST = rjust

__all__ = ['TypeCheck','RangeCheck','OptionCheck','ParConstant','inf']


def value_formatter(value, str_length = 0):
    " Return a formated string of a value"
    ret = None
    if isinstance(value,float):
        if value == inf:
            ret = "\xe2\x88\x9e"
        elif value == -inf:
            ret = "-\xe2\x88\x9e"
    elif isinstance(value,str):
        ret = repr(value)
    if ret is None:
        ret = str(value)
    if str_length == 0:
        return ret
    return VALUE_JUST(ret,str_length)

scalars = (int,float)
option_types = scalars + (str,)

class TypeCheck(object):
    " A simple type checking class for a single value"
    def __init__(self, value, parname = None):
        self._value_type = type(value)
        self._value      = value
        self._repr_str   = "TypeCheck(%s)"
        self._not_in_str = None
        self._in_str     = None
        self._in_range   = None
        self._parname    = None

    def set(self,value):
        " Try to set the value using the check"
        self._value = self.check(value)

    def set_parname(self,parname):
        if not isinstance(parname,str):
            raise TypeError, "Provide a 'str'."
        self._parname = parname

    def get(self):
        " Return the value"
        return self._value
    
    def check(self,value):
        " Check the value using the type and any range check"
        # Fist check if the value is an int or float.
        # (Treat these independently)
        name_str = "" if self._parname is None else " for the parameter '%s'"%self._parname
        if self._value_type in [int,float] and isinstance(value,(int,float)):
            if isinstance(value,float) and self._value_type == int:
                warning("Converting %s to %d%s"%(str(value),int(value),name_str))
            value = self._value_type(value)
        if not isinstance(value,self._value_type):
            raise TypeError, " Please provide a '%s'%s"%(self._value_type.__name__,name_str)
        if not self._in_range is None:
            if not self._in_range(value):
                raise ValueError, "Illegal value%s: %s"%(name_str,self.format(value,True))
        return value
    
    def format(self, value = None, not_in = False, str_length = 0):
        " Print a nice formated version of the value and its range"
        if value is None:
            value = self._value

        # If no '_in_str' is defined
        if self._in_str is None:
            return value_formatter(self._value,str_length)
        
        if not_in:
            return self._not_in_str%(value_formatter(value,str_length))
        else:
            return self._in_str%(value_formatter(value,str_length))

    def length_value_format(self):
        " Return the length of the formated str of value"
        return len(value_formatter(self._value))

    def __repr__(self):
        " Print a executable version of the Check"
        return self._repr_str%repr(self._value)
    
    def __str__(self):
        " Print a nice representation of the value and Check"
        return self.format()

class ParConstant(TypeCheck):
    " A Constant parameter which prevent any change of values" 
    def __init__(self, value):
        TypeCheck.__init__(self,value)
        self._repr_str = "ParConstant(%s)"
        
        # Define a 'check function'
        self._in_range = lambda x:x==self._value
        
        # Define some string used for pretty print
        self._in_str = "%s - Constant"
        self._not_in_str = "%%s != %s"%str(self._value)
        self.set(value)
    
class RangeCheck(TypeCheck):
    " A simple type and range checking class for a single value"
    def __init__(self, value, ge=None, le=None, gt=None, lt=None, parname=None):
        TypeCheck.__init__(self,value,parname)
        
        # If no range is provided
        if ge is None and le is None and gt is None and lt is None:
            raise TypeError, "Provide at least one range"
        
        # Checking valid combinations of kwargs
        if not ge is None and not gt is None:
            raise TypeError, "Cannot create a 'RangeCheck' including "+\
                  "both 'ge' and 'gt'"
        if not le is None and not lt is None:
            raise TypeError, "Cannot create a 'RangeCheck' including "+\
                  "both 'le' and 'lt'"

        # Checking valid types for 'range checks'
        wrong_range_check_type = False
        for arg in [value,ge,le,gt,lt]:
            if not arg is None:
                wrong_range_check_type |= not isinstance(arg,scalars)
                
        if wrong_range_check_type:
            raise TypeError, "Can only provide 'range check' for" + \
                  " 'int' or 'float'"
        
        # Pick the min and max values
        if not ge is None:
            min_val = ge
        elif not gt is None:
            min_val = gt
        else:
            min_val = -inf
        
        if not le is None:
            max_val = le
        elif not lt is None:
            max_val = lt
        else:
            max_val = inf

        # Check the relation between the min and max value
        if min_val > max_val:
            raise ValueError, "Please provide a 'max value' that is "+\
                  "larger than the 'min value'."
        
        range_check_dict = {}

        # Dict for test and repr
        range_check_dict["min_op"] = ">=" if gt is None else ">"
        range_check_dict["max_op"] = "<=" if lt is None else "<"
        range_check_dict["min_value"] = repr(min_val)
        range_check_dict["max_value"] = repr(max_val)
        
        # Dict for pretty print
        range_check_dict["min_op_format"] = "[" if gt is None else "("
        range_check_dict["max_op_format"] = "]" if lt is None else ")"
        range_check_dict["min_format"] = value_formatter(min_val)
        range_check_dict["max_format"] = value_formatter(max_val)

        # Define a 'check function'
        self._in_range = eval(("lambda value:value %(min_op)s %(min_value)s "+\
                                 "and value %(max_op)s %(max_value)s")%range_check_dict)

        # Define some string used for pretty print
        self._in_str = ("%%s \xe2\x88\x88 %(min_op_format)s%(min_format)s," +\
                       "%(max_format)s%(max_op_format)s")%range_check_dict
        
        self._not_in_str = ("%%s \xe2\x88\x89 %(min_op_format)s%(min_format)s," +\
                           "%(max_format)s%(max_op_format)s")%range_check_dict

        # Define a 'repr string'
        op_list = [ge,gt,le,lt]
        op_str_list = ["ge","gt","le","lt"]
        self._repr_str = "RangeCheck(%%s,%s)"%(", ".join(["%s=%s"%(op_str_list[v[0]],\
            repr(v[1])) for v in enumerate(op_list) if not v[1] is None]))
        
        # Set the value using the check functionality
        self.set(value)
    
class OptionCheck(TypeCheck):
    " A simple type and options checking class for a single value"
    def __init__(self, value, options):
        TypeCheck.__init__(self,value)
        
        if not isinstance(options,list):
            raise TypeError, "Provide a 'list' with the valid options."
        
        # Check valid types for an 'option check'
        if not isinstance(value,option_types) or isinstance(value,bool):
            raise TypeError, ("Can only provide 'option check' for 'str'," + \
                  " 'int' and 'float', not for '%s'")%(type(value).__name__)
        
        # Check that all values in options has the same type
        for val in options:
            if not isinstance(val,self._value_type):
                raise TypeError, ("All values of the 'option check' " +\
                      "need to be of type: '%s'")%(type(value).__name__)
        
        # Define a 'check function'
        self._in_range = eval("lambda value:value in %s"%repr(options))
        
        # Define some string used for pretty print
        self._in_str = "%%s \xe2\x88\x88 %s"%repr(options)
        
        self._not_in_str = "%%s \xe2\x88\x89 %s"%repr(options)
        
        # Define a 'repr string'
        self._repr_str = "OptionCheck(%%s, %s)"%repr(options)
        
        # Set the value using the check functionality
        self.set(value)
    

debug("%s loaded"%__name__)

# Test code
if __name__ == "__main__":
    x = TypeCheck(3)
    print x
    print repr(x)
    x.set(5)
    print x
    
    v = RangeCheck(5,0,10)
    print v
    print repr(v)
    
    u = OptionCheck("a",["a","b","c","d"])
    print u
    print repr(u)
