__author__ = "Johan Hake (johan.hake@gmail.com)"
__copyright__ = "Copyright (C) 2008 Johan Hake"
__license__  = "GNU LGPL Version 3.0 or any later"
__date__ = "2009-05-07 -- 2009-05-07"

__all__ = ['SteadyStateActivator',
           'SteadyStateLigand']

from diffsim.diffusiveligand import DiffusiveLigand, DiffsimObject
from diffsim.stochastictimestepper import DiscreteMarkovModel
from diffsim.base            import ContinuousVariable, check_ion_name
from diffsim.common          import (FARADAY,UNIVERSAL_GAS,T_ABS,VALENCE,MILLI,
                                     MICRO,NANO,PICO,NM3_per_DM3,ALLOWED_IONS,Na,Ca)
from diffsim.parameterdict   import ParameterDict, TypeCheck, RangeCheck, OptionCheck

from diffsim.dyadiccleft.markovmodels import (LCCMarkovModel_Greenstein,
                                              LCCMarkovModel_Hake,
                                              RyRMarkovModel_Stern,
                                              RyRMarkovModel_Dura,
                                              RyRMarkovModel_Polakova,
                                              NaMarkovModel_Clancy)

import numpy

# Initialize the logging system
from diffsim.logger import *
init_local_logger(__name__,globals(),INFO)

class SteadyStateActivator(DiffsimObject):
    """ Objects of this class contains informations for a SteadyStateLigand

    """
    def __init__(self, name, MarkovModel, start_marker, *on_and_off_info, **dependency):
        super(SteadyStateActivator,self).__init__(name)
        
        # Check the MarkovModel
        assert issubclass(MarkovModel,DiscreteMarkovModel)
        self._MM = MarkovModel
        
        # The values 0 for off values and 1 for on values
        self._value_map = []
        self._value_map.append({})
        self._value_map.append({})
        self._value_type = {}
        self._value_cache = []
        
        # Check the on and off infos
        size_dependency = None
        all_markers = []

        assert len(on_and_off_info) % 3 == 0, "provide on_and_off_infos in arguments of three"
        num_types_on_and_off = len(on_and_off_info) / 3
        num_markov_models = None
        markers_to_markov_models = {}
        for i in xrange(num_types_on_and_off):
            # The first argument should be list or numpy array of ints containing the markers
            markers = on_and_off_info[i*3]
            assert isinstance(markers,(list,numpy.ndarray))
            if isinstance(markers,list):
                assert all(isinstance(m,int) for m in markers)
            else:
                assert markers.dtype == int
                
            if num_markov_models is None:
                num_markov_models = len(markers)
            else:
                assert num_markov_models == len(markers)
                
            all_markers.extend(markers)
            
            # Check the on_values and off_values
            for j in xrange(1,3):
                values = on_and_off_info[i*3+j]
                assert isinstance(values,(float,numpy.ndarray))
                if isinstance(values,numpy.ndarray):
                    if size_dependency is None:
                        size_dependency = len(values)
                    else:
                        assert size_dependency == len(values)
                else:
                    assert size_dependency == 0
            
            # Store the values
            # NOTE: on values are provided first
            self._value_map[1][i] = on_and_off_info[i*3+1]
            self._value_map[0][i] = on_and_off_info[i*3+2]
            
            # Initialize the value_cache
            self._value_cache.append([0.,0.])

            # Create a map from markers to types of on and off values
            for j, m in enumerate(markers):
                self._value_type[m] = i
                markers_to_markov_models[m] = j

        # Store the markers to markov model map
        self._markers_to_markov_models = markers_to_markov_models        
        
        # Store the dependency size 
        self._size_dependency = size_dependency

        # Store the number of different types of on and off values
        self._num_types_on_and_off = num_types_on_and_off

        # Store the number of markov models
        self._num_markov_models = num_markov_models
            
        # Check that no markers are duplicated
        assert len(all_markers) == len(set(all_markers))

        # If size_dependency > 0 a dependency array needs to be provided
        if size_dependency > 0:
            assert len(dependency) == 1, "Provide a dependency array"
            values = dependency.values()[0]
            assert isinstance(values,numpy.ndarray)
            assert values.dtype == float
            assert len(values) == size_dependency
            
            # Register the dependency
            self._ext_vars[dependency.keys()[0]] = 0.0
            self._dependency_values = values
            
        # Store the markers
        self._markers = sorted(all_markers)
        #debug("markers: %s"%str(self._markers))
        
        # Store the states of the included Markov models
        self._mm_states = numpy.zeros(self._num_markov_models).astype(bool)
        
        # Create the MarkovModels and register dependency
        for i in xrange(self._num_markov_models):
            MM_name = "%s_m%d_activator"%(self._name,i+start_marker)
            mm = MarkovModel(MM_name,i+start_marker)
            
            # Register dependency and state vectors
            self._mm_states[i] = self._ext_vars[MM_name] = mm.value()
        
        self._register_DiffsimObject(SteadyStateActivator)

    def reset_activator(self):
        for i in xrange(self._num_markov_models):
            self._mm_states[i] = 0
        for key in self._ext_vars:
            if "_activator" in key:
                self._ext_vars[key] = 0
    
    def receive(self, sender):
        # Not nesescary to register the values in the _ext_vars dict.
        # If the triggering MarkovModel has changed value
        if isinstance(sender,self._MM):
            marker = sender.marker()

            # Only works for on and off values of markov model
            self._mm_states[self._markers_to_markov_models[marker]] = bool(sender.value())
        elif self._size_dependency > 0 :
            dep_value = sender.value()
            # Iterate over the different types
            for j in xrange(self._num_types_on_and_off):
                # iterate over the on and off arrays
                for i in [0,1]:
                    #debug("dep values %s"%str(self._dependency_values))
                    #debug("value map %s"%str(self._value_map[i][j]))
                    #debug("dep_value %s"%str(dep_value))
                    self._value_cache[i][j] = numpy.interp(numpy.array([dep_value]),
                                                           self._dependency_values,
                                                           self._value_map[i][j])[0]
        else:
            raise ValueError, "The sender is not an instance of '%s' and no other dependencies are registered"%type(sender).__name__

    def value(self,m):
        assert m in self._markers
        val = self._value_cache[self._mm_states[self._markers_to_markov_models[m]]][self._value_type[m]]
        #debug("%s=%s.value(%d)"%(str(val),self.name(),m))
        return val
    
    def markers(self):
        return self._markers

class SteadyStateLigand(ContinuousVariable):
    """ This class defines an interface for a DiffusiveLigand in steady state
    """
    def __init__(self, domain_name, ion_name, steady_state_activators):
        
        check_ion_name(ion_name)
        name = domain_name + "_" + ion_name
        super(SteadyStateLigand,self).__init__(name)
        self._ion_name = ion_name

        if not isinstance(steady_state_activators,list):
            steady_state_activators = [steady_state_activators]

        assert all(isinstance(ssa,SteadyStateActivator) for ssa in steady_state_activators)
        markers = []
        ssa_map = {}
        for i, ssa in enumerate(steady_state_activators):
            markers += ssa.markers()
            ssa_map.update(zip(ssa.markers(),[i]*len(ssa.markers())))
        assert len(markers) == len(set(markers)), "No duplicates of markers are allowed"

        # Dummy default values
        self._values_0 = 0.0; self._values_1 = 0.0; self._values_2 = 0.0

        self._ssa_map = ssa_map
        self._steady_state_activators = steady_state_activators
        self._markers = markers
        
        self._register_DiffsimObject(SteadyStateLigand)

    def reset(self):
        for ssa in self._steady_state_activators:
            ssa.reset_activator()
    
    def solve(self, dt, iteration=0):
        " No solve as the everything is happening in the SteadyStateActivators"
        self._converged = True
        return
        
    def ion_name(self):
        return self._ion_name
    
    def value(self,m,*args):
        return self._steady_state_activators[self._ssa_map[m]].value(m)
    
debug("%s loaded"%__name__)
