__author__ = "Johan Hake (johan.hake@gmail.com)"
__copyright__ = "Copyright (C) 2008 Johan Hake"
__license__  = "GNU LGPL Version 3.0 or any later"
__date__ = "2008-02-19 -- 2008-11-18"


# Defining the import * statement
__all__ = ['DiscreteMarkovChain', 'DiscreteMarkovModel','StochasticTimeStepper',\
           'mul', 'add', 'sub', 'div']

# Importing DiffSim stuff
from diffsim.disk            import save
from diffsim.base            import (DiffsimObject, DiscreteVariable,
                                     Singleton, Q10_scale)
from diffsim.common          import T_ABS
from diffsim.parameterdict   import ParameterDict, TypeCheck, RangeCheck, OptionCheck
from diffsim.diffusiveligand import DiffusiveLigand

# Initialize the logging system
from diffsim.logger import *
init_local_logger(__name__,globals())

set_log_level(INFO,__name__)

# Importing operators for combining MarkovChains
from operator import mul, add, sub, div, imul, or_

# Import numpy stuff
import numpy
import math
import numpy.random

class DiscreteMarkovChainType(type):
    """ DiscreteMarkovChainType
    
    A metaclass for DiscreteMarkovChain class. It reads attributes provided by
    a subclass of DiscreteMarkovChain and processes them and add more usefull
    attributes.
    """
    def __new__(cls,class_name,class_parents,class_dict):
        # Check that the metaclass is building a subclass of the MarkovChain class
        if class_parents[0].__name__ == "DiscreteMarkovChain":
            # Check that the sub class implements the following class variables
            attributes = ["_rate_func_map", "_params", "_ext_vars", "_state_values"]
            attr_not_implemented = filter(lambda key:not key in class_dict.keys(), attributes)
            if attr_not_implemented:
                raise RuntimeError, """
                Provide the following attributes to the subclass of MarkovChain:
                %s"""%(", ".join(attr_not_implemented))

            # Tabulate the state information from the provided class dict
            # Make a list with empty tuples, one for each state
            class_dict["_num_states"] = max(max(class_dict["_rate_func_map"]))+1
            class_dict["_state_connections"] = [[] for i in xrange(class_dict["_num_states"])]
            
            # Fill the tuples with connection information
            connections = class_dict["_rate_func_map"].keys()
            connections.sort()
            for from_state, to_state in connections:
                class_dict["_state_connections"][from_state] += [to_state,]
            
            class_dict["_max_num_transitions"] = max(len(c) for c in class_dict["_state_connections"] )
            state_values = [class_dict["_state_values"].get(state,0)
                        for state in xrange(class_dict["_num_states"])]

            class_dict["_state_values"] = state_values
            
            # Tabulate the states dependencies of external variables
            # Initialize all dependencies to False
            ext_var_list = class_dict["_ext_vars"].keys()
            
            #ext_vars.sort()
            ext_vars = {}.fromkeys(ext_var_list,0)
            ext_var_state_dep = [{}.fromkeys(ext_var_list,False)
                                 for i in xrange(class_dict["_num_states"])]
            
            # Collect the informations
            for ext_var in ext_var_list:
                ext_vars[ext_var] = class_dict["_ext_vars"][ext_var][0]
                for state in class_dict["_ext_vars"][ext_var][1]:
                    ext_var_state_dep[state][ext_var] = True

            # Overwrite the class_dict["_ext_vars"] with the tabulated one
            class_dict["_ext_vars"] = ext_vars
            class_dict["_ext_var_state_dep"] = ext_var_state_dep
            
        return type.__new__(cls,class_name,class_parents,class_dict)

def _new_state(rates):
    """ Returns a random number between zero and len(cum_rate)-1,
    based on the cum_rates."""
    return (rates.cumsum()>rates.sum()*numpy.random.random()).argmax()
    
def _compute_new_dimless_exit_time():
    """
    Returns a new dimensionless exit time for a homogeneuous
    poision process with rate 1."""
    dimlesstime = -math.log(numpy.random.random())
    debug("Computing new dimensionless time: %s"%str(dimlesstime))
    return dimlesstime

class DiscreteMarkovChain(DiffsimObject):
    """
    MarkovChain is a base class with generic functionalities for handling a discrete
    Markov chain. It must be subclassed to be used. The subclass must define the
    following four class attributes:
    
    _rate_func_map : A dict with rate functions. The keys are tuples defining the
                     connections the the functions describe.
                     
                     {(0,1):rate_func_1,(1,0):rate_func_2} is a map of rate functions
                     for a two state chain. The rate functions need to accept a
                     parameter dict as the first argument and a dict containing values
                     of any external variables.
                     
    _params        : A ParameterDict that defines the parameters of the Markov chain.
                     This is sent to the rate functions when these are called. If the
                     rate functions are not dependent on parameters, define an
                     empty dict.
                     
    _ext_vars      : A dict defining the Markov chains dependencies of external
                     variables. The keys are the names of the external variables and
                     the values are tuples with size two, defining first a predefined
                     initial value of the external variable, and the second is a list
                     with the states that have rate functions that are dependent on
                     the variable.
                     
    _state_values  : This is a dict defining the values seen from the outside of a
                     chain. Only define the values of any nonzero states.
                     E.g. {0:1} defines state 0 as an "open" state, with value 1.
                     Other values can also be used.
                     """
    
    __metaclass__ = DiscreteMarkovChainType
    def __init__(self,marker):
        """Instantiate a Markov chain.
        
        The keyword arguments the constructor can be called with, defines
        either optional initial value any external values, and/or optional
        values of any parameters."""
        # If trying to make an instance of this class
        if type(self) == DiscreteMarkovChain:
            raise RuntimeError, """
            DiscreteMarkovChain is an "abstract" base class that cannot be instantiated.
            Define a sub class of DiscreteMarkovChain implementing the right attributes.
            """
        super(DiscreteMarkovChain,self).__init__("%s_m%d"%(type(self).__name__,marker))
        
        self._marker     = marker
        self._rates      = numpy.zeros(self._max_num_transitions,'d')

        # Update the instant _ext_vars from the class.
        self._ext_vars.update(type(self)._ext_vars)

        # Instead of updating delete the instant _ext_vars and use the class one instead
        # FIXME. This introduce a serious BUG! Each markov chain needs its own
        # ext_vars values otherwise they will 
        #del self._ext_vars

        if False:#get_log_level() <= DEBUG:
            # Iterate over the gathered state_connections and print them
            for state, connections in enumerate(self._state_connections):
                debug("state: %2d, connections: %s"%(state,\
                      ", ".join("%d"%con for con in connections)))
            # Iterate over the external state dependencies and print them
            for state, state_dep in enumerate(self._ext_var_state_dep):
                debug("state: %2d, dependency: %s"%(state,\
                      ", ".join("%s : %s"%(var,str(dep)) for var, dep in state_dep.iteritems())))
        
        # Reset the markov chain
        self.reset()
        
        # Register the DiffsimObject
        # (Must be done here, the derived class wont call __init__)
        self._register_DiffsimObject(type(self))
    
    def sum_rates(self):
        """ Returns the sum of the Markov chain rates"""
        # If any external variables has been changed
        if self._ext_vars_changed:

            # Check whether the chain is dependent on the external variable in the
            # present state
            if any(self.ext_var_dep(var) for var in self._ext_vars_changed):
                #debug("Recomputing rates for %s"%self.name())
                #if "Dyadic_cleft_Ca" in self._ext_vars:
                #    debug("Dyadic_cleft_Ca: %s"%str(self._ext_vars['Dyadic_cleft_Ca']))
                #    debug("marker: %d"%self._marker)
                self._ext_vars_changed = []
                self._recompute_transition_rates()
        
        return self._rates.sum()
    
    def execute_transition(self):
        """ Execute a random transition
        
        Computes a new state and recompute the rates
        """
        self._state = self._state_connections[self._state]\
                      [_new_state(self._rates)]
        self._recompute_transition_rates()
    
    def ext_var_dep(self, var):
        return self._ext_var_state_dep[self._state][var]
    
    def reset(self):
        " Reset the state and rates of the Markov Chain."
        # Compute the initial probabilities if None
        N = self._num_states
        b = numpy.zeros((N,),'d')
        b[-1] = 1.
        A = numpy.zeros((N,N),'d')
        for state in xrange(N):
            rates = self._compute_transition_rates(state)
            A[state,state] = -rates.sum()
            A[self._state_connections[state],state] = rates
        A[-1,:] = 1
        #debug("Initital State System Matrix: %s"%str(A))
        init_probs = numpy.linalg.solve(A,b)
        
        self._state =  _new_state(init_probs)
        self._recompute_transition_rates()
        
    def value(self):
        " Return the value of the Markov chain"
        return self._state_values[self._state]

    def _recompute_transition_rates(self):
        " Recompute the transition rates for the present state"
        self._rates.fill(0.)
        for i, trans in enumerate(self._state_connections[self._state]):
            self._rates[i] = self._rate_func_map[(self._state, trans)]\
                             (self._params,self._ext_vars)
        
    
    def _compute_transition_rates(self, state):
        """Returns a numpy array with transition rates
        
        @param state : The state for which the rates should be computed
        """
        assert(isinstance(state,int))
        assert(state>=0 and state < self._num_states)

        #print self._ext_vars
        #print self._params
        debug("Connections of state %d: %s"%
              (state,str([trans for trans in self._state_connections[state]])))

        #debug("Ext vars(%s) %s"%(self.name(),str(self._ext_vars)))
        return numpy.fromiter((self._rate_func_map[(state, trans)]\
                         (self._params,self._ext_vars)\
                         for trans in self._state_connections[state]),"d")
    
    def receive(self,sender):
        " Receive values from external objects"
        if isinstance(sender,DiffusiveLigand):
            value = sender.value(self._marker,'mean')
            #print "%s received %e from %s"%(self._name,value,sender.name())
            if value < 0.0:
                value = 0.0
        else:
            value = sender.value(self._marker)
            #print "%s received %s from %s"%(self._name,str(value),sender.name())
        self._ext_vars[sender.name()] = value
        self._ext_vars_changed.append(sender.name())

    def state(self):
        " Return the present state value"
        return self._state

    def __add__(self,other):
        if issubclass(type(other),DiscreteMarkovChain):
            return self.value() + other.value()
        if isinstance(other,(float,int)):
            return self.value() + other
        return NotImplemented
    
    def __sub__(self,other):
        if issubclass(type(other),DiscreteMarkovChain):
            return self.value() - other.value()
        if isinstance(other,(float,int)):
            return self.value() - other
        return NotImplemented
    
    def __mul__(self,other):
        if issubclass(type(other),DiscreteMarkovChain):
            return self.value() * other.value()
        if isinstance(other,(float,int)):
            return self.value() * other
        return NotImplemented
    
    def __radd__(self,other):
        if isinstance(other,(float,int)):
            return self.value() + other
        return NotImplemented
    
    def __rsub__(self,other):
        if isinstance(other,(float,int)):
            return other - self.value()
        return NotImplemented
    
    def __rmul__(self,other):
        if isinstance(other,(float,int)):
            return self.value() * other
        return NotImplemented
    
class DiscreteMarkovModel(DiffsimObject):
    
    __all_DiscreteMarkovModels = []
    
    def __init__(self, name, marker, stop_sim = False):
        if type(self) == DiscreteMarkovModel:
            raise RuntimeError, "DiscreteMarkovModel is a base class that need to be inherited."
        
        assert(isinstance(marker,int))
        super(DiscreteMarkovModel,self).__init__(name)
        
        self._marker = marker
        self._DMCs = []
        self._stop_sim = stop_sim
        
        # Create the registered DiscreteMarkovChains
        for DMC in self._DiscreteMarkovChains:
            self._DMCs.append(DMC(marker))
        
        self._num_chains = len(self._DMCs)
        self._rates = numpy.zeros(self._num_chains,'d')
        
        # Check the dmc_combine variable
        if self._dmc_combine not in [mul, add, sub]:
            raise RuntimeError, \
        "The operator to combine Markov chains has to be one of: mul, add or sub"
        
        # Register the DiffsimObject
        # (Must be done here, the derived class wont call __init__)
        self._register_DiffsimObject(type(self))

    def reset(self):
        " Reset the MarkovChains" 
        for dmc in self._DMCs:
            debug("Reseting %s"%dmc.name())
            dmc.reset()

    def marker(self):
        return self._marker

    def value(self):
        " Returns the combined value of the owed Markov chains"
        if len(self._DMCs) > 1:
            return reduce(self._dmc_combine,self._DMCs)
        return self._DMCs[0].value()
    
    def sum_rates(self):
        """ Returns the sum of the Markov Model rates"""
        result = 0.0
        for dmc in self._DMCs:
            result += dmc.sum_rates()
        return result

    def stop_sim_check(self):
        if self._stop_sim:
            self.stop_sim()
    
    def execute_transition(self):
        """ Execute a transition

        @return True if a channel transition has occured
        """
        old_value = self.value()
        if self._num_chains > 1:
            for i in xrange(self._num_chains):
                self._rates[i] = self._DMCs[i].sum_rates()
            dmc = self._DMCs[_new_state(self._rates)]
            
        else:
            dmc = self._DMCs[-1]
        
        dmc.execute_transition()
        debug("Stepping %s, new value: %d"%(self.name(), self.value()))
        return old_value != self.value(), dmc

    def _register_DiscreteMarkovModel(self):
        """ Register a DiscreteMarkovModel

        There is no need to call this method as it will be called by
        _registerDiffsimObjects.
        """
        debug("Trying to register MarkovModel: %s"%str(self))
        if not self in DiscreteMarkovModel.__all_DiscreteMarkovModels:
            debug("Actually register MarkovModel: %s"%str(self))
            DiscreteMarkovModel.__all_DiscreteMarkovModels.append(self)
            # Create a StochasticTimeStepper
            StochasticTimeStepper()

    def chain_models(self):
        " Return a shallow copy of the instantiated markov chain models"
        return self._DMCs[:]
    
    @classmethod
    def clear_all(cls):
        " Clear all registered DiscreteMarkovModels."
        cls.__all_DiscreteMarkovModels = []
    
    @classmethod
    def objects(cls):
        " Return a shallow copy of all registered DiscreteMarkovModels" 
        return cls.__all_DiscreteMarkovModels[:]
        
class StochasticTimeStepper(Singleton,DiscreteVariable):
    def __init__(self):
        """ Instantiate the Discrete Variable
        
        This function needs to be called by all derived classes.
        
        @param name : A string defining the name of the DiscreteVariable
        """
        DiscreteVariable.__init__(self,"StochasticTimeStepper")
        self._minimize_dt  = True
        self._stop_sim     = False

        self._case_name    = self.name()
        self._state_info   = []
        self._register_DiffsimObject(StochasticTimeStepper)

    def reset(self):
        " Reset the registered Markov models"

        # Seed the random generator
        numpy.random.seed()
        
        self._accumulated_time  = 0.0 
        self._dimless_time_left = _compute_new_dimless_exit_time()
        self._markov_models     = DiscreteMarkovModel.objects()
        self._num_markov_models = len(self._markov_models)
        self._rates             = numpy.zeros(self._num_markov_models,'d')
        self._rates_updated     = False
        
        # Apply scaling of rates based on any registered Q10 value
        info("Reset and scale any registered MarkovModels with regard to current temperature, T=%s \xc2\xb0K (%s \xc2\xb0C)"%(str(self.get_T()),str(self.get_T()-T_ABS)))
        for dmm in self._markov_models:
            p = dmm.get_params()
            assert isinstance(p,ParameterDict) ,"get_params() must return a ParamterDict"
            if p.has_key("Q10") and p.has_key("T_ref"):
                T_scale = Q10_scale(p.Q10,p.T_ref)
                debug("Scaling %s with %s"%(dmm.name(),str(T_scale)))
                for dmc in dmm.chain_models():
                    p = dmc.get_params()
                    assert isinstance(p,ParameterDict),"get_params() must return a ParamterDict"
                    if p.has_key("T_scale"):
                        p['T_scale'] = T_scale
            debug("Reseting %s"%dmm.name())
            dmm.reset()
        
        # Update the rates of the DiscreteMarkovModels
        self.update_rates()
        
        # Keep track of the time
        self._time              = 0.0
        
        # Broadcast the values of the registered markov models
        for dmm in self._markov_models:
            dmm.send()
        
        # Collect information 
        if _app_params.io.collect_state_info:
            self._state_info.append({})
            for dmm in self._markov_models:
                self._state_info[-1][dmm.name()] = dict(time = [0.0],values = [dmm.value()])
                for dmc in dmm.chain_models():
                    self._state_info[-1][dmm.name()][dmc.name()] = \
                            dict(time = [0.0],values = [dmc.state()])

    def update_rates(self):
        " Collect the rate sum of all registered markov models"
        if not self._rates_updated:
            for i, markov_model in enumerate(self._markov_models):
                self._rates[i] = markov_model.sum_rates()
            self._rates_updated = True
            for i,m in enumerate(self._markov_models):
                debug("%s : %.2e, states: %s"%(m.name(),self._rates[i],
                      ",".join(str(dmc._state) for dmc in m._DMCs)))
    
    def step(self):
        """ Step the DiscreteVariable
        
        This is used to update the internals of the DiscreteVariable when
        it is this variable that is "next event" in line.
        
        @return channel_transition : If the step was a "channel transition"
                                     then True is returned
        """
        # FIXME: Can we speed up the updating of rates after a transition?
        self._accumulated_time  = self.next_event()
        self._dimless_time_left = _compute_new_dimless_exit_time()
        self._rates_updated     = False
        self._last_stepped_dmm  = self._markov_models[_new_state(self._rates)]
        #debug("STEP: Stepping %s"%self._last_stepped_dmm.name())
        dmm_changed, dmc = self._last_stepped_dmm.execute_transition()

        # Register the values and time for the stepped markov chain
        if _app_params.io.collect_state_info:
            self._state_info[-1][self._last_stepped_dmm.name()][dmc.name()]\
                    ['values'].append(dmc.state())
            self._state_info[-1][self._last_stepped_dmm.name()][dmc.name()]\
                    ['time'].append(self._time + self._accumulated_time)
        
        # If the value of the markov model has changed 
        if dmm_changed:
            # Register the value of the markov model and the time
            if _app_params.io.collect_state_info:
                self._state_info[-1][self._last_stepped_dmm.name()]['time'].append(self._time + self._accumulated_time)
                self._state_info[-1][self._last_stepped_dmm.name()]['values'].append(self._last_stepped_dmm.value())
            
            # Check whether the Markov Model triggers a stop in simulation.
            self._last_stepped_dmm.stop_sim_check()
        
        return dmm_changed
    
    def next_event(self):
        " Return the time to the next event for this DiscreteVariable"
        self.update_rates()
        return self._accumulated_time + self._dimless_time_left / self._rates.sum()
    
    def update_time(self,dt):
        """ Update the time
        
        This function is used to update the internal time of the DiscreteVariable
        when another DiscreteVariables is the "next event" in line.

        @param dt : The time to next event
        """
        assert (dt - self._accumulated_time) >= 0, "The size of dt must be larger or the same size as the accumulated time."
        self._rates_updated = False
        self._time += dt
        sum_rates = self._rates.sum()
        self._dimless_time_left -= sum_rates*(dt - self._accumulated_time)
        debug("Dimless time left: %s, sum rates: %s"%(str(self._dimless_time_left),
                                                      str(sum_rates)))
        self._accumulated_time = 0.0
    
    def value(self):
        " Return the value of the last stepped Markov model"
        self._last_stepped_dmm.value()
        
    def minimize_dt(self):
        """ Resets the time if True

        This function is called by the time handler after an event is delivered.
        If it returns True the time is reseted
        """
        return self._minimize_dt
    
    def send(self):
        """ Forward the signal to the last stepped markov model"""
        self._last_stepped_dmm.send()

    def finalize(self):
        " Return the state_info and reset it."
        if not _app_params.io.collect_state_info:
            return None
        ret = self._state_info
        for i in xrange(len(ret)):
            for dmm in self._markov_models:
                ret[i][dmm.name()]['time']   = numpy.array(ret[i][dmm.name()]['time'])
                ret[i][dmm.name()]['values'] = numpy.array(ret[i][dmm.name()]['values'])
                for dmc in dmm.chain_models():
                    ret[i][dmm.name()][dmc.name()]['time']   = \
                          numpy.array(ret[i][dmm.name()][dmc.name()]['time'])
                    ret[i][dmm.name()][dmc.name()]['values'] = \
                          numpy.array(ret[i][dmm.name()][dmc.name()]['values'])
        
        self._state_info = []
        return ret
        
    def set_case_name(self,case_name):
        " Set the case name"
        self._case_name = self.name() + case_name

# Put here to avoid circular dependencies
from diffsim.app             import get_params

_app_params = get_params()


debug("%s loaded"%__name__)
