__author__ = "Johan Hake (johan.hake@gmail.com)"
__copyright__ = "Copyright (C) 2008 Johan Hake"
__license__  = "GNU LGPL Version 3.0 or any later"
__date__ = "2008-08-29 -- 2008-11-18"

# Defining the import * statement
__all__ = ['ValueClamp','VoltageStep','SigmoidalValueClamp','SigmoidalVoltageStep','DynamicValueClamp','DynamicVoltageClamp']

# Importing diffsim stuff
from diffsim.base import DiscreteVariable, Singleton, ContinuousVariable
from diffsim.parameterdict import  ParameterDict, TypeCheck, RangeCheck, OptionCheck

# Importing numpy
import numpy

# Initialize the logging system
from diffsim.logger import *
init_local_logger(__name__,globals())

set_log_level(DEBUG)

def check_argument(self,times,value_list, start_time_with_zero=True):
    # Check the arguments 
    try:
        self._times = numpy.asarray(times, dtype='d')
        len(self._times)
    except:
        raise TypeError, \
        "Please send in an iterable of numbers with length > 0, for 'times'"
    
    try:
        self._value_list = numpy.asarray(value_list, dtype='d')
        len(self._value_list)
    except:
        raise TypeError, \
        "Please send in an iterable of numbers with length > 0, for 'value_list'"
    
    if sum(self._times < 0.0) > 0:
        raise TypeError, "Please provide only positive value_list in 'times'"

    if  start_time_with_zero and (self._times[0] != 0.0):
        raise TypeError, "Please start the value_list in 'times' with 0"

    if not numpy.all(numpy.diff(self._times) > 0):
        raise TypeError, "Please provide a monoton increasing sequence of value for 'times'."

    if len(self._times) != len(self._value_list):
        raise TypeError, \
    "Please provide sequence of value_list for 'times' and 'value_list' with same length."


class ValueClamp(DiscreteVariable):
    " A discrete variable that take on value_list at given times"
    
    def __init__(self,name,times,value_list, minimize_dt = False):
        """ Instantiate the ValueClamp

        @param name   : The unique name
        @param times  : A list of times, that starts with 0
        @param value_list : A list of value_list, with the same length at times with
                        correspondning value_list
        """
        
        super(ValueClamp,self).__init__(name)

        self._check_argument(times,value_list)

        #print times, value_list
        self._params["times"]  = times
        self._params["value_list"] = value_list
        self._params.minimize_dt = minimize_dt
        
        # Initialize some protected attributes
        self.reset()
        
        self._register_DiffsimObject(ValueClamp)

    _check_argument = check_argument

    def minimize_dt(self):
        # NOTE: If time is 0.0 and the first value is zero then do not reset
        # This is implemented so the time is _not_ reseted for initial time
        # when a typical state is zero. One could potentialy imagine that
        # A voltage clamp is used and the initial value is set to 0, but
        # a voltage clamp should probably not reset the time anyway
        if self._t == 0.0 and self._value_list[0] == 0.0:
            return False
        return self._params.minimize_dt
    
    def reset(self):
        " Reset value_list to initial states"
        self._t = 0.
        self._ind = 0
        self._check_argument(self._params["times"], self._params["value_list"])
        self._value = self._value_list[0]
    
    def step(self):
        # Get a new value
        self._value = self._value_list[self._ind]

        # and increase the index
        self._ind += 1
        
        return True
    
    def next_event(self):
        # If we have come to the end of our index list, return inf
        if self._ind >= len(self._times):
            time = numpy.inf
        else:
            time = self._times[self._ind] - self._t
        #debug("ValueClamp: %f"%time)
        return time
    
    def update_time(self,dt):
        #debug("Update time: %f, dt: %f"%(self._t,dt))
        #debug("Next event: %f"%self._times[self._ind])
        self._t += dt
    
    def value(self,*args,**kwargs):
        return self._value

    @classmethod
    def default_params(cls):
        return ParameterDict(
            value_list = [0],
            times  = [0],
            minimize_dt = TypeCheck(False))
    
    step.__doc__        = DiscreteVariable.step.__doc__
    update_time.__doc__ = DiscreteVariable.update_time.__doc__
    next_event.__doc__  = DiscreteVariable.next_event.__doc__
    value.__doc__       = DiscreteVariable.value.__doc__

class VoltageStep(ValueClamp,Singleton):
    def __init__(self,times,value_list,minimize_dt=False):
        ValueClamp.__init__(self,"V",times,value_list,minimize_dt)
        self._register_DiffsimObject(VoltageStep)
    
class SigmoidalValueClampStepper(DiscreteVariable):
    " This class is a time stepper for a SigmoidalValueClamp"
    def __init__(self, SVC, max_step):
        assert(isinstance(SVC,SigmoidalValueClamp))
        assert(isinstance(max_step,float))
        assert(max_step > 0)
        super(SigmoidalValueClampStepper,self).__init__(SVC.name() + "_time_stepper")
        self._SVC        = SVC
        self._max_step   = max_step
        self._next_event = None
        self._register_DiffsimObject(SigmoidalValueClampStepper)

    def reset(self):
        " Reset the stepper."
        self._t = 0.0
        self._next_event = None
        self._max_step = self._SVC.get_params().max_step
        assert(isinstance(self._max_step,float))
        assert(self._max_step > 0)
        
    def next_event(self):
        " Return the time to next event."
        if self._next_event is None:
            self._next_event = min(self._max_step / abs(self._SVC.evaluate_der(self._t)),
                                   self._SVC.time_to_next_step(self._max_step))
        return self._next_event
        
    def step(self):
        " A dummy function completing the interface"
        return True
        
    def update_time(self,dt):
        " Update the the time"
        self._next_event = None
        self._t += dt

class SigmoidalValueClamp(ContinuousVariable):
    """ This class provides a SigmoidalValueClamp,

    The transitions between the value_list are done in a smooth sigmoidal way.
    """
    def __init__(self, name, times, value_list, sigma, max_step = None):
        """ Instantiate the SigmoidalValueClamp

        @param name   : The unique name
        @param times  : A list of times, that starts with 0
        @param value_list : A list of value_list, with the same length at times with
                        correspondning value_list
        @param sigma  : Time parameter defining the slope of the sigmoidal curve
        @max_step     : The maximal step a SigmoidalValueClamp can take, optional
        """
        self._check_argument(times,value_list)
        super(SigmoidalValueClamp,self).__init__(name)
        self._params["times"]  = times
        self._params["value_list"] = value_list
        self._params["sigma"]  = self._sigma = sigma
        
        self._delta_value_list = self._value_list[1:] - self._value_list[0:-1]

        # Two list for storing of value_list and times
        self._stored_value_list = []
        self._stored_times  = []

        self._values_0     = 0.
        self._values_1     = 0.
        self._values_2     = 0.

        # If max_step is set check the value and then create a
        # SigmoidalValueClampStepper
        self._stepper = None
        if not max_step is None:
            self._stepper = SigmoidalValueClampStepper(self, max_step)
            self._params.max_step = max_step
            
        # Initialize some protected attributes
        self._t = 0.
        self._register_DiffsimObject(SigmoidalValueClamp)
    
    def default_params(self):
        " Returns the default parameters" 
        return ParameterDict(
            value_list = [0],
            times  = [0],
            sigma = RangeCheck(0.1,gt=0),
            max_step = None)
    
    def reset(self):
        " Reset value_list to initial states"
        self._t = 0.
        self._check_argument(self._params["times"], self._params["value_list"])
        self._sigma = self._params.sigma
        
        # If there are stored value_list and times from previous run, overwrite them
        self._stored_value_list = [self._value_list[0]]
        self._stored_times  = [0.0]
        
        self._value_list_1 = self._value_list[0]
    
    def time_to_next_step(self,max_step):
        " Return the time to the next step "
        from math import log
        from numpy import inf
        ind_next = ((self._times-self._t) > 0).argmax()
        # If self._times-self._t are all negative, i.e., if self._t is larger
        # than the largest self._times
        if ind_next == 0:
            return inf
        
        dV = abs(self._delta_value_list[ind_next-1])
        assert(dV > max_step)
        start_of_next_step = self._times[ind_next] - self._sigma*log(dV/max_step-1.)
        # If we are beyond the start of next time step
        if self._t >= start_of_next_step:
            ind_next += 1
            if ind_next < len(self._times):
                dV = abs(self._delta_value_list[ind_next-1])
                assert(dV > max_step)
                start_of_next_step = self._times[ind_next] - self._sigma*log(dV/max_step-1.)
            else:
                return inf 
        return start_of_next_step - self._t
    
    def _sigmoid(self,t_loc):
        " Return the value of the sigmoidal function"
        from math import exp
        return 1./(1+exp(min(100,-t_loc/self._sigma)))
    
    def _sigmoid_der(self,t_loc):
        " Return the derivative of the sigmoidal function"
        from math import exp
        return exp(min(100,-t_loc/self._sigma))/self._sigma*self._sigmoid(t_loc)**2
    
    def t(self):
        " Return the" 
        return self._t
    
    def value(self,*args,**kwargs):
        " Return the value"
        return self._value_list_1
        
    def evaluate(self, t):
        " Evaluate the value of the sigmoidal value clamp at the present t"
        ret = self._value_list[0]
        for i, dV in enumerate(self._delta_value_list):
            t_loc = t-self._times[i+1]
            ret += dV*self._sigmoid(t_loc)
        return ret
    
    def evaluate_der(self, t):
        ret = 0.0
        for i, dV in enumerate(self._delta_value_list):
            t_loc = t-self._times[i+1]
            ret += dV*self._sigmoid_der(t_loc)
        return ret
    
    def solve(self, dt, iteration=0):
        " A dummy solve to just compute the next value of the SigmoidalValueClamp"
        if iteration > 0:
            return
        
        # Take the time one step further
        self._t += dt
        self._converged = True
        self._value_list_1 = self.evaluate(self._t)
        self._stored_times.append(self._t)
        self._stored_value_list.append(self._value_list_1)
        
    def finalize(self):
        " Return the stored information as numpy arrays."
        return dict(times  = numpy.array(self._stored_times),
                    value_list = numpy.array(self._stored_value_list))
    
    _check_argument = check_argument

class SigmoidalVoltageStep(SigmoidalValueClamp,Singleton):
    def __init__(self,times,value_list,sigma,max_step=None):
        SigmoidalValueClamp.__init__(self,"V",times,value_list,sigma,max_step)
        self._register_DiffsimObject(SigmoidalVoltageStep)


class DynamicValueClampStepper(DiscreteVariable):
    " This class is a time stepper for a SigmoidalValueClamp"
    def __init__(self, DVC):
        assert(isinstance(DVC,DynamicValueClamp))
        super(DynamicValueClampStepper,self).__init__(DVC.name() + "_time_stepper")
        self._DVC        = DVC
        self._next_event = None
        self._register_DiffsimObject(DynamicValueClampStepper)

    def reset(self):
        " Reset the stepper."
        self._t = 0.0
        self._next_event = None
        
    def next_event(self):
        " Return the time to next event."
        if self._next_event is None:
            self._next_event = self._DVC.time_to_next_step()
        return self._next_event
        
    def step(self):
        " A dummy function completing the interface"
        return True
        
    def update_time(self,dt):
        " Update the the time"
        self._next_event = None
        self._t += dt

class DynamicValueClamp(ContinuousVariable):
    def __init__(self, name, times, value_list, offset = 0.0, max_step=None):
        """ Instantiate the DynamicsValueClamp

        @param name   : The unique name
        @param times  : A list of times, that starts with 0
        @param value_list : A list of value_list, with the same length at times with
                        correspondning value_list
        @param offset : A time offset for the beginning of the dynamic value
                        clamp series, must be positive
        @max_step     : The maximal step a DynamicValueClamp can take, optional
        """
        
        self._check_argument(times,value_list,False)
        assert(isinstance(offset,float))
        self._times += offset
        
        self._value_diff_cumsums = numpy.concatenate((numpy.zeros(1),numpy.absolute(numpy.diff(self._value_list)).cumsum()))

        super(DynamicValueClamp,self).__init__(name)
        
        self._params["times"]  = times
        self._params["value_list"] = value_list
        self._params["offset"] = offset
        
        # Two list for storing of value_list and times
        self._stored_value_list = []
        self._stored_times  = []
        
        self._values_0     = 0.
        self._values_1     = 0.
        self._values_2     = 0.

        # If max_step is set check the value and then create a
        # DynamicsValueClampStepper
        self._stepper = None
        if not max_step is None:
            self._stepper = DynamicValueClampStepper(self)
            self._params.max_step = max_step
            
        # Initialize some protected attributes
        self._t = 0.
        self._register_DiffsimObject(DynamicValueClamp)
    
    def default_params(self):
        " Returns the default parameters" 
        return ParameterDict(
            value_list = [0],
            times  = [0],
            offset = RangeCheck(0.0,ge=0),
            max_step = None)
        
    def reset(self):
        " Reset value_list to initial states"
        self._t = 0.
        self._check_argument(self._params["times"], self._params["value_list"],False)
        
        self._value_diff_cumsums = numpy.concatenate((numpy.zeros(1),numpy.absolute(numpy.diff(self._value_list)).cumsum()))
        
        # If there are stored value_list and times from previous run, overwrite them
        self._stored_value_list = [self._value_list[0]]
        self._stored_times  = [0.0]

        self._value_list_0 = self._value_list[0]
        self._value_list_1 = self._value_list_0
        self._value_list_2 = self._value_list_0
    
    def time_to_next_step(self):
        " Return the time to the next step "
        
        # If the present time is smaller than the first registered time
        if self._t < self._times[0]:
            debug("t smaller than offset")
            ttns = self._times[0] - self._t
        else:
            ttns = numpy.interp([self._params.max_step],self._value_diff_cumsums,self._times)[0] - self._t
            if ttns == 0.0:
                ttns = numpy.inf
            
        debug("Time to next step: %s"%str(ttns))
        return ttns
        
    def t(self):
        " Return the" 
        return self._t
    
    def value(self,*args,**kwargs):
        " Return the value"
        return self._value_list_1

    def evaluate(self, t):
        " Return the value for time t"
        #print "t", t, "self._times", self._times, "self._value_list", self._value_list
        return numpy.interp([t], self._times, self._value_list)[0]
        
    def solve(self, dt, iteration=0):
        " A dummy solve to just compute the next value of the DynamicValueClamp"
        if iteration > 0:
            return
        
        #steady_state = super(DynamicValueClamp,self).solve(dt, iteration, relax)
        #if steady_state:
        #    return
        
        self._converged = True
        
        # Take the time one step further
        self._t += dt
        self._value_list_1 = self.evaluate(self._t)
        self._stored_times.append(self._t)
        self._stored_value_list.append(self._value_list_1)
        # Taking the delta based on the cumsums will eliminate errors done when
        # derivatives of value_list changes sign
        delta = numpy.interp([self._t],self._times,self._value_diff_cumsums)[0]
        if not self._params.max_step is None and delta > 0.0:
            self._value_diff_cumsums -= delta

        debug("Voltage: %s"%str(self._value_list_1))
        
    def finalize(self):
        " Return the stored information as numpy arrays."
        return dict(times  = numpy.array(self._stored_times),
                    value_list = numpy.array(self._stored_value_list))
        
    _check_argument = check_argument
        
class DynamicVoltageClamp(DynamicValueClamp, Singleton):
    def __init__(self, times, value_list, offset=0.0, max_step=None):
        DynamicValueClamp.__init__(self, "V", times, value_list, offset, max_step)
        self._register_DiffsimObject(DynamicVoltageClamp)
        
debug("%s loaded"%__name__)
