#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = "Johan Hake (johan.hake@gmail.com)"
__copyright__ = "Copyright (C) 2008 Johan Hake"
__license__  = "GNU LGPL Version 3.0 or any later"
__date__ = "2008-08-19 -- 2010-01-18"

import unittest

from diffsim import *
import numpy, os

# Disable log output
from diffsim.logger      import *
set_log_level(WARNING,ALL)
set_log_level(WARNING,'instant')

from dolfin import *
logging(False)

mesh_file = "mesh/test_mesh.xml"
d = DiffusionDomain("Test",mesh_file)

class DiffusionDomainTestCase(unittest.TestCase):

    def setUp(self):pass
    
    def test_duplicate(self):
        self.assertRaises(TypeError, DiffsimObject, "Test")
    
    def test_attributes(self):
        self.assertEqual(d.name(),"Test")
        self.assertEqual(len(d.demands()), 0)
        self.assertEqual(len(d.default_params()), 3)
    
    def test_face_markers(self):
        for m in range(1, 8):
            d.check_face_marker(m)
        self.assertRaises(RuntimeError,d.check_face_marker, 8)

    def test_vertex_markers(self):
        for m in range(1, 8):
            d.check_vertex_marker(m)
        self.assertRaises(RuntimeError, d.check_vertex_marker, 8)
    
    def test_dofs(self):
        self.assertEqual(d.num_dofs(), 2382)
        z1 = d.zero_dofs()
        z2 = d.zero_dofs(copy=True)
        self.assertNotEqual(id(z1), id(z2))
        self.assertTrue(isinstance(z1, cpp.GenericVector))
        o1 = d.one_dofs()
        o2 = d.one_dofs(copy=True)
        self.assertNotEqual(id(o1), id(o2))
        self.assertTrue(isinstance(o1, cpp.GenericVector))
        self.assertAlmostEqual(d.one_dofs().sum(), d.num_dofs())
        
    def test_assemble(self):
        d.assemble_tensor("mass boundary")
        s = d.assemble_tensor("stiffnes", return_tensor=True)
        self.assertEqual((2382,2382), (s.size(0),s.size(1)))
        for m in range(8):
            mb1 = d.assemble_tensor("mass boundary", marker=m, return_tensor=True)
            mb2 = d.assemble_tensor("mass boundary", marker=m, return_tensor=True,
                                    reassemble=True)
            self.assertNotEqual(id(mb1), id(mb2))
            self.assertAlmostEqual(((mb1-mb2)*d.one_dofs()).sum(), 0.0)
    
    def test_tensor(self):
        self.assertEqual(id(d.tensor("stiffnes")),
                         id(d.assemble_tensor("stiffnes", return_tensor=True)))
    
    def test_domain_volume(self):
        self.assertAlmostEqual(d.domain_volume(), 0.84375)
    
    def test_domain_area(self):
        self.assertAlmostEqual(d.domain_area(), 7.75)
        self.assertAlmostEqual(d.domain_area(1), 1.0)
        self.assertAlmostEqual(d.domain_area(7), 0.25)
    
    def test_save_field(self):
        u = numpy.arange(d.num_dofs(), dtype='d')
        d.save_field("test_field", u)
        f = File("test_field.xml")
        v = Vector()
        f >> v
        if os.path.isfile("test_field.xml"):
            os.unlink("test_field.xml")
        self.assertAlmostEqual(u[0], v[0])
        self.assertAlmostEqual(u[131], v[131])
        
    def test_rename(self):
        self.assertRaises(RuntimeError, d.rename, "Jada")
        
if __name__ == "__main__":
    unittest.main()
